!!>****h* 
!! SYNOPSIS:  
!!   ECC code utility on GPU
!!   
!! CREATOR: 
!!      Hoang-Trong Minh Tuan 
!! AFFILIATION: 
!!      George Mason University 
!! 
!! HISTORY
!!    Rev: 0.1 
!!    0.2: add sum reduction on GPU, min/max on GPU
!!    0.3: add random number generator on GPU using SARU (Hoang-Trong Minh Tuan - Oct, 18, 2010)
!!    0.4: rename from spark_gpu_utility to ecc_gpu_utility
!!    0.5: working version on Fermi (Tuan, Nov 30, 2010)
!!    0.6: add CaB_htrpn (Tuan, Dec 29, 2010)
!!    0.7: fix bug on I-clamp (Tuan,Feb, 01, 2010)
!!    0.8: revise the code, work with PGI 11.3 only (Tuan, Jul, 11, 2011)
!!    0.9: fix bug with PRNG working with PGI 11.7 (Jafri+Tuan, Jul 25,2011)
!!    1.0: rename data to reflect using the new algorithm (Mar 3rd, 2012 - Tuan)
!!    1.1: fix I_Ktos, I_Ktof, I_K1 formula (Tuan - 2013-Nov-03)
!!    1.2: Add Sun2000 model (use new algorithm) (Tuan - 2013-Nov-05)
!!         Add CaCM to I/O (_IO_2_HDF2) -
!!    1.3: fix number of records in the last I/O (Tuan - 2013-Nov-06)
!!         fix Sun2000 LCC model
!!    1.4: allocate_Data() function to allocate data (Tuan - 2013-Nov-18)
!!         reset_Data() function to reinitialize data
!!    1.5: revise save_summarized_data() --> save_summarized_data_1, save_summarized_data_2, save_summarized_data_core
!!         revise spark_summarize() --> spark_summarize_1: accept OPTIONAL parameter (file2save)
!!                                      spark_summarize_2: accept INTEGER parameter as fileID, and OPTIONAL time_duration
!!         (Tuan, 2013-Nov-21)
!!    1.6: fix the way we calculate GAIN (Tuan, 2013-Dec-02)
!!    1.7: add PEAK_GAIN, and modify the existing GAIN to INTEGRAL_GAIN (Tuan, 2013-Dec-03)
!!_****
MODULE ecc_gpu_utility
  USE precisioninfo
  USE data_global
  USE data_ecc
  USE data_ecc_gpu
  USE rand_utility
  USE hdf5_utility
  USE iso_c_binding
  USE cudafor
  USE cuda_interop
  USE error_codes
  IMPLICIT NONE
#include "macro.h"

#define _GPU_MODEL_ _FERMI
#define _USE_PRNG_ _PRNG_SARU

  
#define _BENCHMARK_ _NO
#define _SAFE_MODE_ _NO
  ! DEBUG
#define _DEBUG_   _NO

#if _TEST_NEW_ORPHAN_CRU_ == _YES 
  #define _USE_CaDS_APPROX_ _NO
#endif
  

  INTERFACE save_summarized_data
     MODULE PROCEDURE save_summarized_data_1
     MODULE PROCEDURE save_summarized_data_2
  END INTERFACE
  INTERFACE spark_summarize
     MODULE PROCEDURE spark_summarize_1
     MODULE PROCEDURE spark_summarize_2
  END INTERFACE
  
  INTEGER :: CHUNK_SIZE   !number of records
  INTEGER :: CHUNKS_PER_FILE = 1 ! set to 1 for now
  INTEGER :: FIELDS_PER_RECORD = 0
  REAL(KIND=dp), DIMENSION(:, :), ALLOCATABLE :: data_buffer, data_2restartfile
  INTEGER :: records_used = 0 !keeping track how many record have been filled, before performing data output  
  INTEGER :: file_index = 0
  INTEGER :: RF_FIELDS_PER_RECORD  !in restart file
  INTEGER, PARAMETER :: copy_interval=100 ! how often to copy data back/forth GPU-CPU

  INTEGER, DIMENSION(:), ALLOCATABLE, device :: CRUs_blocked_dev
  INTEGER, DIMENSION(:), ALLOCATABLE :: CRUs_blocked
#if _EMULATE_CONDITION_ == _CAFFEINE_TYPE1 || _EMULATE_CONDITION_ == _CAFFEINE_TYPE2
  ! use inside a kernel
  REAL(KIND=dp), constant :: const_x_phi = 1000
  LOGICAL :: is_caffeinedump_triggered = .FALSE.
  LOGICAL :: some_caffeineBound_CRU_notfullyopen = .TRUE. 
#endif
  REAL(KIND=dp), constant :: time_start_caffeinedump_dev, duration_caffeinedump_dev !sec
  !!REAL(KIND=dp), PARAMETER :: time_start_caffeinedump = 0.5d0, duration_caffeinedump = 10.0d0 !sec (moved to data_global.f95)

  INTEGER, DIMENSION(:), ALLOCATABLE ::  index_LCCcluster_numLCCopen
  INTEGER, DIMENSION(:), ALLOCATABLE ::  index_RYRcluster_numRYRopen
  INTEGER :: num_triggertime = 0
  LOGICAL :: is_all_CRU_open = .FALSE.
  LOGICAL :: is_opened_LCC = .FALSE.

  CHARACTER(350) :: fstring, file_out

  REAL(KIND=dp) :: Ca_srT ! total calcium in the SR (nSR + jSR)
  REAL(KIND=dp), ALLOCATABLE, DIMENSION(:), pinned:: Ca_ds, Ca_jsr
  REAL(KIND=dp), ALLOCATABLE, DIMENSION(:), pinned :: J_efflux_total, &
       J_refill_total, I_dhpr_total, J_ryr_total, J_ryr_orphan_total
  INTEGER, DIMENSION(:), ALLOCATABLE, pinned :: Nquarks, Nsparks, Nblinks

  REAL(KIND=dp) :: J_dhpr_total, INTEGRAL_GAIN, accumulate_J_ryr, accumulate_J_dhpr, accumulate_J_ryr_orphan,  &
       accumulate_J_SL_in, accumulate_J_SL_out, accumulate_J_SR_in, accumulate_J_SR_out, J_SRrelease_total
  REAL(KIND=dp) :: peak_J_SRrelease_total, peak_J_dhpr_total, PEAK_GAIN
  

  !+ APD
  INTEGER, PARAMETER :: NUM_APD_TYPES = 4 !APD90, APD50, APD20, APD0mV
  REAL(KIND=dp), DIMENSION(NUM_APD_TYPES) :: Vmthreshold  
  REAL(KIND=dp), DIMENSION(NUM_APD_TYPES) :: APD, APstart, APend
  LOGICAL, DIMENSION(NUM_APD_TYPES) :: is_inside_AP  
  INTEGER, DIMENSION(NUM_APD_TYPES) :: APcount
  REAL :: time_start_DiastolicPhase
  LOGICAL :: is_in_DiastolicPhase, is_done_firstbeat
  REAL(KIND=dp) :: dt !timestep
  
  !+ Spark/Quark/Blink analysis
  LOGICAL, DIMENSION(:), ALLOCATABLE, device :: flag1, flag2, flag3, flag4
   ! flag1 (spark/quark start), flag2 (blink start), flag3 (spark/quark rise), flag4 (true spark)
  REAL, DIMENSION(:), ALLOCATABLE, device :: ts_spark, ts_blink
  INTEGER, DIMENSION(:), ALLOCATABLE, device :: Nquarks_dev, Nsparks_dev, Nblinks_dev
  !REAL, DIMENSION(:,:), ALLOCATABLE, device :: SparkPeaks, SparkDurations, BlinkDurations
  REAL, DIMENSION(:), ALLOCATABLE, device :: tot_SparkPeaks, tot_SparkDurations, tot_BlinkDurations !total
  REAL, DIMENSION(:), ALLOCATABLE, device :: tot_SparkTime2Peak
  !NTEGER, PARAMETER :: MAX_NUM_QUARKS_PER_SEC = 10000, MAX_NUM_SPARKS_PER_SEC = 1000
  INTEGER, DIMENSION(:), ALLOCATABLE, device :: prev_RYRgate
  REAL, DIMENSION(:), ALLOCATABLE, device :: prev_Ca_ds  

  INTEGER :: num_Sparks, num_Blinks
  REAL:: avg_SparkDuration, avg_BlinkDuration, avg_SparkPeak, &
       SFWHM, BFWHM  ! spark/blink full half-width maximum
  INTEGER:: num_Quarks
  REAL :: Spark_Freq
  REAL(KIND=dp) :: tsegment, time_prev1, time_prev2, time_start_spark_analysis, time_prev_fb, oldtime


  !+ device data
  REAL(KIND=dp), DIMENSION(:), ALLOCATABLE, device :: Ca_ds_dev, Ca_jsr_dev
  REAL(KIND=dp), DIMENSION(:), ALLOCATABLE, device :: X_r_dev ! vector keeps random-numbers

  !!+ cuda stream
  INTEGER :: stream0, stream1

  !+ device constant data
 #if _LCC_MODEL_ == _WILLIAMS2013 || _LCC_MODEL_ == _WILLIAMS2013_NEWALGORITHM
  INTEGER, PARAMETER :: SIZE_INT_SHARE = 15, SIZE_DP_SHARE = 20
 #elif _LCC_MODEL_ == _SUN2000
  INTEGER, PARAMETER :: SIZE_INT_SHARE = 15, SIZE_DP_SHARE = 25
  ! CaCM = calcium-bound calmodulin, CM   = calmodulin
  REAL(KIND=dp), DIMENSION(:), ALLOCATABLE, device :: CaCM_ds_dev
 #endif
  
  INTEGER, DIMENSION(SIZE_INT_SHARE), constant :: int_ca ! constant array
  REAL(KIND=dp), DIMENSION(SIZE_DP_SHARE), constant :: dp_ca

  REAL(KIND=dp), DIMENSION(:,:), ALLOCATABLE, device :: lcc_compKv_act_dev, lcc_compKv_inact_dev, &
       lcc_compKfree_dev, lcc_compKcal_inact_dev
  INTEGER, DIMENSION(:,:), ALLOCATABLE, device :: lcc_indxKv_act_dev, lcc_indxKv_inact_dev, &
       lcc_indxKfree_dev, lcc_indxKcal_inact_dev
  
  REAL(KIND=dp), DIMENSION(:,:), ALLOCATABLE, device :: ryr_compKfree_dev, ryr_compKcal_act_dev
  INTEGER, DIMENSION(:,:), ALLOCATABLE, device :: ryr_indxKfree_dev, ryr_indxKcal_act_dev

  REAL(KIND=dp), DIMENSION(:,:), ALLOCATABLE, device :: lcc_Ktransitionrate_dev, lcc_Krateconstant_dev, lcc_compK_dev
  INTEGER, DIMENSION(:,:), ALLOCATABLE, device :: lcc_stateFromTo_dev, lcc_indxK_dev, lcc_StateSpace_dev

  !! old: size (NSFU, maxnklm+maxnklp-1)
  !! new method: size (NSFU, 1)
  REAL(KIND=dp), ALLOCATABLE, DIMENSION(:), device :: compPdt_dev
#if _DEBUG_ == _YES
  REAL(KIND=dp), ALLOCATABLE, DIMENSION(:) :: compPdt
#endif  
  ! size (NSFU, maxnklm+maxnklp-1)
  INTEGER, ALLOCATABLE, DIMENSION(:, :), device :: indexPdt_dev
  ! size (irow_R)
  REAL(KIND=dp), ALLOCATABLE, DIMENSION(:), device ::  U_Ro_dev, U_Lo_dev
  ! size (NSFU)  - keep current state index at each SFU
  INTEGER, ALLOCATABLE, DIMENSION(:), device ::  CRU_states_dev
  ! size (NSFU)
  REAL(KIND=dp), ALLOCATABLE, DIMENSION(:), device :: RYRgate_dev, DHPRgate_dev
  ! size (NSFU)
  REAL(KIND=dp), ALLOCATABLE, DIMENSION(:), device :: J_efflux_dev
  REAL(KIND=dp), ALLOCATABLE, DIMENSION(:), device :: J_refill_dev
  REAL(KIND=dp), ALLOCATABLE, DIMENSION(:), device :: I_dhpr_dev
  REAL(KIND=dp), ALLOCATABLE, DIMENSION(:), pinned :: I_dhpr
  !REAL(KIND=dp), ALLOCATABLE, DIMENSION(:), device :: J_dhpr_dev, J_ryr_dev
  REAL(KIND=dp), ALLOCATABLE, DIMENSION(:), device :: J_ryr_dev
!!!!!!!
!REAL(KIND=dp), ALLOCATABLE, device :: new_dev  
REAL(kind=dp)::X_00,new



  !+buffering
  REAL(KIND=dp) :: beta_myo
  !+voltage
  REAL(KIND=dp) :: Vm &    ! membrane voltage
       , E_Ca, E_Na, E_K  ! reversal potential
  !+ ionic currents
  REAL(KIND=dp) :: I_ncx, I_Na, I_K1, I_bK, I_bCa, I_NaK, I_pmca, &
       I_app, I_bNa, I_Ktof, I_Kss, I_Ktos, I_dhpr_nj_T
  !+ global calcium ionic fluxes
  REAL(KIND=dp) :: J_ncx, J_serca, J_ryr_nj_T, J_Ca_sarmem
  REAL(KIND=dp) ::  J_leak, J_CaTrpn, J_ryr_orphan
  REAL(KIND=dp) :: J_CaCM, J_CaSL, J_CaSR, J_CaF
  INTEGER :: NUM_STATES_NJ
  INTEGER :: NUM_FIELDS4CURRENTS, NUM_INDIVIDUAL_FIELDS

  REAL(KIND=dp) :: CaB_htrpn, CaF, CaCM, CaSL, CaSR
  
  !+ summary data
  INTEGER, PARAMETER :: NUM_ENTRIES_DATASUMMARY=100
  REAL(KIND=dp), DIMENSION(NUM_ENTRIES_DATASUMMARY) :: summary_data ! to store GAIN, APD...
  INTEGER :: count_summarydata,BCL1
  CHARACTER(40), DIMENSION(NUM_ENTRIES_DATASUMMARY) :: description_summarydata
  
  REAL(KIND=dp), DIMENSION(:), ALLOCATABLE :: p_nj
  CHARACTER(8)  :: date
  INTEGER,DIMENSION(8) :: values
  
#if _DATAMODE_ == _IO_2_HDF || _DATAMODE_ == _IO_2_HDF2
  TYPE(HDFFILE) :: file_struct
#endif

#if _TEST_ORPHAN_CRU_ == _YES
  REAL(KIND=dp), DIMENSION(:), ALLOCATABLE :: pR_orphan, tmp_pR_orphan
  REAL(KIND=dp), DIMENSION(:), ALLOCATABLE :: chiC_orphan, chiO_orphan
#endif
  LOGICAL :: is_being_clamped

#if _USE_PRNG_ == _PRNG_SARU
  TYPE(c_ptr) :: saruObj
#elif _USE_PRNG_ == _PRNG_MT
  TYPE(c_ptr) :: d_status
  TYPE(c_ptr) :: d_texture
#elif _USE_PRNG_ == _PRNG_MTCPU
  INTEGER, DIMENSION(:), ALLOCATABLE :: mt_STATE(:)
  REAL(KIND=DP) :: mt_LB, mt_UB   !lower-bound, upper-bound mersenne twister
  INTEGER :: INFO
  REAL(KIND=dp), DIMENSION(:), ALLOCATABLE :: X_r
#elif _USE_PRNG_ == _PRNG_CURAND
  !!!! CURAND
#endif

#if _RUNMODE_ == _MODE_V_CLAMP && _INCREMENT_VSTIM_ == _YES    
  !!!!!!!!!!!!!!!! TEST V-stim !!!!!!!!!!!!!!!!!!!!!!!
  !! run and incrementally increase Voltage +10mV every second
  !! use only with _MODE_V_CLAMP
  REAL :: tick_time
  INTEGER :: fileID_gain_analysis
  CHARACTER(100), PARAMETER :: filestring_gain_analysis = "Vstim_gain.txt"
  
  !!!!!!!!!!END    TEST V-stim !!!!!!!!!!!!!!!!!!!!!!!
#endif
  
  REAL(kind=dp), PARAMETER :: ramp_time_reference = 1.0d-3  ! should be less than 1ms (from -80mV to +10mV)
  REAL(KIND=dp) :: ramp_time

  TYPE(c_ptr) :: odata_d
  REAL(KIND=dp) :: dp_arg1, dp_arg2
   
  SAVE
CONTAINS
  !!>****f* ecc_gpu_utility/set_device_constant
  !! SYNOPSIS  
  !!   Copy the data to device constant memory 
  !!   Assign frequently used read-only data to device constant memory
  !! INPUTS
  !!   None
  !! OUTPUTS
  !!   None
  !! SIDE EFFECTS
  !!   None
  !! HISTORY 
  !! BUGS 
  !! TODO 
  !! SEE ALSO 
  !!_****
  SUBROUTINE set_device_constant()
    IMPLICIT NONE
  #if _RUNMODE_ == _MODE_V_CLAMP_TEST_LCC
    v_ryr = 0.0d0
  #endif
    
  #if _LCC_MODEL_ == _WILLIAMS2013 
    int_ca = (/ NSFU, N_R, ryr_Ncol_Kcal_act, ryr_Ncol_Kfree, lcc_Ncol_Kv_act, & !5
         lcc_Ncol_Kv_inact, lcc_Ncol_Kcal_inact, lcc_Ncol_Kfree, lcc_Ncol_Kv_act+lcc_Ncol_Kv_inact+lcc_Ncol_Kfree+ryr_Ncol_Kfree+lcc_Ncol_Kcal_inact+ryr_Ncol_Kcal_act-5, irow_R, & ! 10
         0, 0, 0, 0, 0 & ! 15
         /)
  #elif _LCC_MODEL_ == _SUN2000 || _LCC_MODEL_ == _WILLIAMS2013_NEWALGORITHM
    int_ca = (/ NSFU, N_R, ryr_Ncol_Kcal_act, ryr_Ncol_Kfree, 0, & !5
         0, 0, 0, 0, irow_R, & ! 10
         0, 0, 0, 0, 0 & ! 15
         /)
  #endif
  #if _LCC_MODEL_ == _WILLIAMS2013  || _LCC_MODEL_ == _WILLIAMS2013_NEWALGORITHM
    !beta_ds : 12 --> 6    
    !replace 0.5d0 by eta_CO and (1-eta_CO)
    dp_ca = (/ -Ej*eta_CO, Ecc, Eoo, k_jsr0, k_jsr1, & ! 5
         beta_ds, inv_lambda_ds, inv_lambda_jsr, v_ryr, v_refill, & ! 10
         v_efflux, -Ej*(1.0-eta_CO), P_dhpr, eta_RYR , eta_LCC, &  ! 15
         F2zCa2_RT, inv_2FVmyo, 0.341d0*Ca_o, Km_jsr, B_jsrT &  ! 20
         /)
  #elif _LCC_MODEL_ == _SUN2000
    !(21) : k+_CM
    !(22) : k-_CM
    !(23) : CM_tot
    !(12) : eta_CaCM
!    dp_ca = (/ -Ej*eta_CO, Ecc, Eoo, 2.0*k_jsr0, k_jsr1, & ! 5
!         beta_ds, inv_lambda_ds, inv_lambda_jsr, v_ryr, v_refill, & ! 10
!         v_efflux, -Ej*(1.0-eta_CO), 1.4*P_dhpr, eta_RyR , eta_LCC, &  ! 15
!         F2zCa2_RT, inv_2FVmyo, 0.341d0*Ca_o, Km_jsr, B_jsrT, &  ! 20
!         kp_CM, km_CM, CM_tot, eta_CaCM, 0.0d0 & ! 25
!         /)
 
    dp_ca = (/ -Ej*eta_CO, Ecc, Eoo, k_jsr0, k_jsr1, & ! 5
         beta_ds, inv_lambda_ds, inv_lambda_jsr, v_ryr, v_refill, & ! 10
         v_efflux, -Ej*(1.0-eta_CO), P_dhpr, eta_RyR , eta_LCC, &  ! 15
         F2zCa2_RT, inv_2FVmyo, 0.341d0*Ca_o, Km_jsr, B_jsrT, &  ! 20
         kp_CM, km_CM, CM_tot, eta_CaCM, 0.0d0 & ! 25
         /)
  #endif    
  END SUBROUTINE set_device_constant

  SUBROUTINE allocate_Data()
    IMPLICIT NONE
    LOGICAL :: plog
    INTEGER :: iserror
    INTEGER :: res

    !! memory allocation
    !+ device
    ALLOCATE(ryr_compKfree_dev(irow_R,0:ryr_Ncol_Kfree), &
         ryr_compKcal_act_dev(irow_R,0:ryr_Ncol_Kcal_act), STAT=iserror)
    IF (iserror) THEN
       WRITE(error_message, *) "FAILED: device memory allocation ryr_compK"
       res = ERR_CUDA_ALLOCATE_MEM
       RETURN
    END IF
    ALLOCATE(ryr_indxKfree_dev(irow_R, 0:ryr_Ncol_Kfree),  &
         ryr_indxKcal_act_dev(irow_R, 0:ryr_Ncol_Kcal_act), STAT=iserror)
    IF (iserror) THEN
       WRITE(error_message, *) "FAILED: device memory allocation ryr_indxK"
       res = ERR_CUDA_ALLOCATE_MEM
       RETURN
    END IF
    
  #if _LCC_MODEL_ == _WILLIAMS2013    
    ALLOCATE(lcc_compKv_act_dev(irow_L,0:lcc_Ncol_Kv_act), &
         lcc_compKv_inact_dev(irow_L,0:lcc_Ncol_Kv_inact), &
         lcc_compKcal_inact_dev(irow_L,0:lcc_Ncol_Kcal_inact) , &
         lcc_compKfree_dev(irow_L,0:lcc_Ncol_Kfree), &
         STAT=iserror)
    IF (iserror) THEN
       WRITE(error_message, *) "FAILED: device memory allocation lcc_compK"
       res = ERR_CUDA_ALLOCATE_MEM
       RETURN
    END IF
    ALLOCATE(lcc_indxKv_act_dev(irow_L, 0:lcc_Ncol_Kv_act), &
         lcc_indxKv_inact_dev(irow_L, 0:lcc_Ncol_Kv_inact), lcc_indxKcal_inact_dev(irow_L, 0:lcc_Ncol_Kcal_inact), &
         lcc_indxKfree_dev(irow_L, 0:lcc_Ncol_Kfree), STAT=iserror)
    IF (iserror) THEN
       WRITE(error_message, *) "FAILED: device memory allocation ryr_indxK"
       res = ERR_CUDA_ALLOCATE_MEM
       RETURN
    END IF
  #elif _LCC_MODEL_ == _SUN2000 || _LCC_MODEL_ == _WILLIAMS2013_NEWALGORITHM
    ALLOCATE(lcc_Ktransitionrate_dev(mL,mL), STAT=iserror)
    IF (iserror) THEN
       WRITE(*, *) "FAILED: device memory allocation lcc_Ktransitionrate_dev"
       res = ERR_CUDA_ALLOCATE_MEM
       RETURN
    END IF
    ALLOCATE(lcc_Krateconstant_dev(mL,mL), STAT=iserror)
    IF (iserror) THEN
       WRITE(*, *) "FAILED: device memory allocation lcc_Krateconstant_dev"
       res = ERR_CUDA_ALLOCATE_MEM
       RETURN
    END IF
    ALLOCATE(lcc_stateFromTo_dev(1:irow_L,0:lcc_maxNumNeighbors), lcc_indxK_dev(1:irow_L,0:lcc_maxNumNeighbors), lcc_StateSpace_dev(1:irow_L, 1:mL), STAT=iserror)
    IF (iserror) THEN
       WRITE(*, *) "FAILED: device memory allocation lcc_stateFromTo_dev, lcc_indxK_dev, lcc_StateSpace_dev"
       res = ERR_CUDA_ALLOCATE_MEM
       RETURN
    END IF
    ALLOCATE(lcc_compK_dev(NSFU,lcc_maxNumNeighbors), STAT=iserror)
    IF (iserror) THEN
       WRITE(*, *) "FAILED: device memory allocation lcc_compK_dev"
       res = ERR_CUDA_ALLOCATE_MEM
       RETURN
    END IF
    lcc_compK_dev = 0.0d0
  #endif

    ALLOCATE(U_Ro_dev(irow_R), U_Lo_dev(irow_L), STAT = iserror)
    IF (iserror) THEN
       WRITE(error_message, *) "FAILED: device memory allocation U_Ro, U_Lo "
       res = ERR_CUDA_ALLOCATE_MEM
       RETURN
    END IF

    ALLOCATE(Ca_ds_dev(NSFU), Ca_jsr_dev(NSFU), STAT=iserror)
    IF (iserror) THEN
       WRITE(error_message, *) "FAILED: device memory allocation Cajsr, Cads "
       res = ERR_CUDA_ALLOCATE_MEM
       RETURN
    END IF

  #if _LCC_MODEL_ == _SUN2000
    ALLOCATE(CaCM_ds_dev(NSFU), STAT=iserror)
    IF (iserror) THEN
       WRITE(*, *) "FAILED: device memory allocation CaCM_ds_dev "
       res = ERR_CUDA_ALLOCATE_MEM
       RETURN
    END IF    
  #endif
    
    ALLOCATE(J_efflux_dev(NSFU), J_refill_dev(NSFU), J_ryr_dev(NSFU), STAT=iserror)
    IF (iserror) THEN
       WRITE(error_message, *) "FAILED: device memory allocation fluxes Jrefill, Jefflux, Jryr"
       res = ERR_CUDA_ALLOCATE_MEM
       RETURN
    END IF

    !ALLOCATE(I_dhpr_dev(NSFU), J_dhpr_dev(NSFU), J_ryr_dev(NSFU))
    ALLOCATE(I_dhpr_dev(NSFU), STAT=iserror)
    IF (iserror) THEN
       WRITE(error_message, *) "FAILED: device memory allocation Idhpr"
       res = ERR_CUDA_ALLOCATE_MEM
       RETURN
    END IF

    ALLOCATE(RYRgate_dev(NSFU),  DHPRgate_dev(NSFU), STAT=iserror)
    IF (iserror) THEN
       WRITE(error_message, *) "FAILED: device memory allocation RYRgate, DHPRgate"
       res = ERR_CUDA_ALLOCATE_MEM
       RETURN
    END IF

    ALLOCATE(CRU_states_dev(NSFU), STAT = iserror)
    IF (iserror) THEN
       WRITE(error_message, *) "FAILED: device memory allocation state indices"
       res = ERR_CUDA_ALLOCATE_MEM
       RETURN
    END IF

    
    ALLOCATE(compPdt_dev(NSFU), STAT = iserror)
    IF (iserror) THEN
       WRITE(error_message, *) "FAILED: device memory allocation compPdt"
       res = ERR_CUDA_ALLOCATE_MEM
       RETURN
    END IF
    
    ALLOCATE(X_r_dev(NSFU*copy_interval), STAT = iserror)  ! for random numbers
    IF (iserror) THEN
       WRITE(error_message, *) "FAILED: device memory allocation X_r random numbers"
       res = ERR_CUDA_ALLOCATE_MEM
       RETURN
    END IF

    !+ pinned memory
    ALLOCATE(Ca_ds(NSFU), Ca_jsr(NSFU), PINNED=plog)
    IF (.NOT. plog) THEN
       WRITE(log_string,*) "FAILED: Ca_ds, Ca_jsr on PINNED memory"
       CALL write2log(log_string)
    END IF
    ALLOCATE(J_efflux_total(1), J_refill_total(1), I_dhpr_total(1), J_ryr_total(1), J_ryr_orphan_total(1), PINNED=plog)
    IF (.NOT. plog) THEN
       WRITE(log_string,*) "FAILED: J_efflux_total, J_refill_total, I_dhpr_total on PINNED memory"
       CALL write2log(log_string)
    END IF

    ALLOCATE(I_dhpr(NSFU), PINNED=plog)
    IF (.NOT. plog) THEN
       WRITE(log_string,*) "FAILED: I_dhpr on PINNED memory"
       CALL write2log(log_string)
    END IF
    
    !+ host memory
    ALLOCATE(CRU_states(NSFU), LCCcluster_states(NSFU), RYRcluster_states(NSFU))
    ALLOCATE(RYRgate(NSFU),  DHPRgate(NSFU))
    ALLOCATE(p_nj(irow_L_nj*irow_R_nj))

#if _DEBUG_ == _YES
    ALLOCATE(compPdt(NSFU))
#endif    

    !!+ data for spark analysis
    ALLOCATE(flag1(NSFU), flag2(NSFU), flag3(NSFU), flag4(NSFU))
    ALLOCATE(ts_spark(NSFU), ts_blink(NSFU))
    ALLOCATE(Nquarks_dev(NSFU), Nsparks_dev(NSFU), Nblinks_dev(NSFU))

    ALLOCATE(Nquarks(NSFU), Nsparks(NSFU), Nblinks(NSFU))

    ALLOCATE(tot_SparkPeaks(NSFU), tot_SparkDurations(NSFU), tot_BlinkDurations(NSFU))
    ALLOCATE(tot_SparkTime2Peak(NSFU))
    ALLOCATE(prev_RYRgate(NSFU), prev_Ca_ds(NSFU))

    !! this data is being used for caffeine OR trigger RYR simulation 
    IF (.NOT. ALLOCATED(index_RYRcluster_numRYRopen)) ALLOCATE(index_RYRcluster_numRYRopen(0:N_R))
    !! this data is being used for trigger LCC simulation 
    IF (.NOT. ALLOCATED(index_LCCcluster_numLCCopen)) ALLOCATE(index_LCCcluster_numLCCopen(0:N_L))
    
  END SUBROUTINE allocate_Data
  !!>****f* ecc_gpu_utility/init_Simulation
  !! SYNOPSIS  
  !!   Initialize data for the simulation
  !! INPUTS
  !!    None
  !! OUTPUTS
  !!    None
  !! RETURN 
  !! SIDE EFFECTS 
  !! HISTORY 
  !! BUGS 
  !! TODO 
  !! SEE ALSO 
  !!_****
  INTEGER FUNCTION init_Simulation() RESULT(res)
    IMPLICIT NONE

    LOGICAL :: plog
    INTEGER :: iserror

    WRITE (log_string, *) "start init_Simulation()"
    CALL write2log(log_string)
    PRINT *, log_string

    res = SUCCESS
    
    res = print_mode_info()

    CALL allocate_Data()

!!$    flag1 = .FALSE.
!!$    flag2 = .FALSE.
!!$    flag3 = .FALSE.
!!$    flag4 = .FALSE.
!!$    ts_spark = +0d0
!!$    ts_blink = +0d0
!!$    Nquarks_dev = 0
!!$    Nsparks_dev = 0
!!$    Nblinks_dev = 0
!!$    tot_SparkDurations = 0d0
!!$    tot_SparkPeaks = 0d0
!!$    tot_BlinkDurations = 0d0
!!$    tot_SparkTime2Peak = 0d0
    
    !copy from host to device
    ryr_compKcal_act_dev = ryr_compKcal_act
  !  Print*,ryr_compKcal_act(:,:)
    ryr_compKfree_dev = ryr_compKfree
    ryr_indxKcal_act_dev = ryr_indxKcal_act
    ryr_indxKfree_dev = ryr_indxKfree
! print*,ryr_Ncol_Kcal_act
#if _LCC_MODEL_ == _WILLIAMS2013
    lcc_compKv_act_dev = lcc_compKv_act
    lcc_compKv_inact_dev = lcc_compKv_inact
    lcc_compKfree_dev = lcc_compKfree
    lcc_compKcal_inact_dev = lcc_compKcal_inact
    lcc_indxKv_act_dev = lcc_indxKv_act
    lcc_indxKv_inact_dev = lcc_indxKv_inact
    lcc_indxKfree_dev = lcc_indxKfree
    lcc_indxKcal_inact_dev = lcc_indxKcal_inact
#elif _LCC_MODEL_ == _SUN2000 || _LCC_MODEL_ == _WILLIAMS2013_NEWALGORITHM
    lcc_stateFromTo_dev = lcc_stateFromTo
    lcc_indxK_dev = lcc_indxK
    lcc_StateSpace_dev = lcc_StateSpace
    lcc_Krateconstant_dev = lcc_Krateconstant
    lcc_Ktransitionrate = lcc_Krateconstant
#endif
!    new_dev=new    
    U_Ro_dev = U_Ro  !irow_R length
    U_Lo_dev = U_Lo  !irow_L length

    I_app = +0d0  ! uA/cm^2
    Vm = V_rest   ! mV
  !   X_00 =2.0

!    BCL1 = 1
!!$    !! some internal data
!!$    is_being_clamped = .FALSE.
!!$    is_inside_AP = .FALSE. 
!!$    is_EndSim = .FALSE.

!!$    !!!!!!!!!!!!!!!!!!!!!!!!!!!
!!$    ! data setting
!!$    CALL loaddata_from_conf()
!!$    IF (TRIM(restart_file) .NE. "") THEN
!!$       ! overload the setting from parameter file
!!$       CALL loaddata_from_restartfile()
!!$       CALL state_breakdown_all(CRU_states, irow_R, LCCcluster_states, RYRcluster_states)
!!$    END IF
!!$    
!!$    ! initialize gpu data
!!$    Ca_ds_dev = Ca_ds
!!$    Ca_jsr_dev = Ca_jsr
!!$ #if _LCC_MODEL_ == _SUN2000    
!!$    CaCM_ds_dev = CaCM_initial
!!$ #endif    
      
    compPdt_dev = +0d0

!!$    IF (TRIM(restart_file) .EQ. "") THEN
!!$       ! initialize release site's state
!!$       CALL init_CRU_states(lcc_initialstate, ryr_initialstate, N_L, mL, N_R, mR, &
!!$            CRU_states, LCCcluster_states, RYRcluster_states)
!!$    ENDIF
!!$        
!!$    !copy from host to device
!!$    CRU_states_dev = CRU_states
!!$    RYRgate_dev = U_Ro(RYRcluster_states) !IMPORTANT: now we need to deal with homogeneous cluster state index
!!$    DHPRgate_dev = U_Lo(LCCcluster_states)

    ! Pseudo-random-number-generator
    CALL init_PRNG()
    
    !! Define simulation parameters
    t_start = t_start + real_timestart ! NOTE: read-in value (conf file) for t_start is 0
    time = t_start
    t_end = t_end + real_timestart
!!$ #if _RUNMODE_ == _MODE_V_CLAMP || _RUNMODE_ == _MODE_V_CLAMP_TEST_LCC
!!$    !tsegment = t_start + t_V_stim_start
!!$    tsegment = t_start + t_I_stim_start
!!$ #elif _RUNMODE_ == _MODE_I_CLAMP
!!$    tsegment = t_start + t_I_stim_start
!!$ #endif
    tsegment = t_start
    !+ when to start spark analysis
    time_start_spark_analysis = MAX(t_start, t_end - duration_spark_analysis)
#if _RUNMODE_ == _MODE_V_CLAMP && _INCREMENT_VSTIM_ == _YES
! #if _RUNMODE_ == _MODE_I_CLAMP && _INCREMENT_VSTIM_ == _YES    
    !!!!!!!!!!!!!!!! TEST V-stim !!!!!!!!!!!!!!!!!!!!!!!
    tick_time = t_start + BCL
    time_start_spark_analysis = t_start
    fileID_gain_analysis = get_file_ID()
    !OPEN (unit=fileID_gain_analysis, file=TRIM(where2save) // "/Vstim_gain.txt", ACTION='write')
    OPEN (unit=fileID_gain_analysis, file=TRIM(where2save) // "/" // filestring_gain_analysis, ACTION='write')
    !!!!!!!!!!end    TEST V-stim !!!!!!!!!!!!!!!!!!!!!!!
 #endif
    
 #if _DATAMODE_ != _IO_NONE
    IF (duration_write_lastbeats .GT. t_end-t_start) THEN
       WRITE(error_message, *) "Duration to write lastbeats longer than simulation setting"
       res = ERR_FILE_INVALID_DATA
       RETURN
    ENDIF
 #endif
    time_prev1 =  t_end - duration_write_lastbeats - write_interval ! a time tracker for data I/O
    time_prev_fb = t_start
    time_prev2 = time_start_spark_analysis ! a time tracker to find integrate flux
    IF (write_firstbeat .NE. 0) THEN
       IF (t_start + 1.0d0 .GT. time_start_spark_analysis) THEN
          WRITE(log_string, *) "Duration do spark analysis overlap with first-beat analysis"
          CALL write2log(log_string)
          STOP
       ENDIF
       time_prev2 = t_start
    ENDIF
        BCL1=1.0    

!!$    tsegment = 1.d0/(Hz)
!!$    IF (Hz .GT. 1) THEN
!!$       t_V_stim_start = tsegment/2
!!$       t_V_stim_end = t_V_stim_start + t_V_stim_duration
!!$       t_I_stim_start = tsegment/2
!!$       t_I_stim_end = t_I_stim_start + t_I_stim_duration
!!$    END IF
    
!!$    ! Vm-dependent rate transitions
!!$#if _LCC_MODEL_ == _WILLIAMS2013
!!$    INCLUDE "Vm_dependent.inc"  !!kva,kvi
!!$#elif _LCC_MODEL_ == _SUN2000
!!$    INCLUDE "Update_LCC_Sun2000.inc"
!!$    lcc_Ktransitionrate_dev = lcc_Ktransitionrate
!!$    !CALL print_mat(lcc_Ktransitionrate, 1, mL, 1, mL, "Ktrans_V-00.txt")
!!$    !CALL print_mat(lcc_Ktransitionrate, 1, mL, 1, mL, "Ktrans_test.txt")
!!$    !CALL print_mat(lcc_Ktransitionrate, 1, mL, 1, mL, "Ktrans.txt")
!!$    CALL print_mat(lcc_Ktransitionrate, 1, mL, 1, mL)
!!$#elif _LCC_MODEL_ == _WILLIAMS2013_NEWALGORITHM
!!$    INCLUDE "Update_LCC_Williams2013.inc"
!!$    lcc_Ktransitionrate_dev = lcc_Ktransitionrate
!!$    !CALL print_mat(lcc_Ktransitionrate, 1, mL, 1, mL, "Ktrans_Williams.txt")
!!$    !STOP
!!$#endif
!!$    
!!$    INCLUDE "dp_args.inc"  !! dp_arg1, dp_arg2

    !! data used in reduction
    odata_d = initreduction_double()

    res = setup_KernelConfiguration()
    IF (res .NE. SUCCESS) RETURN
    
    ! copy data to constant memory
    CALL set_device_constant()




    !! this data is being used for caffeine OR trigger RYR simulation 
    CALL get_index_XXXcluster_numXXXopen(N_R, irow_R, NSFU, U_Ro, &
         index_RYRcluster_numRYRopen)

    !! this data is being used for trigger LCC simulation 
    CALL get_index_XXXcluster_numXXXopen(N_L, irow_L, NSFU, U_Lo, &
         index_LCCcluster_numLCCopen)
    
    !!! write a new configuration content  to the parameter file in the destination folder
    CALL write_and_copy_configurationfile()

    CALL reset_Data()
    ramp_time = ramp_time_reference * (V_stim-V_rest)/(+10-(V_rest))
    
    ! I/O init
    CALL data_io_init()
    !!!TUAN: check to make sure I/O for firstbeat working properly
    IF (write_firstbeat .NE. 0) THEN
       CALL data_io_write_fb()
    ENDIF
    IF ((duration_write_lastbeats .GT. 0.0d0) .AND. (time .GE. (t_end - duration_write_lastbeats))) THEN 
       CALL data_io_write()
    ENDIF

    WRITE (log_string, *) "++end init_Simulation()"
    CALL write2log(log_string)    
    PRINT *, log_string
  END FUNCTION init_Simulation

  !!>****f* ecc_gpu_utility/continue_Simulation
  !! SYNOPSIS  
  !!     Calculate the calcium concentration using GPU 
  !! INPUTS 
  !! OUTPUTS 
  !! RETURN 
  !! SIDE EFFECTS 
  !! HISTORY 
  !! BUGS 
  !! TODO 
  !! SEE ALSO 
  !!_****
  attributes(host) SUBROUTINE continue_Simulation()
    IMPLICIT NONE
    !+ other local
    INTEGER, PARAMETER :: MAX_ITER=600000000 ! maximum number of iteration steps
    INTEGER :: ii, nsteps, err
            
    WRITE (log_string, *) "start continue_Simulation()"
    CALL write2log(log_string)    
    PRINT *, log_string
    
    nsteps = copy_interval
    !! !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    !! TIMING LOOP
    OUTERLOOP: DO ii = 1, MAX_ITER
#if _BENCHMARK_ == _YES
       PRINT *, "ii", ii
       IF (ii .EQ. 3) THEN
          EXIT OUTERLOOP
       ENDIF
#endif

       !! run a number of time steps
       err = simulate_NextSteps(nsteps, is_EndSim)
#if _SAFE_MODE_ == _YES
       IF (err .NE. SUCCESS) THEN
          WRITE(log_string, *) "Error in simulate_NextSteps()"
          CALL write2log(log_string)
          EXIT OUTERLOOP
       END IF
#endif

       IF (is_EndSim) EXIT OUTERLOOP
    END DO OUTERLOOP

    WRITE (log_string, *) "++end continue_Simulation()"
    CALL write2log(log_string)    
    PRINT *, log_string
        
#if _BENCHMARK_ == _NO
    !!TUANHOANG
    CALL spark_summarize()
#endif
    
    !! writing the last time
    CALL data_io_finalize()

    RETURN
  END SUBROUTINE continue_Simulation

  SUBROUTINE finalize_Simulation
 #if _RUNMODE_ == _MODE_V_CLAMP && _INCREMENT_VSTIM_ == _YES    
    CLOSE(fileID_gain_analysis)
 #endif
    
    !!clean data on GPU
    CALL finalizereduction_double_tesla(odata_d)
  #if _USE_PRNG_ == _PRNG_SARU    
    CALL finalize_saru(saruObj)
  #endif
    CALL free_device()
  END SUBROUTINE finalize_Simulation

  SUBROUTINE reset_Data()
    IMPLICIT NONE
    INTEGER ::iserror
    LOGICAL :: plog
    INTEGER :: ii
    
    flag1 = .FALSE.
    flag2 = .FALSE.
    flag3 = .FALSE.
    flag4 = .FALSE.
    ts_spark = +0d0
    ts_blink = +0d0
    Nquarks_dev = 0
    Nsparks_dev = 0
    Nblinks_dev = 0
    tot_SparkDurations = 0d0
    tot_SparkPeaks = 0d0
    tot_BlinkDurations = 0d0
    tot_SparkTime2Peak = 0d0

    !! Initial setting values
    beta_myo = 0d0;  
    J_efflux_total(1) = 0d0; J_efflux_dev = 0.0d0
    J_leak = 0d0;
    J_refill_total(1) = 0d0;     J_refill_dev = 0.0d0
    J_Ca_sarmem = 0d0;   J_ryr_nj_T = 0d0;  J_serca = 0d0; J_CaTrpn = 0d0
    J_CaCM = 0.0d0; J_CaSL = 0.0d0; J_CaSR = 0.0d0; J_CaF = 0.0d0
    I_Na = 0d0;   I_dhpr_nj_T = 0d0
    I_dhpr_total(1) = 0d0;     I_dhpr_dev = 0d0; I_dhpr = 0d0
    J_ryr_orphan = 0d0
    J_ryr_total(1) = 0d0;     J_ryr_dev = +0d0; 
    J_ryr_orphan_total(1) = 0d0
    INTEGRAL_GAIN = 0d0
    PEAK_GAIN = 0d0
    peak_J_dhpr_total = 0d0
    peak_J_SRrelease_total = 0d0
    accumulate_J_dhpr = 0.0d0
    accumulate_J_ryr = 0.0d0; accumulate_J_ryr_orphan = 0.0d0
    accumulate_J_SL_in = 0.0d0; accumulate_J_SL_out = 0.0d0
    accumulate_J_SR_in = 0.0d0; accumulate_J_SR_out = 0.0d0
    APD = 0d0; APstart = 0d0; APend = 0d0;
    Vmthreshold = (/ V_rest * 0.9d0, V_rest * 0.5d0, V_rest * 0.2d0, +0.0d0 /)
    APcount = +0
    is_in_DiastolicPhase = .FALSE.
    is_done_firstbeat = .FALSE.
    I_Na = 0.0d0
    I_Ktof = 0.0d0
    I_Ktos = 0.0d0
    I_K1 = 0.0d0
    I_Kss = 0.0d0
    I_bNa = 0.0d0
    I_bK = 0.0d0
    I_bCa = 0.0d0
    I_NaK = 0.0d0
    I_pmca = 0.0d0
    I_dhpr_nj_T = 0.0d0
    
    !! some internal data
    is_being_clamped = .FALSE.
    is_inside_AP = .FALSE. 
    is_EndSim = .FALSE.

    CALL loaddata_from_conf()
    IF (TRIM(restart_file) .NE. "") THEN
       ! overload the setting from parameter file
       CALL loaddata_from_restartfile()
       CALL state_breakdown_all(CRU_states, irow_R, LCCcluster_states, RYRcluster_states)
    END IF

    ! initialize gpu data
    Ca_ds_dev = Ca_ds
    Ca_jsr_dev = Ca_jsr
 #if _LCC_MODEL_ == _SUN2000    
    CaCM_ds_dev = CaCM_initial
 #endif    

    IF (TRIM(restart_file) .EQ. "") THEN
       ! initialize release site's state
       CALL init_CRU_states(lcc_initialstate, ryr_initialstate, N_L, mL, N_R, mR, &
            CRU_states, LCCcluster_states, RYRcluster_states)
    ENDIF
        
    !copy from host to device
    CRU_states_dev = CRU_states
    RYRgate_dev = U_Ro(RYRcluster_states) !IMPORTANT: now we need to deal with homogeneous cluster state index
    DHPRgate_dev = U_Lo(LCCcluster_states)

!!$    Vm = -40
!!$    DO ii=1,20
!!$       INCLUDE "Update_LCC_Sun2000.inc"
!!$       WRITE(*, "(10(1pe12.5,2x))") Vm, Po2*lcc_Krateconstant(1,2), tau_Po2, att
!!$       !WRITE(*, "(10(1pe12.5,2x))") Vm, lcc_Ktransitionrate(1,2), lcc_Ktransitionrate(2,1), lcc_Ktransitionrate(1,2)/(lcc_Ktransitionrate(1,2)+lcc_Ktransitionrate(2,1))
!!$       Vm = Vm+5
!!$    ENDDO
!!$    STOP
       
       
    ! Vm-dependent rate transitions
#if _LCC_MODEL_ == _WILLIAMS2013
    INCLUDE "Vm_dependent.inc"  !!kva,kvi
#elif _LCC_MODEL_ == _SUN2000
    INCLUDE "Update_LCC_Sun2000.inc"
    lcc_Ktransitionrate_dev = lcc_Ktransitionrate
    !CALL print_mat(lcc_Ktransitionrate, 1, mL, 1, mL, "Ktrans_V-00.txt")
    !CALL print_mat(lcc_Ktransitionrate, 1, mL, 1, mL, "Ktrans_test.txt")
    !CALL print_mat(lcc_Ktransitionrate, 1, mL, 1, mL, "Ktrans.txt")
!    CALL print_mat(lcc_Ktransitionrate, 1, mL, 1, mL)
#elif _LCC_MODEL_ == _WILLIAMS2013_NEWALGORITHM
    INCLUDE "Update_LCC_Williams2013.inc"
    lcc_Ktransitionrate_dev = lcc_Ktransitionrate
    !CALL print_mat(lcc_Ktransitionrate, 1, mL, 1, mL, "Ktrans_Williams.txt")
    !STOP
#endif
    
    INCLUDE "dp_args.inc"  !! dp_arg1, dp_arg2
    
  END SUBROUTINE reset_Data

  
  !!>****f* ecc_gpu_utility/simulate_NextSteps
  !! SYNOPSIS  
  !!     Perform a number of time steps (i.e. "nsteps" steps)
  !!     NOTE: make sure nsteps .LE.  copy_interval
  !!   
  !! INPUTS
  !!   nsteps
  !! OUTPUTS 
  !!   is_EndSim : .TRUE. if the simulation is complete
  !! RETURN 
  !! SIDE EFFECTS 
  !! HISTORY 
  !! BUGS 
  !!_****
  INTEGER FUNCTION simulate_NextSteps(nsteps, is_EndSim) RESULT(res)
    IMPLICIT NONE
    LOGICAL, INTENT(INOUT) :: is_EndSim
    INTEGER, INTENT(IN):: nsteps
    !!local 
    INTEGER :: iinner, cuErr
    REAL(KIND=dp), PARAMETER :: tau_time = 1.0d-3
    REAL(KIND=dp) :: tresidue, invdt, min_Pdt, dtmax
    INTEGER :: ii
    REAL(KIND=dp) :: dummy1, dummy2
    !+ parameters for gating variables
    REAL(KIND=dp) :: TCa_i2, TCa_sr2, vcycle_pump
    REAL(KIND=dp), DIMENSION(irow_L_nj) :: tmp_pL_nj
    REAL(KIND=dp), DIMENSION(irow_R_nj) :: tmp_pR_nj


    res = SUCCESS

    IF (is_EndSim) RETURN

#if _SAFE_MODE_ == _YES
    IF (nsteps .GT. copy_interval) THEN
       PRINT *, "ERROR: simulate_NextSteps\n nsteps is larger than copy_interval"
       err = ERR_INVALID_DATA_SETTING
       RETURN
    END IF
#endif
       
#if _USE_PRNG_ == _PRNG_SARU
    !!SARU pseudo-RNG
    CALL gen_rand_saru_c2050(saruObj, X_r_dev, NSFU, nsteps)
#elif _USE_PRNG_ == _PRNG_MT
    CALL generate_mt_data(d_status, d_texture, NSFU*nsteps, X_r_dev)
#elif _USE_PRNG_ ==  _PRNG_MTCPU
    CALL DRANDUNIFORM(NSFU*nsteps,mt_LB,mt_UB,mt_STATE,X_r,INFO)
    X_r_dev = X_r
#endif

 #if _RUNMODE_ == _MODE_V_CLAMP && _INCREMENT_VSTIM_ == _YES    
    !!!!!!!!!!!!!!!! TEST V-stim !!!!!!!!!!!!!!!!!!!!!!!
    IF (time .GE. tick_time) THEN
       WRITE (fileID_gain_analysis, *) "================================================"
       WRITE (fileID_gain_analysis, *) "Vstim = ", V_stim
       CALL spark_summarize(fileID_gain_analysis, BCL)
       CALL reset_Data()
       tick_time = tick_time + BCL
      V_stim = V_stim + 10
  !     V_stim = V_stim + 5
       IF (ABS(V_stim) .LT. EPS)  V_stim = 0.0001d0
       !!adjust to that the slope of ramp_time is the same for all Vstim, given that
       !! the slope from -80mV to +10mV is ramp_time_reference
       !ramp_time = ramp_time_reference * (V_stim-V_rest)/(+10-(-80))
       PRINT *, "Vstim = ", V_stim
    ENDIF
    !!!!!!!!!!END    TEST V-stim !!!!!!!!!!!!!!!!!!!!!!!
 #endif

    
    INNERLOOP: DO iinner = 1, nsteps  ! inner loop
       ! CHECK FOR COMPLETE SIMULATION
       IF (time .GT. t_end) THEN
          is_EndSim = .TRUE.
          EXIT INNERLOOP
       ENDIF
#if _BENCHMARK_ == _YES
       !PRINT *, "iinner", iinner
#endif
       
#if (_RUNMODE_ == _MODE_I_CLAMP || _RUNMODE_ == _MODE_V_CLAMP || _RUNMODE_ == _MODE_V_CLAMP_TEST_LCC)
       tresidue = MODULO(time, tsegment)
IF (time .GE. tsegment+BCL) then
       tsegment = tsegment+BCL
      END IF   


!if((time.GE.0.0).and.(time.LE.1.0))then
!elseif((time.GE.tsegment+BCL)) then
!tsegment = tsegment+BCL
!elseif (time.GE.tsegment+BCL1)then
!tsegment=tsegment+BCL1
!END IF

!if((time.LE.5.0).and.(time.ge.tsegment+BCL))then
!tsegment = tsegment+BCL
!elseif ((time.ge.tsegment+BCL1))then
!tsegment=tsegment+BCL1
!endif

    IF (time .GE. tsegment) tresidue = time - tsegment
 #if _RUNMODE_ == _MODE_V_CLAMP || _RUNMODE_ == _MODE_V_CLAMP_TEST_LCC
       !V-clamp
     #if _LCC_MODEL_ == _WILLIAMS2013
       !INCLUDE "Vclamp_method1.inc"
       !INCLUDE "Vclamp_method2.inc"
       INCLUDE "Vclamp_method3.inc"
       !INCLUDE "Vclamp_method4.inc"
       !INCLUDE "Vclamp_method5.inc"
     #elif _LCC_MODEL_ == _SUN2000
       INCLUDE "Vclamp_method3_LCC_Sun2000.inc"
     #elif _LCC_MODEL_ == _WILLIAMS2013_NEWALGORITHM
       INCLUDE "Vclamp_method3_LCC_Williams2013.inc"
     #endif
       
 #elif _RUNMODE_ == _MODE_I_CLAMP
       !!I-clamp
       IF (tresidue.GT. t_I_stim_start .AND. tresidue.LT. t_I_stim_end) THEN
          IF (.NOT. is_being_clamped) THEN
             !!current-clamp
             I_app = I_stim
             is_being_clamped = .TRUE.
          END IF
!    ELSEIF (tresidue.GT. (t_I_stim_start+0.45) .AND. tresidue.LT. t_I_stim_end+0.45) THEN
!        IF (.NOT. is_being_clamped) THEN
!            !!current-clamp
!             I_app = I_stim
!             is_being_clamped = .TRUE.
!          END IF
    
   
      ELSEIF (is_being_clamped .EQ. .TRUE.) THEN
          I_app = +0d0             
          is_being_clamped = .FALSE.
       END IF
   #if _LCC_MODEL_ == _WILLIAMS2013
       INCLUDE "Vm_dependent.inc"  !!kva,kvi
   #elif _LCC_MODEL_ == _SUN2000
       INCLUDE "Update_LCC_Sun2000.inc"
       cuErr = cudaMemcpy2D(lcc_Ktransitionrate_dev,mL, lcc_Ktransitionrate, mL, mL, mL)
   #elif _LCC_MODEL_ == _WILLIAMS2013_NEWALGORITHM
       INCLUDE "Update_LCC_Williams2013.inc"
       cuErr = cudaMemcpy2D(lcc_Ktransitionrate_dev,mL, lcc_Ktransitionrate, mL, mL, mL)
   #endif    
       INCLUDE "dp_args.inc"  !! dp_arg1, dp_arg2
 #endif

#endif
       
       !!find rate transition
#if _TEST_NEW_ORPHAN_CRU_ == _YES 
       CALL get_compPdt_2b <<<gridSize, blocksize>>> (kva, kvi, NSFU_normal, Ej_orphan, eta_CO)
#else
   #if _LCC_MODEL_ == _WILLIAMS2013       
       CALL get_compPdt_2 <<<gridSize, blocksize>>> (kva, kvi)
   #elif _LCC_MODEL_ == _SUN2000
       CALL get_compPdt_Sun2000 <<<gridSize, blocksize>>> ()
   #elif _LCC_MODEL_ == _WILLIAMS2013_NEWALGORITHM
       CALL get_compPdt_Williams2013 <<< gridSize, blocksize>>>()
   #endif
#endif

       cuErr = cudaDeviceSynchronize()

#if _SAFE_MODE_ == _YES
       IF (cuErr .NE. cudaSuccess) THEN
          WRITE(log_string,*) "Kernel Error: get_compPdt_2"
          CALL write2log(log_string)
          WRITE(error_message,*) cudaGetErrorString(cudaGetLastError())
          res = CUDA_ERROR
          RETURN
       END IF
#endif
       
       !TUAN: use min on GPU
       min_Pdt = min_reduce_double_2host(compPdt_dev(1:NSFU), NSFU, odata_d)
       
#if _DEBUG_ == _YES
       compPdt = compPdt_dev
       IF (min_Pdt .NE. MINVAL(compPdt)) THEN
          WRITE(log_string,*) "Error: min_reduce_double_2host not working"
          CALL write2log(log_string)
          STOP
       END IF
#endif
       
       dtmax =-0.1d0/min_Pdt
       dt = MAX(MIN(dtmax,dt_UB),dt_LB)
       
!!$#if _RUNMODE_ == _MODE_V_CLAMP
!!$       !!!try to see if we can reduce tail current
!!$       IF (tresidue .GT. t_V_stim_end .AND. tresidue .LE. t_V_stim_end + ramp_time) THEN
!!$          dt = 1d-8
!!$       ENDIF
!!$#endif
       oldtime = time
       time = time + dt
       invdt = 1.0d0/dt

#if _EMULATE_CONDITION_ == _CAFFEINE_TYPE1
       !put caffeine after 'time_start_caffeinedump' (sec) of quiescent simulation
       !for the duration of 'duration_caffeinedump' (sec)
       IF (time .GE. time_start_caffeinedump .AND. &
            time .LE. (time_start_caffeinedump + duration_caffeinedump)) THEN
          !INCLUDE "caffeinedump_v1.inc"
          !INCLUDE "caffeinedump_v2.inc"
          !INCLUDE "caffeinedump_v3.inc"
          !INCLUDE "caffeinedump_v4.inc"
          !INCLUDE "caffeinedump_v5.inc"
          IF (is_caffeinedump_triggered .EQ. .FALSE.) THEN
             dt = 5.0d-8 ! set small dt to avoid numerical instability
             time = oldtime + dt
             invdt = 1.0d0/dt
             CRU_states = CRU_states_dev
             CALL set_wholecell_state_caffeine_5(U_Ro, CRUs_blocked, CRU_states(1:NSFU),  NSFU, irow_R,  &
                  LCCcluster_states, RYRcluster_states, some_caffeineBound_CRU_notfullyopen)
             
             RYRgate_dev(1:NSFU) = U_Ro(RYRcluster_states(1:NSFU))
             CRU_states_dev(1:NSFU) = CRU_states(1:NSFU)
             DHPRgate_dev(1:NSFU) = U_Lo(LCCcluster_states(1:NSFU))
	     
             is_caffeinedump_triggered = .NOT. some_caffeineBound_CRU_notfullyopen
          ENDIF
          
          CALL update_SFU_caffeine <<<gridSize, blocksize>>> &
               (dt, Vm, dp_arg1, dp_arg2, &
               Ca_myo, Ca_nsr, &
               )
       ELSE !endif at the next #if _EMULATE_CONDITION_ == _CAFFEINE_TYPE1
#elif _EMULATE_CONDITION_ == _TRIGGER_RYR_CRU
       !trigger CRU
       IF (time .GE. (num_triggertime * interval_trigger_RYR + time_start_trigger_RYR) .AND. &
            is_all_CRU_open .EQ. .FALSE.) THEN
          
          dt = 5.0d-8 ! set small dt to avoid numerical instability
          time = oldtime + dt
          invdt = 1.0d0/dt

          CRU_states = CRU_states_dev
          CALL trigger_CRU(U_Ro, CRU_states(1:NSFU), NSFU, irow_R, &
               LCCcluster_states, RYRcluster_states, is_all_CRU_open)

          RYRgate_dev(1:NSFU) = U_Ro(RYRcluster_states(1:NSFU))
          CRU_states_dev(1:NSFU) = CRU_states(1:NSFU) 
          DHPRgate_dev(1:NSFU) = U_Lo(LCCcluster_states(1:NSFU))

          IF (is_all_CRU_open .EQ. .TRUE. ) THEN
             !PRINT *, "is_all_CRU_open = ", is_all_CRU_open
             num_triggertime = num_triggertime + 1
             is_all_CRU_open = .FALSE.
          ENDIF
       #if _LCC_MODEL_ == _WILLIAMS2013
          CALL update_SFU_nostatechanges <<<gridSize, blocksize, 0, stream0>>> &
               (X_r_dev((iinner-1)*NSFU+1:(iinner)*NSFU), &
               dt, invdt, Vm, kva, kvi, dp_arg1, dp_arg2, &
               Ca_myo, Ca_nsr)
       #elif _LCC_MODEL_ == _SUN2000
          !!call something here
          !! update_SFU_Sun2000_nostatechanges()
          STOP
       #elif _LCC_MODEL_ == _MAHAJAN2008
          !!call something here
          STOP
       #elif _LCC_MODEL_ == _MOROTTI2012
          !!call something here
          STOP
       #endif

       ELSE !endif at the next #if _EMULATE_CONDITION_ == _TRIGGER_RYR_CRU
#elif _EMULATE_CONDITION_ == _TRIGGER_LCC
       !trigger CRU by opening 1 LCC (check trigger_CRU_openLCC())  for a period of duration_open_LCC = 0.5ms 
       IF (time .GE. (num_triggertime * interval_trigger_LCC + time_start_trigger_LCC) .AND. &
            is_all_CRU_open .EQ. .FALSE.) THEN

          dt = 5.0d-8 ! set small dt to avoid numerical instability
          time = oldtime + dt
          invdt = 1.0d0/dt

          IF (time .GT. num_triggertime * interval_trigger_LCC + time_start_trigger_LCC + duration_open_LCC) THEN
             is_all_CRU_open=.TRUE.
          ENDIF

          IF (.NOT. is_opened_LCC .OR. is_all_CRU_open) THEN
             CRU_states = CRU_states_dev
             CALL trigger_CRU_openLCC(U_Lo, CRU_states(1:NSFU), NSFU, irow_R, &
                  LCCcluster_states, RYRcluster_states, is_all_CRU_open)
             RYRgate_dev(1:NSFU) = U_Ro(RYRcluster_states(1:NSFU))
             CRU_states_dev(1:NSFU) = CRU_states(1:NSFU) 
             DHPRgate_dev(1:NSFU) = U_Lo(LCCcluster_states(1:NSFU))
             is_opened_LCC = .TRUE.
          END IF
          
          IF (is_all_CRU_open .EQ. .TRUE. ) THEN
             !PRINT *, "is_all_CRU_open = ", is_all_CRU_open
             num_triggertime = num_triggertime + 1
             is_all_CRU_open = .FALSE.
             is_opened_LCC = .FALSE.
          ENDIF
        #if _LCC_MODEL_ == _WILLIAMS2013
          CALL update_SFU_nostatechanges <<<gridSize, blocksize>>> &
               (X_r_dev((iinner-1)*NSFU+1:iinner*NSFU), &
               dt, invdt, Vm, kva, kvi, dp_arg1, dp_arg2, Ca_myo, Ca_nsr)
        #elif _LCC_MODEL_ == _SUN2000
          !!call something here
          !! update_SFU_Sun2000_nostatechanges()
          STOP
        #elif _LCC_MODEL_ == _MAHAJAN2008
          !!call something here
          STOP
        #elif _LCC_MODEL_ == _MOROTTI2012
          !!call something here
          STOP
        #endif
          
       ELSE !endif at the next #if _EMULATE_CONDITION_ == _TRIGGER_LCC
#endif

       !!TUANHOANG
#if _TEST_NEW_ORPHAN_CRU_ == _YES 
          CALL update_SFU_2b <<<gridSize, blocksize>>> &
               (X_r_dev((iinner-1)*NSFU+1:iinner*NSFU), &
               dt, invdt, Vm, kva, kvi, dp_arg1, dp_arg2, Ca_myo, Ca_nsr, &
               NSFU_normal, inv_lambda_ds_orphan, Km_myo, B_myoT, Ej_orphan, eta_CO)
#else
   #if _LCC_MODEL_ == _WILLIAMS2013       
       CALL update_SFU_2 <<<gridSize, blocksize>>> &
            (X_r_dev((iinner-1)*NSFU+1:iinner*NSFU), &
            dt, invdt, Vm, kva, kvi, dp_arg1, dp_arg2, Ca_myo, Ca_nsr)
   #elif _LCC_MODEL_ == _SUN2000
       !PRINT *, "calling updateSytem_Sun2000"
       CALL update_SFU_Sun2000 <<<gridSize, blocksize>>> &
            (X_r_dev((iinner-1)*NSFU+1:iinner*NSFU), &
            dt, Vm, dp_arg1, dp_arg2, Ca_myo, Ca_nsr,new)
   #elif _LCC_MODEL_ == _WILLIAMS2013_NEWALGORITHM
       CALL update_SFU_Williams2013 <<<gridSize, blocksize>>> &
            (X_r_dev((iinner-1)*NSFU+1:iinner*NSFU), &
            dt, Vm, dp_arg1, dp_arg2, Ca_myo, Ca_nsr)
   #endif
#endif
          
#if _EMULATE_CONDITION_ == _CAFFEINE_TYPE1 || _EMULATE_CONDITION_ == _TRIGGER_RYR_CRU ||  _EMULATE_CONDITION_ == _TRIGGER_LCC
       END IF
#endif

#if _TEST_ORPHAN_CRU_ == _YES
       J_ryr_orphan = SUM(pR_orphan*U_Ro_orphan, MASK=(U_Ro_orphan .GE. 1)) * v_ryr * (Ca_nsr - Ca_myo)
       chiC_orphan = EXP(-Ej*0.5d0*(dfloat(N_R-U_Ro_orphan)*Ecc - dfloat(U_Ro_orphan-1)*Eoo))
       chiO_orphan = EXP(-Ej*0.5d0*(dfloat(U_Ro_orphan)*Eoo - dfloat(N_R-U_Ro_orphan-1)*Ecc))
       tmp_pR_orphan = pR_orphan
       DO ii = 1, irow_R_orphan
          pR_orphan(ii) = tmp_pR_orphan(ii) + dt * SUM(tmp_pR_orphan(:)* (chiC_orphan(ii) * aKm2_orphan(:,ii) + chiO_orphan(ii) * (k_jsr0*Ca_nsr+k_jsr1)*Ca_myo**eta_RYR * aKp2_orphan(:,ii)))
       END DO
#endif

#if _RUNMODE_ != _MODE_CLOSEDCELL && _RUNMODE_ != _MODE_V_CLAMP_TEST_LCC
       !!Resting potential
       E_Na = RT_F*LOG(Na_o/Na_i)
       E_K = RT_F*LOG(K_o/K_i)
       E_Ca = RT_FzCa*LOG(Ca_o/Ca_myo)
       
       !! Fast Na current (Luo-Rudy)
       IF (Vm .GE. -40.d0) THEN
          a_h = 1.d3*0
          a_j = 1.d3*0
          b_h = 1.d3*1/(0.13d0*(1+EXP((Vm+10.66d0)/-11.1d0)))
          b_j = 1.d3*0.3d0*(EXP(-2.535d-7*Vm)/(1.d0+EXP(-0.1d0*(Vm+32.d0))))
       ELSEIF (Vm .LT. -40.d0) THEN 
          a_h = 1.d3*(0.135d0*EXP((80d0+Vm)/-6.8d0))
          a_j = 1.d3*(-1.2714d5*EXP(0.2444d0*Vm)-3.474d-5*EXP(-0.04391d0*Vm))* &
               ( (Vm+37.78d0)/(1.d0+EXP(0.311d0*(Vm+79.23d0))))
          b_h = 1.d3*(3.56d0*EXP(0.079d0*Vm)+3.1d5*EXP(0.35d0*Vm))
          b_j = 1.d3*(0.1212d0*EXP(-0.01052d0*Vm)/(1.d0+EXP(-0.1378d0*(Vm+40.14d0))))
       ELSE ! NaN Blank
          WRITE(log_string,*) "Error: NaN value Vm"
          CALL write2log(log_string)
          WRITE(log_string, *) time, V_rest, Vm
          CALL write2log(log_string)
          STOP
          RETURN
       END IF
       a_m = 1.d3*0.32d0*(Vm+47.13d0)/(1.d0-EXP(-0.1d0*(Vm+47.13d0)))
       b_m = 1.d3*0.08d0*EXP(-Vm/11.d0)  ![1/sec]

       I_Na = g_Na * m_Na**3 * h_Na * j_Na * (Vm - E_Na)  ![uA/cm^2]

       !! K currents [uA/cm^2]
       !!wrong formula
       !!I_Ktof = g_Ktof* a_Ktof**3 *in_Ktof*(Vm-E_K) 
       !!I_Ktos = g_Ktos* a_Ktos**3 *in_Ktos*(Vm-E_K) 
       !!I_K1   = g_K1*(K_o/(K_o+210.0d0))*((Vm-E_K-1.73d0)/(1.0d0+EXP(0.0596d0*(Vm-E_K-1.73d0)))) 
       I_Ktof = g_Ktof* a_Ktof *in_Ktof*(Vm-E_K) 
       I_Ktos = g_Ktos* a_Ktos *in_Ktos*(Vm-E_K) 
       I_K1   = g_K1*(K_o/(K_o+210.0d0))*((Vm-E_K-1.73d0)/(1.0d0+EXP(0.0896d0*(Vm-E_K-1.73d0)))) 
          
       I_Kss  = g_Kss*a_Kss*in_Kss*(Vm-E_K)
        !!!!!!!!!!!!!!!!!!!!!!
          !!! here I introduce a new equation for the increased open probability of
          !!!!! RYR mutation

        ! new =2*0.25*(X_00-new)   
        !  print*,new 
!!!!!!!!!!!!!!!!!!!!!!!
       !! Background currents [uA/cm^2]
       I_bNa = g_bNa*(Vm-E_Na)
       I_bK  = g_bK*(Vm-E_K)
       I_bCa = g_bCa*(Vm-E_Ca)

       !! Sarcolemma pumps and Exchangers
       I_NaK  = I_NaK_bar*(1.d0/(1.d0+0.1245d0*EXP(-0.1d0*(Vm*F_RT))+0.0365d0*(1/7d0*(EXP(Na_o/67300)-1.d0))* &
            EXP(-(Vm*F_RT))))*(1.d0/(1.d0+(21000d0/Na_i)**1.5d0))*(K_o/(K_o+1500d0))
       I_pmca = I_pmca_bar*(Ca_myo**eta_pmca/(Km_pmca**eta_pmca+Ca_myo**eta_pmca))
       I_ncx  = I_ncx_bar*(Na_i**3 * Ca_o*EXP(eta_ncx*Vm*F_RT) - Na_o**3 *Ca_myo*EXP((eta_ncx-1.d0)*Vm*F_RT))/ &
            ((Km_ncxNa**3 + Na_o**3)*(Km_ncxCa+Ca_o)*(1.d0+ksat_ncx*EXP((eta_ncx-1.d0)*Vm*F_RT)))
!print*,time,I_ncx
       !! Euler's==========================================
!       new=new+dt*(X_00-new)
!    print*,new
       m_na = m_na + dt*(a_m*(1-m_na)-b_m*m_na)
       h_na = h_na + dt*(a_h*(1-h_na)-b_h*h_na)
       j_na = j_na + dt*(a_j*(1-j_na)-b_j*j_na)
       a_Ktof = a_Ktof + dt*(1d3*((0.18064d0*EXP(0.03577d0*(Vm+30d0)))*(1-a_Ktof)-(0.3956d0*EXP(-0.06237d0*(Vm+30d0)))*a_Ktof))
       
       in_Ktof = in_Ktof + dt*(1d3*((((0.000152d0*EXP(-(Vm+13.5d0)/7d0))/(0.067083d0*EXP(-(Vm+33.5d0)/7d0)+1.d0)))* &
            (1.d0-in_Ktof)-(((0.00095d0*EXP((Vm+33.5d0)/7d0))/(0.051335d0*EXP((Vm+33.5d0)/7d0)+1.d0)))*in_Ktof))
       
       a_Ktos = a_Ktos + dt*(1d3*((1.d0/(1.d0+EXP(-(Vm+22.5d0)/7.7d0)))-a_Ktos)/(0.493d0*EXP(-0.0629d0*Vm)+2.058d0))
       in_Ktos = in_Ktos + dt*(1d3*((1.d0/(1.d0+EXP( (Vm+45.2d0)/5.7d0)))-in_Ktos)/((270.0d0+1050d0/(1.d0+EXP( (Vm+45.2d0)/5.7d0 )))))
       a_Kss  = a_Kss  + dt*(1d3*((1.d0/(1.d0+EXP(-(Vm+22.5d0)/7.7d0)))-a_Kss)/(0.393d0*EXP(-0.0862d0*Vm)+0.13d0))
       in_Kss  = in_Kss  + dt*0
       
   #if _LCC_MODEL_ == _WILLIAMS2013
       I_dhpr_nj_T = SUM(pL_nj, mask=U_Lo_nj .EQ. 1) * P_dhpr_nj *(F2zCa2_RT*Vm) * &
            (Ca_myo * EXP(Vm*FzCa_RT) - 0.341d0 * Ca_o)/(EXP(Vm*FzCa_RT)-1.d0) ! [uA]
   #else
       !TUAN
       I_dhpr_nj_T = 0.0d0
   #endif

       !! Updating Fluxes
       !+ membrane fluxes inward Ca2+ fluxes
       !J_bCa = -I_bCa*Am_2FVmyo
       !J_pmca = I_pmca*Am_2FVmyo
       !J_ncx = -2.0d0 * I_ncx * Am_2FVmyo
       !J_dhpr_nj_T = -I_dhpr_nj_T * inv_2FVmyo
       !J_Ca_sarmem = J_bCa - J_pmca - J_ncx + J_dhpr_nj_T
       !!TUANHOANG
       J_Ca_sarmem = (-I_bCa - I_pmca + zCa * I_ncx) * Am_2FVmyo  - I_dhpr_nj_T * inv_2FVmyo  ! [uM/s]       
#endif
          
       
          
       !+ internal fluxes
       J_leak = v_leak *(Ca_nsr - Ca_myo)  ![uM/s]

  #if _RUNMODE_ == _MODE_V_CLAMP_TEST_LCC
       J_serca = 0.0d0
       J_ryr_nj_T = +0d0
  #else
       TCa_i2 = (Ca_myo*invKp_myo)**2
       TCa_sr2 = (Ca_nsr*invKp_nsr)**2
       dummy1 = (3.24873d12 * TCa_i2**2 + TCa_i2 * (9.17846d6 - 11478.2d0 * TCa_sr2) - 0.329904d0 * TCa_sr2)
       dummy2 = (0.104217d0 + 17.923d0 * TCa_sr2 + TCa_i2 *(1.75583d6 + 7.61673d6 * TCa_sr2) &
            + TCa_i2**2 *(6.08463d11 + 4.50544d11 * TCa_sr2))  ! D_cycle
       vcycle_pump = dummy1/dummy2  ![1/sec]
       J_serca = cpl_serca*vcycle_pump*Ap  ![uM/sec]

       !!TUANHOANG
       J_ryr_nj_T = SUM(pR_nj, MASK=(U_Ro_nj .EQ. 1 )) * v_ryr_nj_T * (Ca_nsr - Ca_myo) ![uM/s]
  #endif
       
       !flux forming compound CaB_htrpn in myoplasm
       J_CaTrpn = kon_htrpn*Ca_myo*(B_htrpnT - CaB_htrpn) - koff_htrpn*CaB_htrpn  ![uM/s]
       J_CaCM = kon_CM*Ca_myo*(CM_T - CaCM) - koff_CM * CaCM
       J_CaSL = kon_SL*Ca_myo*(B_SL_T - CaSL) - koff_SL * CaSL
       J_CaSR = kon_SR*Ca_myo*(B_SR_T - CaSR) - koff_SR * CaSR
  #if _USE_FLUO_ == _YES
       J_CaF  = kon_F*Ca_myo*(F_T - CaF) - koff_F * CaF
  #endif

       !TUAN: 
       !Na_i = Na_i + dt*(-(I_Na + I_bNa + 3*I_ncx + 3*I_NaK)*Am_FVmyo)  ![uM]
       !K_i = K_i + dt*(-(I_K1 + I_Kss + I_Ktof + I_Ktos + I_bK - 2*I_NaK)*Am_FVmyo)  ![uM]
!print*, ryr_compKfree(50,0), ryr_compKfree(50,1),ryr_compKfree(50,1)

!print*, ryr_compKcal_act(50,0),ryr_compKcal_act(50,1),ryr_compKcal_act(50,2) 
       I_dhpr_total(1) = sum_reduce_double_2host(I_dhpr_dev, NSFU, odata_d)
       
#if _RUNMODE_ == _MODE_I_CLAMP
       !I-clamp (dVm/dt)
       Vm = Vm + dt*(-1.0d3/Csc * (I_Na + (I_dhpr_total(1) + I_dhpr_nj_T)/Am + &
            I_K1 + I_Kss + I_Ktof + I_Ktos + I_ncx  + I_NaK + I_pmca + &
            I_bNa + I_bCa + I_bk + I_app))
       
#endif
       !! synchronize GPU data
       cuErr = cudaDeviceSynchronize()
          
#if _SAFE_MODE_ == _YES
       IF (cuErr .NE. cudaSuccess) THEN
          WRITE(log_string,*) "Error: kernel"
          CALL write2log(log_string)
          WRITE(log_string,*) cudaGetErrorString(cudaGetLastError())
          CALL write2log(log_string)
          STOP
       END IF
#endif

       J_refill_total(1) = sum_reduce_double_2host(J_refill_dev, NSFU, odata_d)
       
       Ca_nsr = Ca_nsr + dt*((beta_nsr*inv_lambda_nsr)  * (J_serca - J_leak - J_refill_total(1) - J_ryr_nj_T - J_ryr_orphan))
          
       !!beta_myo = 1.0d0 /(1.0d0 + (B_myoT * Km_myo)/((Km_myo + Ca_myo)**2))
       dummy1 = (Km_myo + Ca_myo)**2
       dummy2 = dummy1 +  B_myoT * Km_myo
       beta_myo = dummy1 / dummy2
       
       J_efflux_total(1) = sum_reduce_double_2host(J_efflux_dev, NSFU, odata_d)
       
       Ca_myo = Ca_myo + dt*(beta_myo * (J_efflux_total(1) + J_leak + J_Ca_sarmem + J_ryr_nj_T - J_serca  - J_CaTrpn - J_CaCM - J_CaSL - J_CaSR - J_CaF + J_ryr_orphan))

       !!TUANHOANG
       !PRINT *, J_leak, J_ryr_nj_T, time
       
   #if _LCC_MODEL_ == _WILLIAMS2013
       tmp_pL_nj = pL_nj
       DO ii=1, irow_L_nj
          pL_nj(ii) = tmp_pL_nj(ii) + dt*SUM(tmp_pL_nj(:)*(kva*aKv1_nj(:,ii) + kvi*aKv2_nj(:,ii) + aKm1_nj(:,ii) + &
               Ca_myo**eta_LCC * aKp1_nj(:,ii)))
       END DO
       tmp_pR_nj = pR_nj
       DO ii=1, irow_R_nj
          pR_nj(ii) = tmp_pR_nj(ii) + dt*SUM(tmp_pR_nj(:)*(aKm2_nj(:,ii) + (k_jsr0*Ca_nsr+k_jsr1)*Ca_myo**eta_RYR * aKp2_nj(:,ii)))
       ENDDO
   #elif _LCC_MODEL_ == _SUN2000
       !PRINT *, "calculating non-j DHPR"
   #endif

       !CaF  = CaF  + dt*(-koff_F*CaF + kon_F*Ca_myo*(F_T - CaF));
       CaF  = CaF  + dt*J_CaF
       CaB_htrpn = CaB_htrpn + dt*J_CaTrpn
       CaCM = CaCM + dt*J_CaCM
       CaSL = CaSL + dt*J_CaSL
       CaSR = CaSR + dt*J_CaSR


   #if _HOW_OFTEN_DATA_IO_ == _METHOD1
       !!WRITE_METHOD_1: WRITE DATA OUT after a certain time interval (read from parameter)
       !! Can be switched  to using WRITE_METHOD_2
       IF (time .GE. (t_end - duration_write_lastbeats)) THEN 
          IF ((time - time_prev1) .GE. write_interval) THEN 
             CALL data_io_write()
             time_prev1 = time
          END IF
       ENDIF
       !WRITE DATA OUT (First-Beat) every 100 time-steps
       IF (write_firstbeat .NE. 0) THEN
          IF (time .LE. (t_start+1.0d0)) THEN
             !! write out the first beat 
             IF ((time - time_prev_fb) .GE. write_interval) THEN 
                CALL data_io_write_fb()
                time_prev_fb = time
             END IF
          ELSEIF (is_done_firstbeat .EQ. .FALSE.) THEN
             CALL data_io_write_fb_last()
             CALL spark_summarize_fb()
             is_done_firstbeat = .TRUE.
          ENDIF
       ENDIF
   #endif
    END DO INNERLOOP
   #if _HOW_OFTEN_DATA_IO_ == _METHOD2
       !!WRITE_METHOD_2: WRITE DATA OUT (every 100 time steps) 
       !!comment this section and switch to WRITE_METHOF_1 if we want
       !!!TUAN: switch back if want to output every 100 time steps
    IF (write_firstbeat .NE. 0) THEN
       IF (time .LE. (t_start+1.0d0)) THEN
          !! write out the first beat 
          CALL data_io_write_fb()
       ELSEIF (is_done_firstbeat .EQ. .FALSE.) THEN
          CALL data_io_write_fb_last()
          CALL spark_summarize_fb()
          is_done_firstbeat = .TRUE.
          time_prev2 = time_start_spark_analysis
       ENDIF
    ENDIF
    IF ((duration_write_lastbeats .GT. 0.0d0) .AND. (time .GE. (t_end - duration_write_lastbeats))) THEN
       CALL data_io_write()
    ENDIF
   #endif
    
    ! print out data analysis for resting simulation
    IF ((time .GE. time_start_spark_analysis) .OR. &
         (write_firstbeat .NE. 0 .AND. time .LE. t_start+1.0d0)  ) THEN
#if _ANALYZE_SPARKS_ == _YES
   #if _RUNMODE_ == _MODE_RESTING || _RUNMODE_ == _MODE_CLOSEDCELL
       ! for resting or closed-cell, only do spark analysis
       CALL analyze_Sparks <<<gridSize, blocksize>>> (REAL(time), REAL(Ca_nsr))
   #endif
  #if _RUNMODE_ == _MODE_V_CLAMP || _RUNMODE_ == _MODE_I_CLAMP
       !IF (Hz .LE. 1.0d0) THEN
          ! only do diastolic spark analysis when Hz .le. 1
       !   IF (is_in_DiastolicPhase) THEN
             !end an AP, can measure diastolic spark statistics
             !until the end of that second. This works when Hz < 1.
             CALL analyze_Sparks <<<gridSize, blocksize>>> (REAL(time), REAL(Ca_nsr))
             IF (time .GT. CEILING(time_start_DiastolicPhase)) is_in_DiastolicPhase = .FALSE.
       !   END IF
       !ENDIF
   #endif
#endif

#if _ANALYZE_ECC_ == _YES || _ANALYZE_SL_CALCIUM_FLUX_ == _YES || _ANALYZE_SR_CALCIUM_FLUX_ == _YES
       ! apd analysis + spark analysis in 1Hz stimulation of diastolic phase
       J_ryr_total(1) = sum_reduce_double_2host(J_ryr_dev, NSFU, odata_d)   !based on /1L cytosolic volume
              
       IF (NSFU_normal .LT. NSFU)  J_ryr_orphan_total(1) = &
            sum_reduce_double_2host(J_ryr_dev(NSFU_normal+1:NSFU), NSFU_orphan, odata_d)

       J_dhpr_total = -I_dhpr_total(1) * inv_2FVmyo
       J_SRrelease_total = J_leak+J_ryr_total(1)+J_ryr_nj_T+J_ryr_orphan

       IF (tresidue.GT. t_V_stim_start .AND. tresidue.LT. t_V_stim_end-10.0e-3) THEN
          !!!NOTE: Only integrate from the time it triggered to before the end of the V-clamp
          !accumulate_J_ryr = accumulate_J_ryr + (J_leak+J_ryr_total(1)+J_ryr_nj_T+J_ryr_orphan) * (time - time_prev2)
          accumulate_J_ryr = accumulate_J_ryr + (J_SRrelease_total) * (time - time_prev2)
          
          accumulate_J_dhpr = accumulate_J_dhpr + J_dhpr_total * (time - time_prev2)
          accumulate_J_ryr_orphan = accumulate_J_ryr_orphan + J_ryr_orphan_total(1) * (time - time_prev2)
          IF (peak_J_SRrelease_total < J_SRrelease_total) peak_J_SRrelease_total = J_SRrelease_total
          IF (peak_J_dhpr_total < J_dhpr_total) peak_J_dhpr_total = J_dhpr_total
       END IF
   #if _ANALYZE_SL_CALCIUM_FLUX_ == _YES
       !wrong: accumulate_J_SL_in = accumulate_J_SL_in +  zCa * I_ncx * Am_2FVmyo * (time - time_prev2)
       accumulate_J_SL_in = accumulate_J_SL_in +  (-I_bCa * Am_2FVmyo - I_dhpr_nj_T*inv_2FVmyo) * (time - time_prev2)       
       !wrong: accumulate_J_SL_out = accumulate_J_SL_out + ((I_bCa - I_pmca) * Am_2FVmyo + I_dhpr_nj_T * inv_2FVmyo) * (time - time_prev2)
       accumulate_J_SL_out = accumulate_J_SL_out + (I_pmca - zCa * I_ncx) * Am_2FVmyo * (time - time_prev2)
   #endif
   #if _ANALYZE_SR_CALCIUM_FLUX_ == _YES
       accumulate_J_SR_in = accumulate_J_SR_in + J_serca * (time - time_prev2)
       !accumulate_J_SR_out = accumulate_J_SR_out + (J_leak + J_ryr_total(1) + J_ryr_nj_T + J_ryr_orphan) * (time - time_prev2)
       accumulate_J_SR_out = accumulate_J_SR_out + (J_SRrelease_total) * (time - time_prev2)
   #endif
       !!PRINT *, accumulate_J_ryr_orphan
       !WRITE(*, "(5(2x,1pe12.4))") J_ryr_total(1), accumulate_J_ryr, Vol_ref/V_myo_T, t_end-time_start_spark_analysis

       time_prev2 = time
       
       DO ii = 1, NUM_APD_TYPES
          IF (Vm .GE. Vmthreshold(ii) .AND. is_inside_AP(ii) .EQ. .FALSE.) THEN
             is_inside_AP(ii) = .TRUE.
             APstart(ii) = time
          END IF
          IF (Vm .LE. Vmthreshold(ii) .AND. is_inside_AP(ii) .EQ. .TRUE.) THEN
             is_inside_AP(ii) = .FALSE.
             APend(ii) = time
             IF (APstart(ii) .LT. APend(ii)) THEN
                APD(ii) = APD(ii) + (APend(ii) - APstart(ii))
                APcount(ii) = APcount(ii) + 1
             END IF
             IF (ii .EQ. 1) THEN
                is_in_DiastolicPhase = .TRUE.
                time_start_DiastolicPhase = time
             END IF
          END IF
       END DO
#endif       
    ENDIF
    
  END FUNCTION  simulate_NextSteps


!#if _LCC_MODEL_ == _WILLIAMS2013 || _LCC_MODEL_ == _WILLIAMS2013_NEWALGORITHM
#if _LCC_MODEL_ == _WILLIAMS2013  
  !!>****f* tscell_gpu_utility/get_compPdt_2
  !! SYNOPSIS  
  !!    [Device] Get the transition rate at which the SFU doesn't
  !!    change state. This will help determining the appropriate time
  !!    step so that 90% of the CRUs doesn't change the state. 
  !!   
  !! INPUTS
  !!   kva, kvi :: 2 parameters associated with transition rates formula
  !! OUTPUTS 
  !! RETURN 
  !! SIDE EFFECTS 
  !! HISTORY 
  !! BUGS 
  !!_****
  attributes(global) SUBROUTINE get_compPdt_2(kva, kvi)
    USE cudafor
    IMPLICIT NONE
    
    REAL(KIND=DP), VALUE:: kva, kvi
    ! local
    REAL(KIND=dp) :: bigC1, bigC2, chiC, chiO, phi, dtemp, dtemp2
    INTEGER:: tx, ii !,  ct   ! ct = current state
    REAL(KIND=dp), DIMENSION(BLOCKSIZE_1D), shared :: s_cads, s_ryrgate
    INTEGER,VOLATILE :: lcc_ct, ryr_ct
    
    ii =  (blockIdx%x-1) * blockDim%x + threadIdx%x
    
    IF (ii .LE. int_ca(1)) THEN
       tx = threadIdx%x
       s_cads(tx) = Ca_ds_dev(ii)
       s_ryrgate(tx) = RYRgate_dev(ii)
    ENDIF
    CALL syncthreads()
       
    IF (ii .LE. int_ca(1)) THEN
       CALL state_breakdown_ker(CRU_states_dev(ii), int_ca(10), lcc_ct, ryr_ct)
       !bigC2 = Ca_ds_dev(ii)**eta_RYR
       !bigC1 = Ca_ds_dev(ii)**eta_LCC
       dtemp = LOG(s_cads(tx))
       bigC2 = EXP(dtemp*dp_ca(14))
       bigC1 = EXP(dtemp*dp_ca(15))
       
       !chiC = EXP(-Ej*eta_CO*((N_R-RYRgate_dev(ii))*Ecc - (RYRgate_dev(ii)-1.d0)*Eoo))
       dtemp = (int_ca(2)-s_ryrgate(tx))*dp_ca(2)
       dtemp2 = dtemp - (s_ryrgate(tx) - 1.d0)*dp_ca(3)
       dtemp = dp_ca(1)*dtemp2
       chiC = EXP(dtemp)
       
       !chiO = EXP(-Ej*(1-eta_CO)*(RYRgate_dev(ii)*Eoo - (N_R-RYRgate_dev(ii)- 1.d0)*Ecc))
       dtemp = s_ryrgate(tx)*dp_ca(3)
       dtemp2 = dtemp - (int_ca(2) - s_ryrgate(tx) - 1.d0)*dp_ca(2)
       dtemp = dp_ca(12)*dtemp2
       chiO = EXP(dtemp)

       !phi = k_jsr0*Ca_jsr_dev(ii) + k_jsr1
       phi = dp_ca(4)*Ca_jsr_dev(ii) + dp_ca(5)

       ! use 1st column for negative value
       dtemp = chiC * ryr_compKfree_dev(ryr_ct,1)  
       dtemp2 = dtemp + bigC2 * chiO * phi * ryr_compKcal_act_dev(ryr_ct,1)
       dtemp = dtemp2 + kvi * lcc_compKv_inact_dev(lcc_ct,1) 
       dtemp2 = dtemp + kva * lcc_compKv_act_dev(lcc_ct,1)
       dtemp = dtemp2 + bigC1 * lcc_compKcal_inact_dev(lcc_ct,1) 
       compPdt_dev(ii) = dtemp + lcc_compKfree_dev(lcc_ct,1) 
    END IF
  END SUBROUTINE get_compPdt_2

  attributes(global) SUBROUTINE get_compPdt_2b(kva, kvi, NSFU_normal,  Ej_orphan, eta_CO)
    !deal with orphan CRUs
    USE cudafor
    IMPLICIT NONE
    
    REAL(KIND=DP), VALUE:: kva, kvi, Ej_orphan, eta_CO
    INTEGER, VALUE :: NSFU_normal
    ! local
    REAL(KIND=dp) :: bigC1, bigC2, chiC, chiO, phi, dtemp, dtemp2, Ej_half_CO, Ej_half_OC
    INTEGER:: tx, ii
    REAL(KIND=dp), DIMENSION(BLOCKSIZE_1D), shared :: s_cads, s_ryrgate
    INTEGER,VOLATILE :: lcc_ct, ryr_ct
    
    ii =  (blockIdx%x-1) * blockDim%x + threadIdx%x
    
    IF (ii .LE. int_ca(1)) THEN
       tx = threadIdx%x
       s_cads(tx) = Ca_ds_dev(ii)
       s_ryrgate(tx) = RYRgate_dev(ii)
       IF (ii .LE. NSFU_normal) THEN
          Ej_half_CO = dp_ca(1)
          Ej_half_OC = dp_ca(12)
       ELSE
          Ej_half_CO = -Ej_orphan*eta_CO
          Ej_half_OC = -Ej_orphan*(1-eta_CO)
       ENDIF
    ENDIF
    CALL syncthreads()
       
    IF (ii .LE. int_ca(1)) THEN
       CALL state_breakdown_ker(CRU_states_dev(ii), int_ca(10), lcc_ct, ryr_ct)
       !bigC2 = Ca_ds_dev(ii)**eta_RYR
       !bigC1 = Ca_ds_dev(ii)**eta_LCC
       dtemp = LOG(s_cads(tx))
       bigC2 = EXP(dtemp*dp_ca(14))
       bigC1 = EXP(dtemp*dp_ca(15))
       
       !chiC = EXP(-Ej*eta_CO*((N_R-RYRgate_dev(ii))*Ecc - (RYRgate_dev(ii)-1.d0)*Eoo))
       dtemp = (int_ca(2)-s_ryrgate(tx))*dp_ca(2)
       dtemp2 = dtemp - (s_ryrgate(tx) - 1.d0)*dp_ca(3)
       dtemp = Ej_half_CO*dtemp2
       chiC = EXP(dtemp)
       
       !chiO = EXP(-Ej*(1-eta_CO)*(RYRgate_dev(ii)*Eoo - (N_R-RYRgate_dev(ii)- 1.d0)*Ecc))
       dtemp = s_ryrgate(tx)*dp_ca(3)
       dtemp2 = dtemp - (int_ca(2) - s_ryrgate(tx) - 1.d0)*dp_ca(2)
       dtemp = Ej_half_OC*dtemp2
       chiO = EXP(dtemp)

       !phi = k_jsr0*Ca_jsr_dev(ii) + k_jsr1
       phi = dp_ca(4)*Ca_jsr_dev(ii) + dp_ca(5)

       ! use 1st column for negative value
       dtemp = chiC * ryr_compKfree_dev(ryr_ct,1)  
       dtemp2 = dtemp + bigC2 * chiO * phi * ryr_compKcal_act_dev(ryr_ct,1)
       dtemp = dtemp2 + kvi * lcc_compKv_inact_dev(lcc_ct,1) 
       dtemp2 = dtemp + kva * lcc_compKv_act_dev(lcc_ct,1)
       dtemp = dtemp2 + bigC1 * lcc_compKcal_inact_dev(lcc_ct,1) 
       compPdt_dev(ii) = dtemp + lcc_compKfree_dev(lcc_ct,1) 
    END IF
  END SUBROUTINE get_compPdt_2b

  !=============================================================================
  ! Update CRU's state + calcium concentration (Ca-ds, Ca-jsr)
  !=============================================================================
  attributes(global) SUBROUTINE update_SFU_2(Xrnd, dt, invdt, &
       Vm, kva, kvi, dp_arg1, dp_arg2, Ca_myo, Ca_nsr) 
    !new method
    USE cudafor
    IMPLICIT NONE
    REAL(KIND=dp), DIMENSION(:), device :: Xrnd
    REAL(KIND=dp), VALUE :: dt,  invdt, Ca_nsr, Ca_myo, Vm, kva, kvi, dp_arg1, dp_arg2

    !local
    INTEGER :: ii, jj, kk, tx
    REAL(KIND=dp) :: dtemp, dtemp2, compP, J_ryr
    !INTEGER :: nt, ct, cnt
    !nt = new state, ct = current state, cnt= candidate next state
    REAL(KIND=dp) :: bigC1, bigC2, chiC, chiO, phi, J_dhpr
    REAL(KIND=dp) :: nryr ! new RYRgate
    REAL(KIND=dp) :: ndhpr !new dhprgate
    INTEGER :: itemp
    REAL(KIND=dp), DIMENSION(BLOCKSIZE_1D), shared :: s_cads, s_ryrgate
    INTEGER, VOLATILE :: lcc_ct, ryr_ct
    INTEGER :: lcc_nt, ryr_nt
    INTEGER :: is_examining
    LOGICAL :: found_nt
    INTEGER :: candidate_new_state !can be any cluster types (LCC or RYR)
    INTEGER, PARAMETER :: EXAMINING_LCC = 1, EXAMINING_RYR = 2

    ! thread Index
    ii =  (blockIdx%x-1) * blockDim%x + threadIdx%x

    IF (ii .LE. int_ca(1)) THEN
       tx = threadIdx%x
       s_cads(tx) = Ca_ds_dev(ii)
       s_ryrgate(tx) = RYRgate_dev(ii)
    ENDIF
    CALL syncthreads()

    IF (ii .LE. int_ca(1)) THEN
       CALL state_breakdown_ker(CRU_states_dev(ii), int_ca(10), lcc_ct, ryr_ct)

       lcc_nt = lcc_ct
       ryr_nt = ryr_ct
       found_nt = .FALSE.
       compP = (compPdt_dev(ii) + invdt) * dt

       IF (compP .GE. Xrnd(ii)) THEN ! channel doesn't change state (easy)
          found_nt = .TRUE.
       ELSE ! channel change to some new state
          dtemp = LOG(s_cads(tx))
          dtemp2 = dtemp *dp_ca(14)
          bigC2 = EXP(dtemp2)
          dtemp2 = dtemp*dp_ca(15)
          bigC1 = EXP(dtemp2)

          dtemp = (int_ca(2)-s_ryrgate(tx))*dp_ca(2)
          dtemp2 = dtemp - (s_ryrgate(tx) - 1.d0)*dp_ca(3)
          dtemp = dp_ca(1)*dtemp2
          chiC = EXP(dtemp)
          !dtemp = (int_ca(2)-RYRgate_dev(ii))*dp_ca(2)
          !dtemp2 = dtemp - (RYRgate_dev(ii) - 1.d0)*dp_ca(3)
          !dtemp = dp_ca(1)*dtemp2
          !chiC = EXP(dtemp)

          dtemp = s_ryrgate(tx)*dp_ca(3)
          dtemp2 = dtemp - (int_ca(2) - s_ryrgate(tx) - 1.d0)*dp_ca(2)
          dtemp = dp_ca(12)*dtemp2
          chiO = EXP(dtemp)
          !dtemp = RYRgate_dev(ii)*dp_ca(3)
          !dtemp2= dtemp - (int_ca(2)-RYRgate_dev(ii)- 1.d0)*dp_ca(2)
          !dtemp = dp_ca(12)*dtemp2
          !chiO = EXP(dtemp)

          phi = dp_ca(4)*Ca_jsr_dev(ii) + dp_ca(5)

          DO jj=2, int_ca(9)
             IF (found_nt .EQ. .FALSE.) THEN
                IF (jj .LE. int_ca(3)) THEN
                   kk = jj
                   dtemp = bigC2 * chiO * phi * ryr_compKcal_act_dev(ryr_ct,kk)
                   candidate_new_state = ryr_indxKcal_act_dev(ryr_ct,kk)
                   is_examining = EXAMINING_RYR
                ELSE IF (jj .LE. int_ca(3)+int_ca(4)-1) THEN
                   kk = jj-(int_ca(3)-1)
                   dtemp = chiC * ryr_compKfree_dev(ryr_ct,kk)
                   candidate_new_state = ryr_indxKfree_dev(ryr_ct,kk)
                   is_examining = EXAMINING_RYR
                   
                ELSE IF (jj .LE. int_ca(3)+int_ca(4)+int_ca(5)-2) THEN
                   kk = jj-(int_ca(3)+int_ca(4)-2)
                   dtemp = kva*lcc_compKv_act_dev(lcc_ct,kk)
                   candidate_new_state = lcc_indxKv_act_dev(lcc_ct,kk)
                   is_examining = EXAMINING_LCC

                ELSE IF (jj .LE. int_ca(3)+int_ca(4)+int_ca(5)+int_ca(6)-3) THEN
                   kk = jj-(int_ca(3)+int_ca(4)+int_ca(5)-3)
                   dtemp = kvi*lcc_compKv_inact_dev(lcc_ct,kk)
                   candidate_new_state = lcc_indxKv_inact_dev(lcc_ct,kk)
                   is_examining = EXAMINING_LCC
                   
                ELSE  IF (jj .LE. int_ca(3)+int_ca(4)+int_ca(5)+int_ca(6)+int_ca(7)-4) THEN
                   kk = jj-(int_ca(3)+int_ca(4)+int_ca(5)+int_ca(6)-4)
                   dtemp = bigC1 * lcc_compKcal_inact_dev(lcc_ct,kk)
                   candidate_new_state = lcc_indxKcal_inact_dev(lcc_ct,kk)
                   is_examining = EXAMINING_LCC                   
                   
                ELSE !IF (jj .LE. int_ca(3)+int_ca(4)+int_ca(5)+int_ca(6)+int_ca(7)+int_ca(8)-5) THEN  ! the last
                   kk = jj-(int_ca(3)+int_ca(4)+int_ca(5)+int_ca(6)+int_ca(7)-5)
                   dtemp = lcc_compKfree_dev(lcc_ct,kk)
                   candidate_new_state = lcc_indxKfree_dev(lcc_ct,kk)
                   is_examining = EXAMINING_LCC                   
                   
                   !if we need more else, need to retain IF in the previous one
                END IF
                compP = compP + dtemp * dt 
                IF (compP .GE. Xrnd(ii)) THEN
                   found_nt = .TRUE.
                   SELECT CASE (is_examining)
                   CASE(EXAMINING_LCC)
                      !to avoid LCC gating at negative potential
                      !!!IF (Vm .GE. -70.0d0)  lcc_nt = candidate_new_state
                      lcc_nt = candidate_new_state
                   CASE(EXAMINING_RYR)
                      ryr_nt = candidate_new_state
                   END SELECT
                END IF
             END IF
          END DO
       END IF

       nryr = U_Ro_dev(ryr_nt)
       ndhpr = U_Lo_dev(lcc_nt)

       J_efflux_dev(ii) = dp_ca(11)*(s_cads(tx)-Ca_myo)
       J_refill_dev(ii) = dp_ca(10)*(Ca_nsr-Ca_jsr_dev(ii))

#if (_RUNMODE_ == _MODE_I_CLAMP || _RUNMODE_ == _MODE_V_CLAMP || _RUNMODE_ == _MODE_V_CLAMP_TEST_LCC)
       !I-clamp or V-clamp
       I_dhpr_dev(ii)  = ndhpr * dp_ca(13) * dp_ca(16) * Vm * &
            (s_cads(tx) * dp_arg1 - dp_ca(18)) * dp_arg2  ! [uA]
       !J_dhpr_dev(ii)  = -I_dhpr_dev(ii) * dp_ca(17)  ! [uA/s]
       J_dhpr  = -I_dhpr_dev(ii) * dp_ca(17)  ! [uA/s]
#else
       I_dhpr_dev(ii)  = 0.0d0
       !J_dhpr_dev(ii) = 0d0
       J_dhpr = 0d0
#endif

       !!J_ryr = nryr * dp_ca(9) * (Ca_jsr_dev(ii)-Ca_ds_dev(ii)) !J_ryr
       J_ryr = nryr * dp_ca(9) * (Ca_jsr_dev(ii)-s_cads(tx))
       J_ryr_dev(ii) = J_ryr
#if _USE_CaDS_APPROX_ == _NO
       !exact formula
       !s_cads(tx) = s_cads(tx) + dt * dp_ca(6) * dp_ca(7) * (J_ryr - J_efflux_dev(ii) + J_dhpr_dev(ii))
       s_cads(tx) = s_cads(tx) + dt * dp_ca(6) * dp_ca(7) * (J_ryr - J_efflux_dev(ii) + J_dhpr)
#else
       dtemp = ndhpr * dp_ca(13) * dp_ca(16) * Vm * dp_arg2 * dp_ca(17)
       dtemp2 = nryr * dp_ca(9)
       s_cads(tx) = (dtemp2 * Ca_jsr_dev(ii) + dp_ca(11) * Ca_myo + dtemp * dp_ca(18))/&
            (dtemp2 + dp_ca(11) + dtemp)
       !dtemp = (dp_ca(13)*dp_ca(16) * Vm * dp_ca(18) * dp_arg2) * &
       !     dp_ca(17) * ndhpr + dp_ca(11) * Ca_myo + nryr * dp_ca(9) * Ca_jsr_dev(ii)
       !dtemp2 = nryr * dp_ca(9) + dp_ca(11) + (dp_ca(13) * dp_ca(16) * Vm * dp_arg1 * dp_arg2)  * dp_ca(17) * ndhpr
       !s_cads(tx) = dtemp / dtemp2       
#endif
       dtemp = dp_ca(19) + Ca_jsr_dev(ii)
       dtemp2 = dtemp * dtemp
       dtemp = dtemp2 / (dp_ca(19)*dp_ca(20) + dtemp2)   !beta_jsr
       Ca_jsr_dev(ii) = Ca_jsr_dev(ii) + dt*(dtemp * dp_ca(8)) * (J_refill_dev(ii) - J_ryr)

       CRU_states_dev(ii) =  state_combine_single_ker(lcc_nt, ryr_nt, int_ca(10))
       !CRU_states_dev(ii) = (lcc_nt-1) * int_ca(10) + ryr_nt
       RYRgate_dev(ii) = nryr
       DHPRgate_dev(ii) = ndhpr
       Ca_ds_dev(ii) = s_cads(tx)
    END IF
  END SUBROUTINE update_SFU_2

  !! to work with testing orphan CRUs in which V_ds increase in these orphan CRUs 
  attributes(global) SUBROUTINE update_SFU_2b(Xrnd, dt, invdt, &
       Vm, kva, kvi, dp_arg1, dp_arg2, Ca_myo, Ca_nsr, &
       NSFU_normal, inv_lambda_ds_orphan, Km_myo, B_myoT, Ej_orphan, eta_CO) 
    !new method
    USE cudafor
    IMPLICIT NONE
    REAL(KIND=dp), DIMENSION(:), device :: Xrnd
    REAL(KIND=dp), VALUE :: dt,  invdt, Ca_nsr, Ca_myo, Vm, kva, kvi, dp_arg1, dp_arg2, inv_lambda_ds_orphan, &
         Km_myo, B_myoT, Ej_orphan, eta_CO
    INTEGER, VALUE ::  NSFU_normal
    
    !local
    INTEGER :: ii, jj, kk, tx
    REAL(KIND=dp) :: dtemp, dtemp2, compP, J_ryr, Ej_half_CO, Ej_half_OC
    !INTEGER :: nt, ct !, cnt
    !nt = new state, ct = current state, cnt= candidate next state
    REAL(KIND=dp) :: bigC1, bigC2, chiC, chiO, phi, J_dhpr
    INTEGER :: nryr ! new RYRgate
    INTEGER :: ndhpr !new dhprgate
    INTEGER :: itemp
    REAL(KIND=dp), DIMENSION(BLOCKSIZE_1D), shared :: s_cads, s_ryrgate
    INTEGER, VOLATILE :: lcc_ct, ryr_ct
    INTEGER :: lcc_nt, ryr_nt
    INTEGER :: is_examining
    LOGICAL :: found_nt
    INTEGER :: candidate_new_state !can be any cluster types (LCC or RYR)
    INTEGER, PARAMETER :: EXAMINING_LCC = 1, EXAMINING_RYR = 2

    ! thread Index
    ii =  (blockIdx%x-1) * blockDim%x + threadIdx%x

    IF (ii .LE. int_ca(1)) THEN
       tx = threadIdx%x
       s_cads(tx) = Ca_ds_dev(ii)
       s_ryrgate(tx) = RYRgate_dev(ii)
       IF (ii .LE. NSFU_normal) THEN
          Ej_half_CO = dp_ca(1)
          Ej_half_OC = dp_ca(12)
       ELSE
          Ej_half_CO = -Ej_orphan*eta_CO
          Ej_half_OC = -Ej_orphan*(1-eta_CO)
       ENDIF
    ENDIF
    CALL syncthreads()

    IF (ii .LE. int_ca(1)) THEN
       CALL state_breakdown_ker(CRU_states_dev(ii), int_ca(10), lcc_ct, ryr_ct)

       lcc_nt = lcc_ct
       ryr_nt = ryr_ct
       found_nt = .FALSE.
       compP = (compPdt_dev(ii) + invdt) * dt

       IF (compP .GE. Xrnd(ii)) THEN ! channel doesn't change state (easy)
          found_nt = .TRUE.
       ELSE ! channel change to some new state
          dtemp = LOG(s_cads(tx))
          dtemp2 = dtemp *dp_ca(14)
          bigC2 = EXP(dtemp2)
          dtemp2 = dtemp*dp_ca(15)
          bigC1 = EXP(dtemp2)

          dtemp = (int_ca(2)-s_ryrgate(tx))*dp_ca(2)
          dtemp2 = dtemp - (s_ryrgate(tx) - 1.d0)*dp_ca(3)
          dtemp = Ej_half_CO*dtemp2
          chiC = EXP(dtemp)
          !dtemp = (int_ca(2)-RYRgate_dev(ii))*dp_ca(2)
          !dtemp2 = dtemp - (RYRgate_dev(ii) - 1.d0)*dp_ca(3)
          !dtemp = dp_ca(1)*dtemp2
          !chiC = EXP(dtemp)

          dtemp = s_ryrgate(tx)*dp_ca(3)
          dtemp2 = dtemp - (int_ca(2) - s_ryrgate(tx) - 1.d0)*dp_ca(2)
          dtemp = Ej_half_OC*dtemp2
          chiO = EXP(dtemp)
          !dtemp = RYRgate_dev(ii)*dp_ca(3)
          !dtemp2= dtemp - (int_ca(2)-RYRgate_dev(ii)- 1.d0)*dp_ca(2)
          !dtemp = dp_ca(12)*dtemp2
          !chiO = EXP(dtemp)

          phi = dp_ca(4)*Ca_jsr_dev(ii) + dp_ca(5)

          DO jj=2, int_ca(9)
             IF (found_nt .EQ. .FALSE.) THEN
                IF (jj .LE. int_ca(3)) THEN
                   kk = jj
                   dtemp = bigC2 * chiO * phi * ryr_compKcal_act_dev(ryr_ct,kk)
                   candidate_new_state = ryr_indxKcal_act_dev(ryr_ct,kk)
                   is_examining = EXAMINING_RYR
                ELSE IF (jj .LE. int_ca(3)+int_ca(4)-1) THEN
                   kk = jj-(int_ca(3)-1)
                   dtemp = chiC * ryr_compKfree_dev(ryr_ct,kk)
                   candidate_new_state = ryr_indxKfree_dev(ryr_ct,kk)
                   is_examining = EXAMINING_RYR
                   
                ELSE IF (jj .LE. int_ca(3)+int_ca(4)+int_ca(5)-2) THEN
                   kk = jj-(int_ca(3)+int_ca(4)-2)
                   dtemp = kva*lcc_compKv_act_dev(lcc_ct,kk)
                   candidate_new_state = lcc_indxKv_act_dev(lcc_ct,kk)
                   is_examining = EXAMINING_LCC

                ELSE IF (jj .LE. int_ca(3)+int_ca(4)+int_ca(5)+int_ca(6)-3) THEN
                   kk = jj-(int_ca(3)+int_ca(4)+int_ca(5)-3)
                   dtemp = kvi*lcc_compKv_inact_dev(lcc_ct,kk)
                   candidate_new_state = lcc_indxKv_inact_dev(lcc_ct,kk)
                   is_examining = EXAMINING_LCC
                   
                ELSE  IF (jj .LE. int_ca(3)+int_ca(4)+int_ca(5)+int_ca(6)+int_ca(7)-4) THEN
                   kk = jj-(int_ca(3)+int_ca(4)+int_ca(5)+int_ca(6)-4)
                   dtemp = bigC1 * lcc_compKcal_inact_dev(lcc_ct,kk)
                   candidate_new_state = lcc_indxKcal_inact_dev(lcc_ct,kk)
                   is_examining = EXAMINING_LCC                   
                   
                ELSE !IF (jj .LE. int_ca(3)+int_ca(4)+int_ca(5)+int_ca(6)+int_ca(7)+int_ca(8)-5) THEN  ! the last
                   kk = jj-(int_ca(3)+int_ca(4)+int_ca(5)+int_ca(6)+int_ca(7)-5)
                   dtemp = lcc_compKfree_dev(lcc_ct,kk)
                   candidate_new_state = lcc_indxKfree_dev(lcc_ct,kk)
                   is_examining = EXAMINING_LCC                   
                   
                   !if we need more else, need to retain IF in the previous one
                END IF
                compP = compP + dtemp * dt 
                IF (compP .GE. Xrnd(ii)) THEN
                   found_nt = .TRUE.
                   SELECT CASE (is_examining)
                   CASE(EXAMINING_LCC)
                      !to avoid LCC gating at negative potential
                      !IF (Vm .GE. -70.0d0)  lcc_nt = candidate_new_state
                      lcc_nt = candidate_new_state
                   CASE(EXAMINING_RYR)
                      ryr_nt = candidate_new_state
                   END SELECT
                END IF
             END IF
          END DO
       END IF

       nryr = U_Ro_dev(ryr_nt)
       ndhpr = U_Lo_dev(lcc_nt)

       J_efflux_dev(ii) = dp_ca(11)*(s_cads(tx)-Ca_myo)
       J_refill_dev(ii) = dp_ca(10)*(Ca_nsr-Ca_jsr_dev(ii))

#if (_RUNMODE_ == _MODE_I_CLAMP || _RUNMODE_ == _MODE_V_CLAMP || _RUNMODE_ == _MODE_V_CLAMP_TEST_LCC)
       !I-clamp or V-clamp
       I_dhpr_dev(ii)  = ndhpr * dp_ca(13) * dp_ca(16) * Vm * &
            (s_cads(tx) * dp_arg1 - dp_ca(18)) * dp_arg2  ! [uA]
       !J_dhpr_dev(ii)  = -I_dhpr_dev(ii) * dp_ca(17)  ! [uA/s]
       J_dhpr  = -I_dhpr_dev(ii) * dp_ca(17)  ! [uA/s]
#else
       I_dhpr_dev(ii)  = 0.0d0
       !J_dhpr_dev(ii) = 0d0
       J_dhpr = 0d0
#endif

       !!J_ryr = nryr * dp_ca(9) * (Ca_jsr_dev(ii)-Ca_ds_dev(ii)) !J_ryr
       J_ryr = nryr * dp_ca(9) * (Ca_jsr_dev(ii)-s_cads(tx))
       J_ryr_dev(ii) = J_ryr
#if _USE_CaDS_APPROX_ == _NO
       !exact formula
       !s_cads(tx) = s_cads(tx) + dt * dp_ca(6) * dp_ca(7) * (J_ryr - J_efflux_dev(ii) + J_dhpr_dev(ii))
       dtemp = (Km_myo + s_cads(tx))**2
       dtemp2 = dtemp +  B_myoT * Km_myo
       dtemp = dtemp/dtemp2  ! == beta_ds
       IF (ii .LE. NSFU_normal) THEN 
          dtemp2 = dp_ca(7)
       ELSE 
          dtemp2 = inv_lambda_ds_orphan
       ENDIF
       s_cads(tx) = s_cads(tx) + dt * dtemp * dtemp2 * (J_ryr - J_efflux_dev(ii) + J_dhpr)
#else
       dtemp = ndhpr * dp_ca(13) * dp_ca(16) * Vm * dp_arg2 * dp_ca(17)
       dtemp2 = nryr * dp_ca(9)
       s_cads(tx) = (dtemp2 * Ca_jsr_dev(ii) + dp_ca(11) * Ca_myo + dtemp * dp_ca(18))/&
            (dtemp2 + dp_ca(11) + dtemp)
       !dtemp = (dp_ca(13)*dp_ca(16) * Vm * dp_ca(18) * dp_arg2) * &
       !     dp_ca(17) * ndhpr + dp_ca(11) * Ca_myo + nryr * dp_ca(9) * Ca_jsr_dev(ii)
       !dtemp2 = nryr * dp_ca(9) + dp_ca(11) + (dp_ca(13) * dp_ca(16) * Vm * dp_arg1 * dp_arg2)  * dp_ca(17) * ndhpr
       !s_cads(tx) = dtemp / dtemp2       
#endif
       dtemp = dp_ca(19) + Ca_jsr_dev(ii)
       dtemp2 = dtemp * dtemp
       dtemp = dtemp2 / (dp_ca(19)*dp_ca(20) + dtemp2)   !beta_jsr
       Ca_jsr_dev(ii) = Ca_jsr_dev(ii) + dt*(dtemp * dp_ca(8)) * (J_refill_dev(ii) - J_ryr)

       !CRU_states_dev(ii) =  state_combine_single_ker(lcc_cnt, ryr_cnt, int_ca(10))
       CRU_states_dev(ii) = (lcc_nt-1) * int_ca(10) + ryr_nt
       RYRgate_dev(ii) = nryr
       DHPRgate_dev(ii) = ndhpr
       Ca_ds_dev(ii) = s_cads(tx)
    END IF
  END SUBROUTINE update_SFU_2b

  attributes(global) SUBROUTINE update_SFU_noLCCstatechanges(Xrnd, dt, invdt, &
       Vm, kva, kvi, dp_arg1, dp_arg2, Ca_myo, Ca_nsr) 
    !new method
    USE cudafor
    IMPLICIT NONE
    REAL(KIND=dp), DIMENSION(:), device :: Xrnd
    REAL(KIND=dp), VALUE :: dt,  invdt, Ca_nsr, Ca_myo, Vm, kva, kvi, dp_arg1, dp_arg2

    !local
    INTEGER :: ii, jj, kk, tx
    REAL(KIND=dp) :: dtemp, dtemp2, compP, J_ryr
    !INTEGER :: nt, ct, cnt
    !nt = new state, ct = current state, cnt= candidate next state
    REAL(KIND=dp) :: bigC1, bigC2, chiC, chiO, phi, J_dhpr
    REAL(KIND=dp) :: nryr ! new RYRgate
    REAL(KIND=dp) :: ndhpr !new dhprgate
    INTEGER :: itemp
    REAL(KIND=dp), DIMENSION(BLOCKSIZE_1D), shared :: s_cads, s_ryrgate
    INTEGER, VOLATILE :: lcc_ct, ryr_ct
    INTEGER :: lcc_nt, ryr_nt
    INTEGER :: is_examining
    LOGICAL :: found_nt
    INTEGER :: candidate_new_state !can be any cluster types (LCC or RYR)
    INTEGER, PARAMETER :: EXAMINING_LCC = 1, EXAMINING_RYR = 2

    ! thread Index
    ii =  (blockIdx%x-1) * blockDim%x + threadIdx%x

    IF (ii .LE. int_ca(1)) THEN
       tx = threadIdx%x
       s_cads(tx) = Ca_ds_dev(ii)
       s_ryrgate(tx) = RYRgate_dev(ii)
    ENDIF
    CALL syncthreads()

    IF (ii .LE. int_ca(1)) THEN
       CALL state_breakdown_ker(CRU_states_dev(ii), int_ca(10), lcc_ct, ryr_ct)

       lcc_nt = lcc_ct
       ryr_nt = ryr_ct
       found_nt = .FALSE.
       compP = (compPdt_dev(ii) + invdt) * dt

       IF (compP .GE. Xrnd(ii)) THEN ! channel doesn't change state (easy)
          found_nt = .TRUE.
       ELSE ! channel change to some new state
          dtemp = LOG(s_cads(tx))
          dtemp2 = dtemp *dp_ca(14)
          bigC2 = EXP(dtemp2)
          dtemp2 = dtemp*dp_ca(15)
          bigC1 = EXP(dtemp2)

          dtemp = (int_ca(2)-s_ryrgate(tx))*dp_ca(2)
          dtemp2 = dtemp - (s_ryrgate(tx) - 1.d0)*dp_ca(3)
          dtemp = dp_ca(1)*dtemp2
          chiC = EXP(dtemp)
          !dtemp = (int_ca(2)-RYRgate_dev(ii))*dp_ca(2)
          !dtemp2 = dtemp - (RYRgate_dev(ii) - 1.d0)*dp_ca(3)
          !dtemp = dp_ca(1)*dtemp2
          !chiC = EXP(dtemp)

          dtemp = s_ryrgate(tx)*dp_ca(3)
          dtemp2 = dtemp - (int_ca(2) - s_ryrgate(tx) - 1.d0)*dp_ca(2)
          dtemp = dp_ca(12)*dtemp2
          chiO = EXP(dtemp)
          !dtemp = RYRgate_dev(ii)*dp_ca(3)
          !dtemp2= dtemp - (int_ca(2)-RYRgate_dev(ii)- 1.d0)*dp_ca(2)
          !dtemp = dp_ca(12)*dtemp2
          !chiO = EXP(dtemp)

          phi = dp_ca(4)*Ca_jsr_dev(ii) + dp_ca(5)

          DO jj=2, int_ca(9)
             IF (found_nt .EQ. .FALSE.) THEN
                IF (jj .LE. int_ca(3)) THEN
                   kk = jj
                   dtemp = bigC2 * chiO * phi * ryr_compKcal_act_dev(ryr_ct,kk)
                   candidate_new_state = ryr_indxKcal_act_dev(ryr_ct,kk)
                   is_examining = EXAMINING_RYR
                ELSE IF (jj .LE. int_ca(3)+int_ca(4)-1) THEN
                   kk = jj-(int_ca(3)-1)
                   dtemp = chiC * ryr_compKfree_dev(ryr_ct,kk)
                   candidate_new_state = ryr_indxKfree_dev(ryr_ct,kk)
                   is_examining = EXAMINING_RYR
                   
                ELSE IF (jj .LE. int_ca(3)+int_ca(4)+int_ca(5)-2) THEN
                   kk = jj-(int_ca(3)+int_ca(4)-2)
                   dtemp = kva*lcc_compKv_act_dev(lcc_ct,kk)
                   candidate_new_state = lcc_indxKv_act_dev(lcc_ct,kk)
                   is_examining = EXAMINING_LCC

                ELSE IF (jj .LE. int_ca(3)+int_ca(4)+int_ca(5)+int_ca(6)-3) THEN
                   kk = jj-(int_ca(3)+int_ca(4)+int_ca(5)-3)
                   dtemp = kvi*lcc_compKv_inact_dev(lcc_ct,kk)
                   candidate_new_state = lcc_indxKv_inact_dev(lcc_ct,kk)
                   is_examining = EXAMINING_LCC
                   
                ELSE  IF (jj .LE. int_ca(3)+int_ca(4)+int_ca(5)+int_ca(6)+int_ca(7)-4) THEN
                   kk = jj-(int_ca(3)+int_ca(4)+int_ca(5)+int_ca(6)-4)
                   dtemp = bigC1 * lcc_compKcal_inact_dev(lcc_ct,kk)
                   candidate_new_state = lcc_indxKcal_inact_dev(lcc_ct,kk)
                   is_examining = EXAMINING_LCC                   
                   
                ELSE !IF (jj .LE. int_ca(3)+int_ca(4)+int_ca(5)+int_ca(6)+int_ca(7)+int_ca(8)-5) THEN  ! the last
                   kk = jj-(int_ca(3)+int_ca(4)+int_ca(5)+int_ca(6)+int_ca(7)-5)
                   dtemp = lcc_compKfree_dev(lcc_ct,kk)
                   candidate_new_state = lcc_indxKfree_dev(lcc_ct,kk)
                   is_examining = EXAMINING_LCC                   
                   
                   !if we need more else, need to retain IF in the previous one
                END IF
                compP = compP + dtemp * dt 
                IF (compP .GE. Xrnd(ii)) THEN
                   found_nt = .TRUE.
                   SELECT CASE (is_examining)
                   CASE(EXAMINING_LCC)
                      !! DO NOTHING AS LCC doesn't CHANGE
                   CASE(EXAMINING_RYR)
                      ryr_nt = candidate_new_state
                   END SELECT
                END IF
             END IF
          END DO
       END IF

       nryr = U_Ro_dev(ryr_nt)
       ndhpr = U_Lo_dev(lcc_nt)

       J_efflux_dev(ii) = dp_ca(11)*(s_cads(tx)-Ca_myo)
       J_refill_dev(ii) = dp_ca(10)*(Ca_nsr-Ca_jsr_dev(ii))

#if (_RUNMODE_ == _MODE_I_CLAMP || _RUNMODE_ == _MODE_V_CLAMP  || _EMULATE_CONDITION_ == _TRIGGER_LCC)
       !I-clamp or V-clamp
       I_dhpr_dev(ii)  = ndhpr * dp_ca(13) * dp_ca(16) * Vm * &
            (s_cads(tx) * dp_arg1 - dp_ca(18)) * dp_arg2  ! [uA]
       !J_dhpr_dev(ii)  = -I_dhpr_dev(ii) * dp_ca(17)  ! [uA/s]
       J_dhpr  = -I_dhpr_dev(ii) * dp_ca(17)  ! [uA/s]
#else
       !J_dhpr_dev(ii) = 0d0
       J_dhpr = 0d0
#endif

       !!J_ryr = nryr * dp_ca(9) * (Ca_jsr_dev(ii)-Ca_ds_dev(ii)) !J_ryr
       J_ryr = nryr * dp_ca(9) * (Ca_jsr_dev(ii)-s_cads(tx))
       J_ryr_dev(ii) = J_ryr
#if _USE_CaDS_APPROX_ == _NO
       !exact formula
       !s_cads(tx) = s_cads(tx) + dt * dp_ca(6) * dp_ca(7) * (J_ryr - J_efflux_dev(ii) + J_dhpr_dev(ii))
       s_cads(tx) = s_cads(tx) + dt * dp_ca(6) * dp_ca(7) * (J_ryr - J_efflux_dev(ii) + J_dhpr)
#else
       dtemp = ndhpr * dp_ca(13) * dp_ca(16) * Vm * dp_arg2 * dp_ca(17)
       dtemp2 = nryr * dp_ca(9)
       s_cads(tx) = (dtemp2 * Ca_jsr_dev(ii) + dp_ca(11) * Ca_myo + dtemp * dp_ca(18))/&
            (dtemp2 + dp_ca(11) + dtemp)
       !dtemp = (dp_ca(13)*dp_ca(16) * Vm * dp_ca(18) * dp_arg2) * &
       !     dp_ca(17) * ndhpr + dp_ca(11) * Ca_myo + nryr * dp_ca(9) * Ca_jsr_dev(ii)
       !dtemp2 = nryr * dp_ca(9) + dp_ca(11) + (dp_ca(13) * dp_ca(16) * Vm * dp_arg1 * dp_arg2)  * dp_ca(17) * ndhpr
       !s_cads(tx) = dtemp / dtemp2       
#endif
       dtemp = dp_ca(19) + Ca_jsr_dev(ii)
       dtemp2 = dtemp * dtemp
       dtemp = dtemp2 / (dp_ca(19)*dp_ca(20) + dtemp2)   !beta_jsr
       Ca_jsr_dev(ii) = Ca_jsr_dev(ii) + dt*(dtemp * dp_ca(8)) * (J_refill_dev(ii) - J_ryr)

       CRU_states_dev(ii) =  state_combine_single_ker(lcc_nt, ryr_nt, int_ca(10))
       RYRgate_dev(ii) = nryr
       DHPRgate_dev(ii) = ndhpr
       Ca_ds_dev(ii) = s_cads(tx)
    END IF
  END SUBROUTINE update_SFU_noLCCstatechanges
  
  attributes(global) SUBROUTINE update_SFU_nostatechanges(Xrnd, dt, invdt, &
       Vm, kva, kvi, dp_arg1, dp_arg2, Ca_myo, Ca_nsr) 
    !new method
    USE cudafor
    IMPLICIT NONE
    REAL(KIND=dp), DIMENSION(:), device :: Xrnd
    REAL(KIND=dp), VALUE :: dt,  invdt, Ca_nsr, Ca_myo, Vm, kva, kvi, dp_arg1, dp_arg2

    !local
    INTEGER :: ii, jj, kk, tx
    REAL(KIND=dp) :: dtemp, dtemp2, compP, J_ryr
    !INTEGER :: nt, ct, cnt
    !nt = new state, ct = current state, cnt= candidate next state
    REAL(KIND=dp) :: bigC1, bigC2, chiC, chiO, phi, J_dhpr
    REAL(KIND=dp) :: nryr ! new RYRgate
    REAL(KIND=dp) :: ndhpr !new dhprgate
    INTEGER :: itemp
    REAL(KIND=dp), DIMENSION(BLOCKSIZE_1D), shared :: s_cads, s_ryrgate
    INTEGER, VOLATILE :: lcc_ct, ryr_ct
    INTEGER :: lcc_nt, ryr_nt
    INTEGER :: is_examining
    LOGICAL :: found_nt
    INTEGER :: candidate_new_state !can be any cluster types (LCC or RYR)
    INTEGER, PARAMETER :: EXAMINING_LCC = 1, EXAMINING_RYR = 2

    ! thread Index
    ii =  (blockIdx%x-1) * blockDim%x + threadIdx%x

    IF (ii .LE. int_ca(1)) THEN
       tx = threadIdx%x
       s_cads(tx) = Ca_ds_dev(ii)
       s_ryrgate(tx) = RYRgate_dev(ii)
    ENDIF
    CALL syncthreads()

    IF (ii .LE. int_ca(1)) THEN
       nryr = RYRgate_dev(ii)
       ndhpr = DHPRgate_dev(ii)

       J_efflux_dev(ii) = dp_ca(11)*(s_cads(tx)-Ca_myo)
       J_refill_dev(ii) = dp_ca(10)*(Ca_nsr-Ca_jsr_dev(ii))

#if (_RUNMODE_ == _MODE_I_CLAMP || _RUNMODE_ == _MODE_V_CLAMP || _EMULATE_CONDITION_ == _TRIGGER_LCC)
       !I-clamp or V-clamp
       I_dhpr_dev(ii)  = ndhpr * dp_ca(13) * dp_ca(16) * Vm * &
            (s_cads(tx) * dp_arg1 - dp_ca(18)) * dp_arg2  ! [uA]
       !J_dhpr_dev(ii)  = -I_dhpr_dev(ii) * dp_ca(17)  ! [uA/s]
       J_dhpr  = -I_dhpr_dev(ii) * dp_ca(17)  ! [uA/s]
#else
       !J_dhpr_dev(ii) = 0d0
       J_dhpr = 0d0
#endif

       !!J_ryr = nryr * dp_ca(9) * (Ca_jsr_dev(ii)-Ca_ds_dev(ii)) !J_ryr
       J_ryr = nryr * dp_ca(9) * (Ca_jsr_dev(ii)-s_cads(tx))
       J_ryr_dev(ii) = J_ryr
#if _USE_CaDS_APPROX_ == _NO
       !exact formula
       !s_cads(tx) = s_cads(tx) + dt * dp_ca(6) * dp_ca(7) * (J_ryr - J_efflux_dev(ii) + J_dhpr_dev(ii))
       s_cads(tx) = s_cads(tx) + dt * dp_ca(6) * dp_ca(7) * (J_ryr - J_efflux_dev(ii) + J_dhpr)
#else
       dtemp = ndhpr * dp_ca(13) * dp_ca(16) * Vm * dp_arg2 * dp_ca(17)
       dtemp2 = nryr * dp_ca(9)
       s_cads(tx) = (dtemp2 * Ca_jsr_dev(ii) + dp_ca(11) * Ca_myo + dtemp * dp_ca(18))/&
            (dtemp2 + dp_ca(11) + dtemp)
       !dtemp = (dp_ca(13)*dp_ca(16) * Vm * dp_ca(18) * dp_arg2) * &
       !     dp_ca(17) * ndhpr + dp_ca(11) * Ca_myo + nryr * dp_ca(9) * Ca_jsr_dev(ii)
       !dtemp2 = nryr * dp_ca(9) + dp_ca(11) + (dp_ca(13) * dp_ca(16) * Vm * dp_arg1 * dp_arg2)  * dp_ca(17) * ndhpr
       !s_cads(tx) = dtemp / dtemp2       
#endif
       dtemp = dp_ca(19) + Ca_jsr_dev(ii)
       dtemp2 = dtemp * dtemp
       dtemp = dtemp2 / (dp_ca(19)*dp_ca(20) + dtemp2)   !beta_jsr
       Ca_jsr_dev(ii) = Ca_jsr_dev(ii) + dt*(dtemp * dp_ca(8)) * (J_refill_dev(ii) - J_ryr)

       CRU_states_dev(ii) =  state_combine_single_ker(lcc_nt, ryr_nt, int_ca(10))
       !CRU_states_dev(ii) = (lcc_nt-1) * int_ca(10) + ryr_nt
       RYRgate_dev(ii) = nryr
       DHPRgate_dev(ii) = ndhpr
       Ca_ds_dev(ii) = s_cads(tx)
    END IF
  END SUBROUTINE update_SFU_nostatechanges
  
#endif !_LCC_MODEL_ == _WILLIAMS2013 || _WILLIAMS2013_NEWALGORITHM
  
  attributes(global) SUBROUTINE analyze_Sparks(time, Ca_nsr)
    USE cudafor
    IMPLICIT NONE
    REAL, VALUE :: time
    REAL, VALUE :: Ca_nsr

    INTEGER :: ii
    REAL :: dtemp1
    
    ! thread Index
    ii =  (blockIdx%x-1) * blockDim%x + threadIdx%x

    IF (ii .LE. int_ca(1)) THEN
       IF (RYRgate_dev(ii) .GE. 1 .AND. flag1(ii) .EQ. .FALSE.) THEN
          ! spark/blink/quark start
          flag1(ii) = .TRUE.; flag2(ii) = .TRUE.; flag3(ii) = .TRUE.;
          ts_spark(ii) = time; ! time start spark/quark
          ts_blink(ii) = time; ! time start blink
       ELSEIF (RYRgate_dev(ii) .EQ. 0 .AND. flag1(ii) .EQ. .TRUE.  & 
            .AND. flag4(ii) .EQ. .FALSE.) THEN
          ! detect quarks
          ! count # of quarks with 1,2,3... opening RYRs
          Nquarks_dev(ii)= Nquarks_dev(ii) + 1 
          ! reset flags
          flag1(ii) = .FALSE.; flag2(ii) = .FALSE.; flag3(ii) = .FALSE.;
       !ELSEIF  (RYRgate_dev(ii) .GE. float(int_ca(2))*0.909d0 .AND.  &
       ELSEIF  (RYRgate_dev(ii) .GE. float(int_ca(2)) .AND.  &
            flag1(ii) .EQ. .TRUE. .AND. flag3(ii) .EQ. .TRUE.) THEN
          ! mature spark
!!$          IF (RYRgate_dev(ii) .LT. prev_RYRgate(ii)) THEN
!!$             ! spark peak (increase spark count)
!!$             Nsparks_dev(ii) = Nsparks_dev(ii) + 1
!!$             SparkPeaks(ii, Nsparks_dev(ii)) = prev_Ca_ds(ii)
!!$             flag3(ii) = .FALSE.; flag4(ii) = .TRUE.;
!!$          ELSE
!!$             prev_Ca_ds(ii) = Ca_ds_dev(ii)
!!$          END IF
          flag3(ii) = .FALSE.
          flag4(ii) = .TRUE.
          Nsparks_dev(ii) = Nsparks_dev(ii) + 1
          tot_SparkPeaks(ii) = tot_SparkPeaks(ii) + Ca_ds_dev(ii)
          tot_SparkTime2Peak(ii) = time - ts_spark(ii)
       ELSE
          IF (RYRgate_dev(ii) .EQ. 0 .AND. flag1(ii) .EQ. .TRUE. .AND. &
            flag4(ii) .EQ. .TRUE.) THEN
             ! spark termination 
             flag1(ii) = .FALSE.; 
             tot_SparkDurations(ii) = tot_SparkDurations(ii) + (time-ts_spark(ii))
          ENDIF
          IF  (Ca_jsr_dev(ii) .GT. 0.99 * Ca_nsr .AND. flag2(ii) .EQ. .TRUE. .AND. &
               flag4(ii) .EQ. .TRUE.)  THEN
             ! blink detection
             flag4(ii) = .FALSE.; flag2(ii) = .FALSE.
             Nblinks_dev(ii) = Nblinks_dev(ii) + 1;
             tot_BlinkDurations(ii) = tot_BlinkDurations(ii) + (time - ts_blink(ii))
          ENDIF
       ENDIF
       !prev_RYRgate(ii) = RYRgate_dev(ii)
    END IF
  END SUBROUTINE analyze_Sparks
  
  SUBROUTINE free_device()
    CALL cleanup_gpudata_ecc()
    DEALLOCATE(Ca_ds_dev, Ca_jsr_dev)
  END SUBROUTINE free_device

  SUBROUTINE cleanup_gpudata_ecc()
    IF (ALLOCATED(lcc_compKv_act_dev)) DEALLOCATE(lcc_compKv_act_dev)
    IF (ALLOCATED(lcc_compKv_inact_dev)) DEALLOCATE(lcc_compKv_inact_dev)
    IF (ALLOCATED(lcc_compKfree_dev)) DEALLOCATE(lcc_compKfree_dev)
    IF (ALLOCATED(ryr_compKfree_dev)) DEALLOCATE(ryr_compKfree_dev)
    IF (ALLOCATED(lcc_compKcal_inact_dev)) DEALLOCATE(lcc_compKcal_inact_dev)
    IF (ALLOCATED(ryr_compKcal_act_dev)) DEALLOCATE(ryr_compKcal_act_dev)
    
    IF (ALLOCATED(lcc_indxKv_act_dev)) DEALLOCATE(lcc_indxKv_act_dev)
    IF (ALLOCATED(lcc_indxKv_inact_dev)) DEALLOCATE(lcc_indxKv_inact_dev)
    IF (ALLOCATED(lcc_indxKfree_dev)) DEALLOCATE(lcc_indxKfree_dev)
    IF (ALLOCATED(ryr_indxKfree_dev)) DEALLOCATE(ryr_indxKfree_dev)
    IF (ALLOCATED(lcc_indxKcal_inact_dev)) DEALLOCATE(lcc_indxKcal_inact_dev)
    IF (ALLOCATED(ryr_indxKcal_act_dev)) DEALLOCATE(ryr_indxKcal_act_dev)
    IF (ALLOCATED(compPdt_dev)) DEALLOCATE(compPdt_dev)
    IF (ALLOCATED(U_Ro_dev)) DEALLOCATE(U_Ro_dev)
    IF (ALLOCATED(U_Lo_dev)) DEALLOCATE(U_Lo_dev)
    IF (ALLOCATED(CRU_states_dev)) DEALLOCATE(CRU_states_dev)
    IF (ALLOCATED(ryrgate_dev)) DEALLOCATE(ryrgate_dev)
    IF (ALLOCATED(dhprgate_dev)) DEALLOCATE(dhprgate_dev)
    IF (ALLOCATED(J_efflux_dev)) DEALLOCATE(J_efflux_dev)
    IF (ALLOCATED(J_refill_dev)) DEALLOCATE(J_refill_dev)
    IF (ALLOCATED(J_efflux_total)) DEALLOCATE(J_efflux_total)
    IF (ALLOCATED(J_refill_total)) DEALLOCATE(J_refill_total)
    IF (ALLOCATED(I_dhpr_total)) DEALLOCATE(I_dhpr_total)
    IF (ALLOCATED(I_dhpr_dev)) DEALLOCATE(I_dhpr_dev)
  END SUBROUTINE cleanup_gpudata_ecc

  SUBROUTINE data_io_init()
    LOGICAL :: is_exist
    INTEGER :: ii

#if _DATAMODE_ == _IO_TEST
    CALL datamode1_init()
#elif _DATAMODE_ == _IO_2_HDF
    CALL datamode2_init()
#elif _DATAMODE_ == _IO_2_ASCII
    CALL datamode3_init()
#elif _DATAMODE_ == _IO_2_SILO_3D
    CALL datamode4_init()
#elif _DATAMODE_ == _IO_2_SILO_SLICE
    CALL datamode5_init()
#elif _DATAMODE_ == _IO_NONE
    WRITE(log_string,*) "No I/O"
    CALL write2log(log_string)
#elif _DATAMODE_ == _IO_2_HDF2
    CALL datamode7_init()
#endif  !DATAMODE
    
  END SUBROUTINE data_io_init

#if _DATAMODE_ == _IO_TEST
  SUBROUTINE datamode1_init()
      
    WRITE(log_string,*) "Mode 1"
    CALL write2log(log_string)
    
    fstring = TRIM(where2save) // "/ca_ds.dat"
    OPEN(unit=11, file=TRIM(fstring))
    WRITE(log_string,"(A,A)") "calcium diadic subspace: ", TRIM(fstring)
    CALL write2log(log_string)

    OPEN(unit=12, file=TRIM(where2save) // "/new.dat")
    WRITE(log_string,*) "ca_jsr.dat: calcium in junctional SR"
    CALL write2log(log_string)
    
    OPEN(unit=13, file=TRIM(where2save) // "/nRo.dat")
    WRITE(log_string,*) "nRo.dat: number of RYR open"
    CALL write2log(log_string)

    OPEN(unit=14, file=TRIM(where2save) // "/nLo.dat")      
    WRITE(log_string,*) "nLo.dat: number of DHPR open"
    CALL write2log(log_string)
    
    OPEN(unit=15, file=TRIM(where2save) // "/Vm.dat")    
    OPEN(unit=16, file=TRIM(where2save) // "/Cmyo.dat")
    OPEN(unit=17, file=TRIM(where2save) // "/Jryr.dat")
    OPEN(unit=18, file=TRIM(where2save) // "/Idhprtotal.dat")
    OPEN(unit=21, file=TRIM(where2save) // "/I_na.dat")
  END SUBROUTINE datamode1_init
  !! !!!!!!!!!!!!!!!!!!!!!!!!!!
  SUBROUTINE datamode1_print()
    INTEGER :: jj
    CHARACTER(100) :: formatstring

    WRITE(formatstring, "(i0)") NSFU+1
    formatstring = "(" // TRIM(formatstring) //  "(2x,1pe12.5))"
    
    ! write out the last 2 second only
   ! IF (time .GE. t_end - 2.0d0) THEN          
       !! print ca_ds
    !     J_ryr_total = J_ryr_dev
       Ca_ds = Ca_ds_dev
       Ca_jsr = Ca_jsr_dev
       RYRgate = Ryrgate_dev
       DHPRgate = DHPRgate_dev
       WRITE(11,TRIM(formatstring)) time,(Ca_ds(jj), jj=1,NSFU)
       WRITE(12,TRIM(formatstring)) time,(new)!, jj=1,NSFU)
       WRITE(13,TRIM(formatstring)) time,(RYRgate(jj), jj=1,NSFU)
       WRITE(14,TRIM(formatstring)) time,(DHPRgate(jj), jj=1,NSFU)
       WRITE(15,TRIM(formatstring)) time, Vm
       WRITE(16,TRIM(formatstring)) time, Ca_myo
  !!    WRITE(17,TRIM(formatstring)) time,(J_ryr_total(jj),jj=1,NSFU)  ! (J_ryr(jj),jj=1,NSFU)
      WRITE(18,TRIM(formatstring)) time, I_dhpr_total(1)             
   ! END IF
    !Ca_ds = CaCM_ds_dev
    WRITE(21,TRIM(formatstring)) time,I_Na   !(Ca_ds(jj), jj=1,NSFU)
    !Ca_ds = Ca_ds_dev
    !WRITE(22,TRIM(formatstring)) time,(Ca_ds(jj), jj=1,NSFU)
    !WRITE(22,TRIM(formatstring)) dt
    CRU_states = CRU_states_dev
    CALL state_breakdown_all (CRU_states, irow_R, LCCcluster_states, RYRcluster_states)
    WRITE(formatstring, "(i0)") NSFU+1
    formatstring = "(1pe12.5,(" // TRIM(formatstring) //  "(2x,I5)))"
    WRITE(22, TRIM(formatstring)) dt, LCCcluster_states
    CALL FLUSH()
    !PRINT *, "TUAN"
    
  END SUBROUTINE datamode1_print
  SUBROUTINE datamode1_close()
    ! close output files
    CLOSE(11)
    CLOSE(12)
    CLOSE(13)
    CLOSE(14)
    CLOSE(15)
    CLOSE(16)
    CLOSE(17)
    CLOSE(18)
  END SUBROUTINE datamode1_close
#endif
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
#if _DATAMODE_ == _IO_2_HDF
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  SUBROUTINE datamode2_init()
    IMPLICIT NONE
    LOGICAL :: is_exist
    INTEGER :: hdf_error
    
    WRITE(log_string,*) "Mode 2 (HDF)"
    CALL write2log(log_string)

    hdf_error =  hdf_init()
    CHUNK_SIZE = TRACK_LENGTH
    CHUNKS_PER_FILE = 1 ! set it to one for now
    NUM_STATES_NJ = mR * mL
    NUM_FIELDS4CURRENTS = 11
    !4 columns = 1 for time + 1 for Vm + 1 for Camyo + 1 for Cansr
    !next (NUM_FIELDS4CURRENTS) = ionic current gating parameters
    !next (NUM_STATES_NJ) = p_nj
    FIELDS_PER_RECORD = FIELDS_PER_RECORD + 4 + NUM_STATES_NJ + NUM_FIELDS4CURRENTS
    ! + 2 = CaF, J_ryr_orphan
    FIELDS_PER_RECORD = FIELDS_PER_RECORD + 2
    !first NSFU = CRU_states
    !second NSFU = cds
    !third NSFU = cjsr
    FIELDS_PER_RECORD = FIELDS_PER_RECORD + 3*NSFU
    
    records_used = 0
    file_index = 0
    ALLOCATE(data_buffer(FIELDS_PER_RECORD, CHUNK_SIZE))
    data_buffer = 0.0d0
    !remove existing files
    sys_command = "rm " // TRIM(where2save) // "/*.h5"
    CALL system(sys_command)
  END SUBROUTINE datamode2_init
  SUBROUTINE datamode2_close()
    INTEGER :: hdf_error
    
    hdf_error = hdf_close()
  END SUBROUTINE datamode2_close
  !! !!!!!!!!!!!!!!!!!!!!!!!!!!!
  SUBROUTINE datamode2_print()
    INTEGER :: offset
    INTEGER :: ii
    ! track and output data
    IF (records_used .EQ. CHUNK_SIZE) THEN
       file_index = file_index + 1
       WRITE(file_out, "(a,'/data-', i3.3,'.h5')") TRIM(where2save), file_index         
       CALL hdf_open_simple(file_out, NSFU, CHUNK_SIZE, CHUNKS_PER_FILE, file_struct, FIELDS_PER_RECORD)
       CALL hdf_write_simple(file_struct, data_buffer, CHUNK_SIZE, FIELDS_PER_RECORD)
       CALL hdf_spark_close(file_struct)
       records_used = 0
    ENDIF
    records_used = records_used + 1
    data_buffer(1, records_used) = time
    data_buffer(2, records_used) = Vm
    data_buffer(3, records_used) = Ca_myo
    data_buffer(4, records_used) = Ca_nsr
    offset = 4
    data_buffer(offset+1:offset+NSFU, records_used) = CRU_states_dev
    offset = offset + NSFU
    data_buffer(offset+1:offset+NSFU, records_used) =  Ca_ds_dev
    offset = offset + NSFU
    data_buffer(offset+1:offset+NSFU, records_used) = Ca_jsr_dev
    offset = offset + NSFU
  #if _LCC_MODEL_ == _WILLIAMS2013
    DO ii = 1, mL
       p_nj((ii-1)*mR+1:ii*mR) = pL_nj(ii) * pR_nj(1:mR)
    END DO
    data_buffer(offset+1:offset+NUM_STATES_NJ, records_used) = p_nj
  #endif
    offset = offset + NUM_STATES_NJ
    data_buffer(offset+1:offset+NUM_FIELDS4CURRENTS, records_used) = (/ Na_i, K_i, m_na, h_na, j_na, a_Ktof, in_Ktof, a_Ktos, in_Ktos, a_Kss, in_Kss /)
    
    offset = offset + NUM_FIELDS4CURRENTS
    data_buffer(offset+1,records_used) = CaF
    data_buffer(offset+2,records_used) = J_ryr_orphan
  END SUBROUTINE datamode2_print
  !! !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  SUBROUTINE datamode2_print_last()
    file_index = file_index + 1
    WRITE(file_out, "(a,'/data-', i3.3,'.h5')") TRIM(where2save), file_index
    CALL hdf_open_simple(file_out, NSFU, records_used, CHUNKS_PER_FILE, file_struct, FIELDS_PER_RECORD)
    CALL hdf_write_simple(file_struct, data_buffer, records_used, FIELDS_PER_RECORD)
    CALL hdf_spark_close(file_struct)
  END SUBROUTINE datamode2_print_last
  
  SUBROUTINE datamode2_print_fb()
    !print out the first-beat
    INTEGER :: offset
    INTEGER :: ii
    ! track and output data
    IF (records_used .EQ. CHUNK_SIZE) THEN
       file_index = file_index + 1
       WRITE(file_out, "(a,'/data-', i3.3,'.h5')") TRIM(where2save)//'/'//TRIM(first_beat), file_index         
       CALL hdf_open_simple(file_out, NSFU, CHUNK_SIZE, CHUNKS_PER_FILE, file_struct, FIELDS_PER_RECORD)
       CALL hdf_write_simple(file_struct, data_buffer, CHUNK_SIZE, FIELDS_PER_RECORD)
       CALL hdf_spark_close(file_struct)
       records_used = 0
    ENDIF
    records_used = records_used + 1
    data_buffer(1, records_used) = time
    data_buffer(2, records_used) = Vm
    data_buffer(3, records_used) = Ca_myo
    data_buffer(4, records_used) = Ca_nsr
    offset = 4
    DO ii = 1, mL
       p_nj((ii-1)*mR+1:ii*mR) = pL_nj(ii) * pR_nj(1:mR)
    END DO
    data_buffer(offset+1:offset+NUM_STATES_NJ, records_used) = p_nj
    offset = offset + NUM_STATES_NJ
    data_buffer(offset+1:offset+NUM_FIELDS4CURRENTS, records_used) = (/ Na_i, K_i, m_na, h_na, j_na, a_Ktof, in_Ktof, a_Ktos, in_Ktos, a_Kss, in_Kss /)

    offset = offset + NUM_FIELDS4CURRENTS
    data_buffer(offset+1,records_used) = CaF
    data_buffer(offset+2,records_used) = J_ryr_orphan
    offset = offset + 2    
    data_buffer(offset+1:offset+NSFU, records_used) = CRU_states_dev
    offset = offset + NSFU
    data_buffer(offset+1:offset+NSFU, records_used) =  Ca_ds_dev
    offset = offset + NSFU
    data_buffer(offset+1:offset+NSFU, records_used) = Ca_jsr_dev
    offset = offset + NSFU
  END SUBROUTINE datamode2_print_fb
  SUBROUTINE datamode2_print_fb_last()
    file_index = file_index + 1
    WRITE(file_out, "(a,'/data-', i3.3,'.h5')") TRIM(where2save)//'/'//TRIM(first_beat), file_index
    CALL hdf_open_simple(file_out, NSFU, records_used, CHUNKS_PER_FILE, file_struct, FIELDS_PER_RECORD)
    CALL hdf_write_simple(file_struct, data_buffer, records_used, FIELDS_PER_RECORD)
    CALL hdf_spark_close(file_struct)
    file_index =0
    records_used = 0
  END SUBROUTINE datamode2_print_fb_last
#endif
    
  !! !!!!!!!!!!!!!!!!!!!!!!!!!!!
#if _DATAMODE_ == _IO_2_ASCII
  SUBROUTINE datamode3_init()
    IMPLICIT NONE
    LOGICAL :: is_exist
    INTEGER :: offset
      
    WRITE(log_string,*) "Mode 3"
    CALL write2log(log_string)
      
    INQUIRE(FILE = TRIM(where2save), EXIST = is_exist)
    IF (.NOT. is_exist) THEN 
       sys_command = 'mkdir ' // TRIM(where2save) 
       CALL system(sys_command)
    END IF
    WRITE(log_string,*) "location: ", TRIM(where2save)
    CALL write2log(log_string)

    records_used = 0
    file_index = 0
    offset = 5 ! should be the value the same as the one being used in datamode3_print
    NUM_STATES_NJ = mR*mL
    NUM_FIELDS4CURRENTS = 11
    FIELDS_PER_RECORD = offset + 3*NSFU + NUM_STATES_NJ + NUM_FIELDS4CURRENTS
    ALLOCATE(data_buffer(FIELDS_PER_RECORD, TRACK_LENGTH) )
    data_buffer = 0.0d0

    WRITE(fstring, "(i0a)") offset, PRECISION_OUTPUT
    WRITE(fstring, "(a,',',i0a)") TRIM(fstring), 3*NSFU, PRECISION_OUTPUT
    WRITE(fstring, "(a,',',i0a)") TRIM(fstring), NUM_STATES_NJ, PRECISION_OUTPUT
    WRITE(fstring, "(a,',',i0a)") TRIM(fstring), NUM_FIELDS4CURRENTS, PRECISION_OUTPUT
      
    fstring = "(" // TRIM(fstring) // ")"
  END SUBROUTINE datamode3_init
  !! !!!!!!!!!!!!!!!!!!!!!!!!!!!!
  SUBROUTINE datamode3_close()
    CLOSE(unit=11)
  END SUBROUTINE datamode3_close
  !! !!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  SUBROUTINE datamode3_print()
    INTEGER :: offset
    INTEGER :: fileID
    
    fileID = get_file_ID()
    IF (records_used .EQ. TRACK_LENGTH) THEN
       file_index = file_index + 1
       WRITE(file_out, "(a,'/data-', i3.3,'.txt')") TRIM(where2save), file_index
       OPEN(unit=fileID, file=TRIM(file_out))
       DO ii = 1, TRACK_LENGTH
          WRITE(fileID,TRIM(fstring)) data_buffer(:, ii)
       END DO
       CLOSE(unit = fileID)
       records_used = 0
    END IF
    records_used = records_used + 1
    data_buffer(1, records_used) = time
    data_buffer(2, records_used) = Vm
    data_buffer(3, records_used) = I_dhpr_total(1)
    data_buffer(4, records_used) = Ca_myo
    data_buffer(5, records_used) = Ca_nsr
    offset = 5
    data_buffer(offset+1:offset+NSFU, records_used) = CRU_states_dev
    offset = offset + NSFU
    data_buffer(offset+1:offset+NSFU, records_used) =  Ca_ds_dev
    offset = offset + NSFU
    data_buffer(offset+1:offset+NSFU, records_used) = Ca_jsr_dev
    offset = offset + NSFU
    DO ii = 1, mL
       p_nj((ii-1)*mR+1:ii*mR) = pL_nj(ii) * pR_nj(1:mR)
    END DO
    data_buffer(offset+1:offset+NUM_STATES_NJ, records_used) = p_nj
    offset = offset + NUM_STATES_NJ
    data_buffer(offset+1:offset+NUM_FIELDS4CURRENTS, records_used) = (/ Na_i, K_i, m_na, h_na, j_na, a_Ktof, in_Ktof, a_Ktos, in_Ktos, a_Kss, in_Kss /)
    offset = offset + NUM_FIELDS4CURRENTS
  END SUBROUTINE datamode3_print
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  SUBROUTINE datamode3_print_last()
    INTEGER :: fileID
    
    fileID = get_file_ID()
    
    file_index = file_index + 1
    WRITE(file_out, "(a,'/data-', i3.3,'.txt')") TRIM(where2save), file_index
    OPEN(unit=fileID, file=TRIM(file_out))
    
    DO ii = 1, records_used
       WRITE(fileID,TRIM(fstring)) data_buffer(:, ii)
    END DO

    CLOSE(unit = fileID)
    
  END SUBROUTINE datamode3_print_last

#endif
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  SUBROUTINE data_io_write()
    !WRITE INITIAL DATA OUT
#if _DATAMODE_ == _IO_TEST
    CALL datamode1_print()
#elif _DATAMODE_ == _IO_2_HDF
    CALL datamode2_print()
#elif _DATAMODE_ == _IO_2_ASCII
    CALL datamode3_print()
#elif _DATAMODE_ == _IO_2_HDF2
    CALL datamode7_print()
#endif
  END SUBROUTINE data_io_write

  SUBROUTINE data_io_finalize()
    INCLUDE "lib3f.h"
    INTEGER :: ierr, ii, fileID
      
    ! writing the last time
#if _DATAMODE_ == _IO_TEST
    CALL datamode1_close()
#elif _DATAMODE_ == _IO_2_HDF
    CALL datamode2_print_last()
    CALL datamode2_close()
#elif  _DATAMODE_ == _IO_2_ASCII
    CALL datamode3_print_last()  ! print the last time
    CALL datamode3_close()
      !   CALL datamode3_finalize()
!!$#elif _DATAMODE_ == 3
!!$    CLOSE(11)
!!$    sys_command = 'gzip -c ' // trim(filename) // ' > ' // trim(filename) // '.gz'
!!$    CALL system(sys_command)
!!$    sys_command = 'rm ' // trim(filename)
!!$    CALL system(sys_command)
#elif _DATAMODE_ == _IO_2_HDF2
    CALL datamode7_print_last()
    CALL datamode7_close()      
#endif
    IF (write_restartfile .NE. 0) THEN
       CALL write2restartfile()
    ENDIF
            
    !!send email message
!    ierr = hostnm(hostname)
!    sys_command = 'echo "' // TRIM(full_command) // ' \n run on machine ' //  TRIM(hostname) // ' complete" | mail -s "PROGRAM COMPLETED" hoangtrongminhtuan@gmail.com'
!    ierr = system(TRIM(sys_command))
!
  END SUBROUTINE data_io_finalize

  !!>****f*  ecc_gpu_utility/init_Fluo
  !! SYNOPSIS  
  !!   Generate necessary information for fluorescence data
  !!   e.g. equilibrium concentrations + rate constants
  !! INPUTS 
  !! OUTPUTS 
  !! RETURN 
  !! SIDE EFFECTS 
  !! HISTORY 
  !! BUGS 
  !! TODO 
  !! SEE ALSO 
  !!_****
  SUBROUTINE init_Fluo()
    REAL(KIND=dp) :: Kd
    
    WRITE (log_string, *) "start init_Fluo()"
    CALL write2log(log_string)    
    PRINT *, log_string
    IF (ABS(kon_F) .GE. EPS) THEN 
       Kd = koff_F / kon_F ! uM
    ELSE
       Kd = 1d12   !! very low affinity
    ENDIF

    !!F_T = read from parameter file
    CaF_infty =  F_T * Ca_myo / (Kd + Ca_myo)
    
    
#if _USE_FLUO_ == _YES    
    WRITE (log_string, *) "Use Dyes (Fluo-?)" 
    CALL write2log(log_string)
    WRITE(log_string, *) "[CaF]_inf = F0 = ", CaF_infty
    CALL write2log(log_string)
    WRITE(log_string, *) "[F_myo]_0 = ", F_T - CaF_infty   ! equilibrium F_myo
    CALL write2log(log_string)
    WRITE(log_string, *) "Kd_F = ", Kd
    CALL write2log(log_string)
#else
    WRITE (log_string, *) "No Dyes (Fluo)" 
    CALL write2log(log_string)
#endif
    
    WRITE (log_string, *) "++end init_Fluo()"
    CALL write2log(log_string)
    PRINT *, log_string
  END SUBROUTINE init_Fluo
  
  SUBROUTINE loaddata_from_conf()
    IMPLICIT NONE

    CALL init_Fluo()
    
    ! Concentration parameters + initial condition
    Ca_myo = Ca_myo_initial
    Ca_nsr = Ca_nsr_initial
    Ca_ds = Ca_myo
    Ca_jsr = Ca_nsr
    CaF = CaF_infty
    
    ! init data use for EC-coupling model
    CALL init_data_ECC()

    ! init data
    CaB_htrpn = (Ca_myo * B_htrpnT) / ( koff_htrpn/kon_htrpn + Ca_myo)
    CaCM = (Ca_myo * CM_T) / (koff_CM/kon_CM + Ca_myo)
    CaSL = (Ca_myo * B_SL_T) / (koff_SL/kon_SL + Ca_myo)
    CaSR = (Ca_myo * B_SR_T) / (koff_SR/kon_SR + Ca_myo)

!!$    Kd = koff_CM/kon_CM
!!$    CaCM_infinity = CM_T * Ca_myo / (Kd + Ca_Myo)
!!$    Kd = koff_SL/kon_SL
!!$    CaSL_infinity = B_SL_T * Ca_myo / (Kd + Ca_Myo)
!!$    Kd = koff_SR/kon_SR
!!$    CaSR_infinity = B_SR_T * Ca_myo / (Kd + Ca_Myo)
!!$    CaCM = CaCM_infinity
!!$    CaSL = CaSL_infinity
!!$    CaSR = CaSR_infinity
    
#if _TEST_ORPHAN_CRU_ == _YES
    ALLOCATE(pR_orphan(irow_R_orphan), tmp_pR_orphan(irow_R_orphan))
    pR_orphan = 0d0
    ALLOCATE(chiC_orphan(irow_R_orphan), chiO_orphan(irow_R_orphan))
#endif
  END SUBROUTINE loaddata_from_conf

  !! we need
  !!  - gating variables
  !!  - Camyo, Cansr, Cads, Cajsr, Na_i, K_i...
  !!  - CRU_states
  SUBROUTINE loaddata_from_restartfile()
    IMPLICIT NONE
    LOGICAL :: is_exist
    INTEGER :: offset, ii, fileID
    CHARACTER(100) :: formatstring
    
    WRITE(log_string, *) "Running simulation using input from restart file"
    CALL write2log(log_string)
    ! check if restart file existed
    INQUIRE(FILE = TRIM(restart_file), EXIST = is_exist)
    IF (.NOT. is_exist) THEN
       WRITE(log_string, *) "Restart " // TRIM(restart_file) // " file not found"
       CALL write2log(log_string)
       STOP
    END IF

    ! check if restart file data fit wit conf. data file
    NUM_FIELDS4CURRENTS = 11
    NUM_INDIVIDUAL_FIELDS = 7+NUM_APD_TYPES
    
#if _TEST_ORPHAN_CRU_ == _YES
    RF_FIELDS_PER_RECORD = 3*FULL_NSFU + mL + mR + NUM_FIELDS4CURRENTS + NUM_INDIVIDUAL_FIELDS
#else
    RF_FIELDS_PER_RECORD = 3*NSFU + mL + mR + NUM_FIELDS4CURRENTS + NUM_INDIVIDUAL_FIELDS
#endif

    ALLOCATE(data_buffer(RF_FIELDS_PER_RECORD,1))
    data_buffer = 0.0d0
    
    WRITE(formatstring, "(i0)") RF_FIELDS_PER_RECORD  
    formatstring = "(" // TRIM(formatstring) //  "(2x,1pe21.13))"

    fileID = get_file_ID()
    OPEN (unit = fileID, file = TRIM(restart_file))
    READ (fileID, TRIM(formatstring)) data_buffer
    CLOSE (fileID)

#if _TEST_ORPHAN_CRU_ == _YES
    ! a fraction of CRU becomes orphan, i.e. LCC becomes non-junctional
    ! and RYR clusters see Ca_myo, not Ca_ds
    IF (TRIM(restart_file) .EQ. "") THEN
       WRITE (6, *) "Orphan mode only work with restart file, for now"
       WRITE (6, *) "and requires 20,000 CaRUs in control simul"
       STOP
    END IF

    offset = 0
    CRU_states = data_buffer(offset+1:offset+NSFU,1)
    DO ii = 1, irow_R_orphan
       pR_orphan(ii) = COUNT(data_buffer(4+NSFU+1:4+FULL_NSFU,1) .EQ. ii) 
    END DO

    offset = offset + FULL_NSFU
    Ca_ds = data_buffer(offset+1:offset+NSFU,1)
    offset = offset+ FULL_NSFU
    Ca_jsr = data_buffer(offset+1:offset+NSFU,1)
    offset = offset + FULL_NSFU
    pL_nj = data_buffer(offset+1:offset+mL,1)
    offset = offset + mL
    pR_nj = data_buffer(offset+1:offset+mR,1)
    offset = offset + mR
               
#else
    ! set up data
    offset = 0
    CRU_states = data_buffer(offset+1:offset+NSFU,1)
    offset = offset + NSFU
    Ca_ds = data_buffer(offset+1:offset+NSFU,1)
    offset = offset+NSFU
    Ca_jsr = data_buffer(offset+1:offset+NSFU,1)
    offset = offset + NSFU
    pL_nj = data_buffer(offset+1:offset+mL,1)
    offset = offset + mL
    pR_nj = data_buffer(offset+1:offset+mR,1)
    offset = offset + mR
#endif
      
    Na_i = data_buffer(offset + 1, 1)
    K_i = data_buffer(offset + 2,1)
    m_na = data_buffer(offset + 3,1)
    h_na = data_buffer(offset + 4,1)
    j_na = data_buffer(offset + 5,1)
    a_Ktof = data_buffer(offset + 6,1)
    in_Ktof = data_buffer(offset + 7,1)
    a_Ktos = data_buffer(offset + 8,1)
    in_Ktos = data_buffer(offset + 9,1)
    a_Kss = data_buffer(offset + 10,1)
    in_Kss = data_buffer(offset + 11,1)
    offset = offset + NUM_FIELDS4CURRENTS

    Vm = data_buffer(offset+1,1)
    Ca_myo = data_buffer(offset+2,1)
    Ca_nsr = data_buffer(offset+3,1)
    is_being_clamped = data_buffer(offset + 4,1)
    CaB_htrpn = data_buffer(offset + 5, 1)
    CaF = data_buffer(offset + 6, 1)
    is_inside_AP = data_buffer(offset + 7:offset+6+NUM_APD_TYPES, 1)
    offset = offset + 6 + NUM_APD_TYPES
    real_timestart = data_buffer(offset + 1, 1)
    offset =  offset + 1
    CaCM = data_buffer(offset+1,1)
    CaSL = data_buffer(offset+2,1)
    CaSR = data_buffer(offset+3,1)
    offset = offset + 3

  END SUBROUTINE loaddata_from_restartfile

  !!>****f* 
  !! SYNOPSIS  
  !!   Setup kernel configuration 
  !! INPUTS 
  !! OUTPUTS 
  !! RETURN 
  !! SIDE EFFECTS 
  !! HISTORY 
  !! BUGS 
  !! TODO 
  !! SEE ALSO 
  !!_****
  FUNCTION setup_KernelConfiguration() RESULT(ierr)
    IMPLICIT NONE
    INTEGER :: ierr
    !local
    INTEGER :: err
    
    ierr = SUCCESS
    ! create cuda stream
    err = cudaStreamCreate(stream0)
    IF (err .NE. cudaSuccess) THEN
       WRITE (error_message,*) "ERROR: Create cuda stream"
       ierr = ERR_CUDA_CREATE_STREAM
       RETURN
    ENDIF
    err = cudaStreamCreate(stream1)
    IF (err .NE. cudaSuccess) THEN
       WRITE (error_message,*) "ERROR: Create cuda stream"
       ierr = ERR_CUDA_CREATE_STREAM
       RETURN
    ENDIF

    !for 1D thread BLOCK
    blocksize = BLOCKSIZE_1D
    !gridSize = INT((NSFU+blocksize-1)/blocksize)  ! ceiling value
    gridSize = CEILING(REAL(NSFU)/blocksize)  ! ceiling value
    RETURN
  END FUNCTION  setup_KernelConfiguration

  
  SUBROUTINE init_PRNG()
    IMPLICIT NONE
    INTEGER :: global_seed
    REAL(KIND=dp) :: tempsec, tempsec2

    !choose PRNG
#if _USE_PRNG_ == _PRNG_SARU
    WRITE(6,*) "Use SARU"
    ! initialize the pseudo-RNG
    IF (fixed_seed .EQ. 1) THEN
       global_seed = rand_seed
       WRITE(log_string,*) " -- use fixed-seed: ", global_seed
       CALL write2log(log_string)
    ELSE
       tempsec = 372.d0*SECNDS(0.0)-1.d0
       tempsec2 = 2.d0*SECNDS(0.0)-1.d0
       global_seed = INT(tempsec) + MOD(INT(tempsec2),127)*975121
       WRITE(log_string,*) "use random-seed: ", global_seed
       CALL write2log(log_string)
    END IF
    saruObj = init_saru(NSFU, global_seed)
#elif _USE_PRNG_ == _PRNG_MT
    WRITE(log_string,*) "Use MT GPU"
    CALL write2log(log_string)
    d_status = init_mt_status()
    d_texture = init_mt_texture(d_status)
!!$    num_data = 1000
!!$    res = cudaMalloc(d_data, num_data)
#elif _USE_PRNG_ == _PRNG_MTCPU
    WRITE(log_string,*) "Use MT CPU"
    CALL write2log(log_string)
    IF (fixed_seed .EQ. 1) THEN
       CALL init_rndgen_testonly(mt_STATE, INFO, rand_seed)
    ELSE
       CALL init_rndgen(mt_STATE, INFO)
    END IF
    ALLOCATE(X_r(NSFU*copy_interval))
    mt_LB = 0.0
    mt_UB = 1.0
#endif

  END SUBROUTINE init_PRNG

  SUBROUTINE save_summarized_data_1(file2save, time_duration)
    !!automatically closing ID
    CHARACTER(*) :: file2save    
    REAL(KIND=dp) :: time_duration
    !!local
    INTEGER :: fileID

    fileID = get_file_ID()
    !! write summary data to file
    OPEN(unit=fileID, file=TRIM(file2save))
    
    CALL save_summarized_data_core(fileID, time_duration)
    CLOSE(fileID) 
  END SUBROUTINE save_summarized_data_1
  
  SUBROUTINE save_summarized_data_2(fileID, time_duration)
    INTEGER :: fileID
    REAL(KIND=dp) :: time_duration
    CALL save_summarized_data_core(fileID, time_duration)
    !!not doing the job of closing ID
  END SUBROUTINE save_summarized_data_2
  
  SUBROUTINE save_summarized_data_core(fileID, time_duration)
    !! called by spark_summarize_fb() and spark_summarize()
    IMPLICIT NONE
    INTEGER :: fileID
    REAL(KIND=dp) :: time_duration
    !!local
    INTEGER :: ii
    REAL(KIND=dp), DIMENSION(NSFU) :: tmp
    REAL(KIND=dp) :: dvalue

    count_summarydata = 0
    
#if _RUNMODE_ == _MODE_V_CLAMP || _RUNMODE_ == _MODE_I_CLAMP
    IF (accumulate_J_dhpr .LE. EPS) THEN
       INTEGRAL_GAIN = 0.0d0
       PEAK_GAIN = 0.0d0
    ELSE
       INTEGRAL_GAIN = accumulate_J_ryr / accumulate_J_dhpr
       PEAK_GAIN = peak_J_SRrelease_total/peak_J_dhpr_total
    ENDIF
    DO ii = 1, NUM_APD_TYPES
       IF (APcount(ii) .EQ. 0) THEN
          APD(ii) = 0
       ELSE
          APD(ii) = APD(ii)/ APcount(ii)
       ENDIF
    END DO
#else
    INTEGRAL_GAIN = 0.0d0
    PEAK_GAIN = 0.0d0
    APD = 0
#endif
    summary_data(1) = INTEGRAL_GAIN
    description_summarydata(1) = 'INTEGRAL_GAIN'
    summary_data(2) = PEAK_GAIN
    description_summarydata(2) = 'PEAK_GAIN'
    count_summarydata = count_summarydata + 2
    
    summary_data(count_summarydata+1:count_summarydata+NUM_APD_TYPES) = APD
    description_summarydata(count_summarydata+1:count_summarydata+NUM_APD_TYPES) = (/ "APD90   ", "APD50", 'APD20', "APD0mV" /)
    count_summarydata = count_summarydata + NUM_APD_TYPES
    
    summary_data(count_summarydata+1) = t_start 
    summary_data(count_summarydata+2) = t_end
    description_summarydata(count_summarydata+1) = "time_start"
    description_summarydata(count_summarydata+2) = "time_end"
    count_summarydata = count_summarydata + 2
    
    summary_data(count_summarydata+1) = (accumulate_J_ryr-accumulate_J_ryr_orphan)/time_duration * Vol_ref/V_myo_T
    summary_data(count_summarydata+2) = accumulate_J_ryr_orphan/time_duration * Vol_ref/V_myo_T  
    description_summarydata(count_summarydata+1) = "mass_Ca_release_normal_per_sec"
    description_summarydata(count_summarydata+2) = "mass_Ca_release_orphan_per_sec"
    count_summarydata = count_summarydata + 2

    Nsparks = Nsparks_dev
    Nblinks = Nblinks_dev
    Nquarks = Nquarks_dev
#if _TEST_NEW_ORPHAN_CRU_ == _YES    
    num_Sparks = SUM(Nsparks(1:NSFU_normal))
    num_Blinks = SUM(Nblinks(1:NSFU_normal))
    num_Quarks = SUM(Nquarks(1:NSFU_normal))
    IF (num_Sparks .EQ. 0)  num_Sparks = 1
    IF (num_Blinks .EQ. 0) num_Blinks = 1
    IF (num_Quarks .EQ. 0) num_Quarks = 1

    tmp = tot_SparkDurations
    avg_SparkDuration = SUM(tmp(1:NSFU_normal))/num_Sparks
    tmp = tot_BlinkDurations
    avg_BlinkDuration = SUM(tmp(1:NSFU_normal))/num_Blinks
    tmp = tot_SparkPeaks
    avg_SparkPeak = SUM(tmp(1:NSFU_normal))/num_Sparks

#else    
    num_Sparks = SUM(Nsparks)
    num_Blinks = SUM(Nblinks)
    num_Quarks = SUM(Nquarks)
    IF (num_Sparks .EQ. 0)  num_Sparks = 1
    IF (num_Blinks .EQ. 0) num_Blinks = 1
    IF (num_Quarks .EQ. 0) num_Quarks = 1

    tmp = tot_SparkDurations
    avg_SparkDuration = SUM(tmp)/num_Sparks
    tmp = tot_BlinkDurations
    avg_BlinkDuration = SUM(tmp)/num_Blinks
    tmp = tot_SparkPeaks
    avg_SparkPeak = SUM(tmp)/num_Sparks
#endif

    IF (num_Quarks+num_Sparks .GT. 0) spark_Freq = float(num_Sparks)/(num_Quarks+num_Sparks)

    summary_data(count_summarydata+1) = num_Sparks/time_duration * (float(MAX_CaRUs)/NSFU)
    summary_data(count_summarydata+2) = num_Blinks/time_duration * (float(MAX_CaRUs)/NSFU)
    summary_data(count_summarydata+3) = num_Quarks/time_duration * (float(MAX_CaRUs)/NSFU)
    summary_data(count_summarydata+4) = avg_SparkDuration
    summary_data(count_summarydata+5) = avg_BlinkDuration
    summary_data(count_summarydata+6) = avg_SparkPeak
    description_summarydata(count_summarydata+1) = "num_DiastolicSparks_per_sec"
    description_summarydata(count_summarydata+2) = "num_DiastolicBlinks_per_sec"
    description_summarydata(count_summarydata+3) = "num_DiastolicQuarks_per_sec"
    description_summarydata(count_summarydata+4) = "avg_SparksDuration"
    description_summarydata(count_summarydata+5) = "avg_BlinkDuration"
    description_summarydata(count_summarydata+6) = "avg_SparkPeak"
    count_summarydata = count_summarydata + 6

    summary_data(count_summarydata+1) = spark_Freq
!!$    summary_data(count_summarydata+2) = SFWHM
!!$    summary_data(count_summarydata+3) = BFWHM
    description_summarydata(count_summarydata+1) = "Spark_Freq"
!!$    description_summarydata(count_summarydata+2) = "Spark_FWHM"
!!$    description_summarydata(count_summarydata+3) = "Blink_FWHM"
!!$    count_summarydata = count_summarydata + 3
    count_summarydata = count_summarydata + 1

    summary_data(count_summarydata+1) = time_start_DiastolicPhase
    description_summarydata(count_summarydata+1) = "time_start_DiastolicPhase"
    count_summarydata = count_summarydata + 1
    
#if _TEST_NEW_ORPHAN_CRU_ == _YES
    num_Sparks = SUM(Nsparks(NSFU_normal+1:))
    num_Blinks = SUM(Nblinks(NSFU_normal+1:))
    num_Quarks = SUM(Nquarks(NSFU_normal:))
    avg_SparkDuration = SUM(tot_SparkDurations(NSFU_normal+1:))/num_Sparks
    avg_BlinkDuration = SUM(tot_BlinkDurations(NSFU_normal+1:))/num_Blinks
    avg_SparkPeak = SUM(tot_SparkPeaks(NSFU_normal+1:))/num_Sparks
    
    IF (num_Quarks+num_Sparks .GT. 0) spark_Freq = float(num_Sparks)/(num_Quarks+num_Sparks)
    
    summary_data(count_summarydata+1) = num_Sparks/time_duration * (float(MAX_CaRUs)/NSFU)
    summary_data(count_summarydata+2) = num_Blinks/time_duration * (float(MAX_CaRUs)/NSFU)
    summary_data(count_summarydata+3) = num_Quarks/time_duration * (float(MAX_CaRUs)/NSFU)
    summary_data(count_summarydata+4) = avg_SparkDuration
    summary_data(count_summarydata+5) = avg_BlinkDuration
    summary_data(count_summarydata+6) = avg_SparkPeak
    description_summarydata(count_summarydata+1) = "num_Sparks_per_sec_Orphan"
    description_summarydata(count_summarydata+2) = "num_Blinks_per_sec_Orphan"
    description_summarydata(count_summarydata+3) = "num_Quarks_per_sec_Orphan"
    description_summarydata(count_summarydata+4) = "avg_SparksDuration_Orphan"
    description_summarydata(count_summarydata+5) = "avg_BlinkDuration_Orphan"
    description_summarydata(count_summarydata+6) = "avg_SparkPeak_Orphan"
    count_summarydata = count_summarydata + 6
#endif
    
    !write Camyo, Cansr at the end
    Ca_jsr = Ca_jsr_dev
    summary_data(count_summarydata+1) = Ca_myo
    summary_data(count_summarydata+2) = Ca_nsr
    summary_data(count_summarydata+3) = (Ca_nsr*V_nsr_T + SUM(Ca_jsr(1:NSFU)) * V_jsr) / (V_nsr_T+V_jsr_T)
    description_summarydata(count_summarydata+1) = "Camyo"
    description_summarydata(count_summarydata+2) = "Cansr"
    description_summarydata(count_summarydata+3) = "Ca_SRT"
    count_summarydata = count_summarydata + 3

    tmp = tot_SparkTime2Peak
    dvalue = 0.0d0
    !WHERE (tmp .GT. 0d0)  dvalue=dvalue + tmp
    IF (COUNT(tmp .GT. 0d0) .GT. 0) dvalue = SUM(tmp, mask=(tmp .GT. 0d0))/count(tmp .gt. 0d0)
    summary_data(count_summarydata+1) = dvalue
    description_summarydata(count_summarydata+1) = "time2peak_Spark"
    summary_data(count_summarydata+2) = COUNT(tmp .GT. 0d0)
    description_summarydata(count_summarydata+2) = "num_Sparks_detected"
    count_summarydata = count_summarydata + 2


    summary_data(count_summarydata+1) = (accumulate_J_SL_in)/time_duration * Vol_ref/V_myo_T
    summary_data(count_summarydata+2) = accumulate_J_SL_out/time_duration * Vol_ref/V_myo_T  
    summary_data(count_summarydata+3) = accumulate_J_SR_in/time_duration * Vol_ref/V_myo_T  
    summary_data(count_summarydata+4) = accumulate_J_SR_out/time_duration * Vol_ref/V_myo_T  
    description_summarydata(count_summarydata+1) = "mass_J_SL_in"
    description_summarydata(count_summarydata+2) = "mass_J_SL_out"
    description_summarydata(count_summarydata+3) = "mass_J_SR_in"
    description_summarydata(count_summarydata+4) = "mass_J_SR_out"
    count_summarydata = count_summarydata + 4

    IF (count_summarydata .GT. NUM_ENTRIES_DATASUMMARY) THEN
       WRITE(*, log_string) "RED ALERT: Increase size NUM_ENTRIES_DATASUMMARY"
       CALL write2log(log_string)
       count_summarydata = NUM_ENTRIES_DATASUMMARY
    ENDIF
    
    !! write config file name to the first line
    WRITE(fileID, *) TRIM(conf_file), count_summarydata
    DO ii = 1, count_summarydata
       WRITE(fileID,"(1pe12.5,'   ', A)") summary_data(ii), TRIM(description_summarydata(ii))
    END DO
  END SUBROUTINE save_summarized_data_core
  
  SUBROUTINE spark_summarize_1(file2save)
    IMPLICIT NONE
    REAL(KIND=dp) :: time_duration
    CHARACTER(*), OPTIONAL :: file2save

    IF (PRESENT(file2save)) THEN
       fstring = TRIM(where2save) // "/" // TRIM(file2save)
       !!fstring = file2save
    ELSE
       fstring = TRIM(where2save) // "/" // TRIM(summary_filename)
    ENDIF
    
    time_duration = (t_end-time_start_spark_analysis)
    CALL save_summarized_data(fstring, time_duration)
  END SUBROUTINE spark_summarize_1

  SUBROUTINE spark_summarize_2(fileID, duration)
    IMPLICIT NONE
    INTEGER :: fileID
    REAL(KIND=dp), OPTIONAL :: duration
    !!local
    REAL(KIND=dp) :: time_duration

    IF (PRESENT(duration)) THEN
       time_duration = duration
    ELSE    
       time_duration = (t_end-time_start_spark_analysis)
    ENDIF
    CALL save_summarized_data(fileID, time_duration)
  END SUBROUTINE spark_summarize_2
  
  !! !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
#if _DATAMODE_ == _IO_2_HDF2  
  SUBROUTINE datamode7_init()
    IMPLICIT NONE
    LOGICAL :: is_exist
    INTEGER :: hdf_error
    
    WRITE(log_string,*) "Mode 7 (HDF2)"
    CALL write2log(log_string)
    
    hdf_error =  hdf_init()
    CHUNK_SIZE = TRACK_LENGTH
    CHUNKS_PER_FILE = 1 ! set it to one for now
    NUM_STATES_NJ = mR * mL
    NUM_FIELDS4CURRENTS = 11
    !4 columns = 1 for time + 1 for Vm + 1 for Camyo + 1 for Cansr
    !next (NUM_STATES_NJ) = p_nj
    !next (NUM_FIELDS4CURRENTS) = ionic current gating parameters
    FIELDS_PER_RECORD = FIELDS_PER_RECORD + 4 + NUM_STATES_NJ + NUM_FIELDS4CURRENTS
    ! + 2 = CaF, J_ryr_orphan
    FIELDS_PER_RECORD = FIELDS_PER_RECORD + 2
    ! + 2 = J_SRrelease_total, J_dhpr_total
    FIELDS_PER_RECORD = FIELDS_PER_RECORD + 2
    !first NSFU = CRU_states
    !second NSFU = cds
    !third NSFU = cjsr
    FIELDS_PER_RECORD = FIELDS_PER_RECORD + 3*NSFU
 #if _LCC_MODEL_ == _SUN2000
    ! 4th NSFU = CaCM
    FIELDS_PER_RECORD = FIELDS_PER_RECORD + NSFU
 #endif
    
    records_used = 0
    file_index = 0
    ALLOCATE(data_buffer(FIELDS_PER_RECORD, CHUNK_SIZE))
    data_buffer = 0.0d0
    !remove existing files
    sys_command = "rm " // TRIM(where2save) // "/*.h5"
    CALL system(sys_command)
  END SUBROUTINE datamode7_init

  SUBROUTINE datamode7_add_data()
    INTEGER :: ii
    INTEGER :: offset
    
    data_buffer(1, records_used) = time
    data_buffer(2, records_used) = Vm
    data_buffer(3, records_used) = Ca_myo
    data_buffer(4, records_used) = Ca_nsr
!    data_buffer(5, records_used) = I_dhpr_total(1)
   offset = 4
  #if _LCC_MODEL_ == _WILLIAMS2013
    DO ii = 1, mL
       p_nj((ii-1)*mR+1:ii*mR) = pL_nj(ii) * pR_nj(1:mR)
    END DO
    data_buffer(offset+1:offset+NUM_STATES_NJ, records_used) = p_nj
  #endif
    offset = offset + NUM_STATES_NJ
    data_buffer(offset+1:offset+NUM_FIELDS4CURRENTS, records_used) = (/ Na_i, K_i, m_na, h_na, j_na, a_Ktof, in_Ktof, a_Ktos, in_Ktos, a_Kss, in_Kss /)
    offset = offset + NUM_FIELDS4CURRENTS
    data_buffer(offset+1,records_used) = CaF
    data_buffer(offset+2,records_used) = J_ryr_orphan
    offset = offset + 2
    data_buffer(offset+1,records_used) = J_SRrelease_total
    data_buffer(offset+2,records_used) = J_dhpr_total
    offset = offset + 2

   #if _LCC_MODEL_ == _SUN2000
    data_buffer(offset+1:offset+NSFU, records_used) = CaCM_ds_dev
    offset = offset + NSFU
   #endif
    
    data_buffer(offset+1:offset+NSFU, records_used) = CRU_states_dev
    offset = offset + NSFU
    data_buffer(offset+1:offset+NSFU, records_used) =  Ca_ds_dev
    offset = offset + NSFU
    data_buffer(offset+1:offset+NSFU, records_used) = Ca_jsr_dev
    offset = offset + NSFU
  END SUBROUTINE datamode7_add_data
  
  SUBROUTINE datamode7_close()
    INTEGER :: hdf_error
    
    hdf_error = hdf_close()
  END SUBROUTINE datamode7_close
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  SUBROUTINE datamode7_print()
    ! track and output data
    IF (records_used .EQ. CHUNK_SIZE) THEN
       file_index = file_index + 1
       WRITE(file_out, "(a,'/data-', i3.3,'.h5')") TRIM(where2save), file_index         
       CALL hdf_ecc_open(file_out, NSFU, CHUNK_SIZE, CHUNKS_PER_FILE, file_struct, FIELDS_PER_RECORD)
     #if _LCC_MODEL_ == _SUN2000
       CALL hdf_ecc_write_CaCM(file_struct, data_buffer, CHUNK_SIZE, FIELDS_PER_RECORD, NSFU)
     #else
       CALL hdf_ecc_write(file_struct, data_buffer, CHUNK_SIZE, FIELDS_PER_RECORD, NSFU)
     #endif
       CALL hdf_ecc_close(file_struct)
       records_used = 0
    ENDIF
    records_used = records_used + 1
    CALL datamode7_add_data()
    
  END SUBROUTINE datamode7_print
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  SUBROUTINE datamode7_print_last()
    file_index = file_index + 1
    WRITE(file_out, "(a,'/data-', i3.3,'.h5')") TRIM(where2save), file_index
    CALL hdf_ecc_open(file_out, NSFU, records_used, CHUNKS_PER_FILE, file_struct, FIELDS_PER_RECORD)
  #if _LCC_MODEL_ == _SUN2000
    CALL hdf_ecc_write_CaCM(file_struct, data_buffer, records_used, FIELDS_PER_RECORD, NSFU)
  #else
    CALL hdf_ecc_write(file_struct, data_buffer, records_used, FIELDS_PER_RECORD, NSFU)
  #endif
    CALL hdf_ecc_close(file_struct)
  END SUBROUTINE datamode7_print_last

  SUBROUTINE datamode7_print_fb()
    !print out the first-beat
    ! track and output data
    IF (records_used .EQ. CHUNK_SIZE) THEN
       file_index = file_index + 1
       WRITE(file_out, "(a,'/data-', i3.3,'.h5')") TRIM(where2save)//'/'//TRIM(first_beat), file_index         
       CALL hdf_ecc_open(file_out, NSFU, CHUNK_SIZE, CHUNKS_PER_FILE, file_struct, FIELDS_PER_RECORD)
   #if _LCC_MODEL_ == _SUN2000
       CALL hdf_ecc_write_CaCM(file_struct, data_buffer, CHUNK_SIZE, FIELDS_PER_RECORD, NSFU)
   #else
       CALL hdf_ecc_write(file_struct, data_buffer, CHUNK_SIZE, FIELDS_PER_RECORD, NSFU)
   #endif
       CALL hdf_ecc_close(file_struct)
       records_used = 0
    ENDIF
    records_used = records_used + 1
    CALL datamode7_add_data()
  END SUBROUTINE datamode7_print_fb
  SUBROUTINE datamode7_print_fb_last()
    file_index = file_index + 1
    WRITE(file_out, "(a,'/data-', i3.3,'.h5')") TRIM(where2save)//'/'//TRIM(first_beat), file_index
    CALL hdf_ecc_open(file_out, NSFU, records_used, CHUNKS_PER_FILE, file_struct, FIELDS_PER_RECORD)
  #if _LCC_MODEL_ == _SUN2000
    CALL hdf_ecc_write_CaCM(file_struct, data_buffer, records_used, FIELDS_PER_RECORD, NSFU)
  #else
    CALL hdf_ecc_write(file_struct, data_buffer, records_used, FIELDS_PER_RECORD, NSFU)
  #endif
    CALL hdf_ecc_close(file_struct)
    file_index =0
    records_used = 0
  END SUBROUTINE datamode7_print_fb_last
#endif

  SUBROUTINE data_io_write_fb()
    !WRITE INITIAL DATA OUT
#if _DATAMODE_ == _IO_TEST
    !CALL datamode1_print()
#elif _DATAMODE_ == _IO_2_HDF
    CALL datamode2_print_fb()
#elif _DATAMODE_ == _IO_2_ASCII
    !CALL datamode3_print()
#elif _DATAMODE_ == _IO_2_HDF2
    CALL datamode7_print_fb()
#endif
  END SUBROUTINE data_io_write_fb
  
  SUBROUTINE data_io_write_fb_last()
    !WRITE INITIAL DATA OUT
#if _DATAMODE_ == _IO_TEST
    !CALL datamode1_print()
#elif _DATAMODE_ == _IO_2_HDF
    CALL datamode2_print_fb_last()
#elif _DATAMODE_ == _IO_2_ASCII
    !CALL datamode3_print()
#elif _DATAMODE_ == _IO_2_HDF2
    CALL datamode7_print_fb_last()
#endif
  END SUBROUTINE data_io_write_fb_last

  SUBROUTINE write2restartfile()
    IMPLICIT NONE
    INTEGER :: fileID
    CHARACTER(100) :: formatstring, fstring
    INTEGER :: offset
    
    WRITE(log_string,*) "Print out restart file"
    CALL write2log(log_string)
      
    NUM_FIELDS4CURRENTS = 11
    NUM_INDIVIDUAL_FIELDS = 7+NUM_APD_TYPES
    RF_FIELDS_PER_RECORD = 3*NSFU + mL + mR + NUM_FIELDS4CURRENTS + NUM_INDIVIDUAL_FIELDS

    ALLOCATE(data_2restartfile(RF_FIELDS_PER_RECORD,1))
    ! write out data to restart file
    offset = 0
    data_2restartfile(offset+1:offset+NSFU,1) = CRU_states_dev
    offset = offset + NSFU
    data_2restartfile(offset+1:offset+NSFU,1) =  Ca_ds_dev
    offset = offset + NSFU
    data_2restartfile(offset+1:offset+NSFU,1) = Ca_jsr_dev
    offset = offset + NSFU
    data_2restartfile(offset+1:offset+mL, 1) = pL_nj
    offset = offset + mL
    data_2restartfile(offset+1:offset+mR, 1) = pR_nj
    offset = offset + mR
    data_2restartfile(offset+1:offset+NUM_FIELDS4CURRENTS, 1) = (/ Na_i, K_i, m_na, h_na, j_na, a_Ktof, in_Ktof, a_Ktos, in_Ktos, a_Kss, in_Kss /)
    offset = offset + NUM_FIELDS4CURRENTS
    data_2restartfile(offset+1,1) = Vm
    data_2restartfile(offset+2,1) = Ca_myo
    data_2restartfile(offset+3,1) = Ca_nsr
    data_2restartfile(offset+4,1) = is_being_clamped
    data_2restartfile(offset+5,1) = CaB_htrpn
    data_2restartfile(offset+6,1) = CaF
    data_2restartfile(offset+7:offset+6+NUM_APD_TYPES,1) = is_inside_AP
    offset = offset + 6 + NUM_APD_TYPES
    data_2restartfile(offset+1, 1) = t_end
    offset = offset + 1
    data_2restartfile(offset + 1, 1) = CaCM
    data_2restartfile(offset + 2, 1) = CaSL
    data_2restartfile(offset + 3, 1) = CaSR
    offset = offset + 3
    
    WRITE(formatstring, "(i0)") RF_FIELDS_PER_RECORD
    formatstring = "(" // TRIM(formatstring) //  "(2x,1pe21.13))"
    
    fstring = TRIM(where2save) // "/" // TRIM(restart_filename) 
    fileID = get_file_ID()
    OPEN (unit = fileID, file = TRIM(fstring))
    WRITE (fileID, TRIM(formatstring)) data_2restartfile
    CLOSE (fileID)
  END SUBROUTINE write2restartfile
    
  SUBROUTINE spark_summarize_fb()
    IMPLICIT NONE
    REAL(KIND=dp) :: time_duration

    fstring = TRIM(where2save) // "/" // TRIM(first_beat) // "/" // TRIM(summary_filename)
    time_duration = 1.0d0
    CALL save_summarized_data(fstring, time_duration)
    
    INTEGRAL_GAIN = 0d0
    PEAK_GAIN = 0.0d0
    peak_J_SRrelease_total = 0d0
    peak_J_dhpr_total = 0d0
    accumulate_J_dhpr = 0.0d0
    accumulate_J_ryr = 0.0d0; accumulate_J_ryr_orphan = 0.0d0
    accumulate_J_SL_out = 0.0d0;  accumulate_J_SL_in = 0.0d0
    accumulate_J_SR_out = 0.0d0;  accumulate_J_SR_in = 0.0d0
    APD = 0d0; APstart = 0d0; APend = 0d0;
    Vmthreshold = (/ V_rest * 0.9d0, V_rest * 0.5d0, V_rest * 0.2d0, +0.0d0 /)
    flag1 = .FALSE.
    flag2 = .FALSE.
    flag3 = .FALSE.
    flag4 = .FALSE.
    Nquarks_dev = 0
    Nsparks_dev = 0
    Nblinks_dev = 0
    tot_SparkPeaks = 0d0
    tot_SparkTime2Peak = 0d0
    tot_BlinkDurations = 0d0
    is_inside_AP = .FALSE.
    APstart = 0d0
    APend = 0d0
    APcount = +0
    is_in_DiastolicPhase = .FALSE.
    is_done_firstbeat = .FALSE.
    ts_spark = 0d0
    ts_blink = 0d0
  END SUBROUTINE spark_summarize_fb

  attributes(device) SUBROUTINE state_breakdown_ker(cru_state, irow_R, lcc_state, ryr_state)
    USE cudafor
    IMPLICIT NONE
    INTEGER, INTENT(IN) :: cru_state, irow_R
    !INTEGER, INTENT(INOUT) :: lcc_state, ryr_state
    INTEGER :: lcc_state, ryr_state
    
    lcc_state = (cru_state-1)/irow_R + 1
    ryr_state = cru_state - (lcc_state-1)*irow_R

  END SUBROUTINE state_breakdown_ker

  attributes(device) FUNCTION state_combine_single_ker(LCCcluster_state, RYRcluster_state, irow_R) RESULT(cru_state)
    USE cudafor
    INTEGER :: cru_state
    INTEGER, INTENT(IN) :: LCCcluster_state, RYRcluster_state
    INTEGER, INTENT(IN) :: irow_R
    
    cru_state = (LCCcluster_state-1) * irow_R + RYRcluster_state

    RETURN
  END FUNCTION state_combine_single_ker

  INTEGER FUNCTION print_mode_info() RESULT(res)
    !! print out information which mode is being used
    IMPLICIT NONE

    res = SUCCESS

#if _RUNMODE_ == _MODE_RESTING
    WRITE(log_string,*) "Resting"
    CALL write2log(log_string)
#elif _RUNMODE_ == _MODE_V_CLAMP
    WRITE(log_string,*) "V-clamped"
    CALL write2log(log_string)
!!$    WRITE(log_string, "(A, F6.2, A, F6.2, A, A, F3.0)") "V_stim = ", V_stim, "mV; duration = ", (t_V_stim_end - t_V_stim_start) * 1d3, "ms", "; Hz=", Hz
    WRITE(log_string, "(A, F6.2, A, F6.2, A, A, F3.0)") "V_stim = ", V_stim, "mV; duration = ", (t_V_stim_end - t_V_stim_start) * 1d3, "ms", "; BCL=", BCL
    CALL write2log(log_string)

#elif _RUNMODE_ == _MODE_I_CLAMP
    WRITE(log_string,*) "I-clamped"
    CALL write2log(log_string)
!!$    WRITE(log_string, "(A, F6.2, A, F6.2, A, A, F3.0)") "I_app = ", I_stim, "uA/cm^2; duration = ", (t_I_stim_end - t_I_stim_start)*1d3, "ms", "; Hz=", Hz
    WRITE(log_string, "(A, F6.2, A, F6.2, A, A, F3.0)") "I_app = ", I_stim, "uA/cm^2; duration = ", (t_I_stim_end - t_I_stim_start)*1d3, "ms", "; BCL=", BCL
    CALL write2log(log_string)

#elif _RUNMODE_ == _MODE_CLOSEDCELL
    WRITE(log_string,*) "Closed cell (no ion in/out)"
    CALL write2log(log_string)

#elif _RUNMODE_ == _MODE_V_CLAMP_TEST_LCC
    WRITE(log_string,*) "V-clamped to test LCC, but disable other membrane ionic currents, RYR, SERCA"
    CALL write2log(log_string)
!!$    WRITE(log_string, "(A, F6.2, A, F6.2, A, A, F3.0)") "V_stim = ", V_stim, "mV; duration = ", (t_V_stim_end - t_V_stim_start) * 1d3, "ms", "; Hz=", Hz
    WRITE(log_string, "(A, F6.2, A, F6.2, A, A, F3.0)") "V_stim = ", V_stim, "mV; duration = ", (t_V_stim_end - t_V_stim_start) * 1d3, "ms", "; BCL=", BCL
    CALL write2log(log_string)
    
#else
    WRITE(log_string,*) "Error: unknown mode"
    CALL write2log(log_string)
    STOP
#endif    

#if _USE_CaDS_APPROX_ == _NO
    IF (dt_UB .GT. 1.0d-7) THEN
       WRITE(log_string,*) "WARNING: Need to reduce dt for exact calculation"
       CALL write2log(log_string)
    END IF
#endif
        
  END FUNCTION print_mode_info

  SUBROUTINE gpu_cleanup()
    IMPLICIT NONE
    INTEGER :: err
    
    err = cudaStreamDestroy(stream0)
    err = cudaStreamDestroy(stream1)
    
    err = cudaDeviceReset()
  END SUBROUTINE gpu_cleanup

  SUBROUTINE trigger_CRU_openLCC(U_Lo, CRU_states, NSFU, irow_R, LCCcluster_states, RYRcluster_states, is_all_CRU_open)
    ! this version: open LCC for a period of time (default: 0.5ms) and hopefully it can trigger the whole CRU
    IMPLICIT NONE
    REAL(KIND=dp), DIMENSION(:), INTENT(IN) :: U_Lo
    INTEGER, DIMENSION(:):: CRU_states
    INTEGER, INTENT(IN) :: NSFU, irow_R
    INTEGER, DIMENSION(:) :: RYRcluster_states, LCCcluster_states
    LOGICAL, INTENT(IN) :: is_all_CRU_open
    ! local
    INTEGER :: ii
    INTEGER :: number_channels_open
    
    number_channels_open = 1  !set 1 channel open
    !!PRINT *, "start trigger_CRU() by opening ", number_channels_open, " LCC"
    IF (is_all_CRU_open) number_channels_open = 0 !! done, now we close LCC channels
    
    CALL state_breakdown_all(CRU_states(1:NSFU), irow_R, LCCcluster_states, RYRcluster_states)
 #if _NUM_CRU_2_TRIGGER == _CRU3_TRIGGERED
    WRITE (log_string, *) "Trigger only CRU #3" 
    CALL write2log(log_string)
    PRINT *, log_string
    ii = 3
    IF (U_Lo(LCCcluster_states(ii)) .LT. number_channels_open) THEN
       !IF ((MOD(ii,4) .EQ. 0) .AND. U_Ro(RYRcluster_states(ii)) .LT. number_channels_open) THEN !(just for testing)
       LCCcluster_states(ii) = index_LCCcluster_numLCCopen(number_channels_open)
    ENDIF
 #elif _NUM_CRU_2_TRIGGER == _CRUS_ALL_TRIGGERED
    WRITE (log_string, *) "Trigger all CRUs" 
    CALL write2log(log_string)
    DO ii = 1, NSFU
       IF (U_Lo(LCCcluster_states(ii)) .LT. number_channels_open) THEN
       !IF ((MOD(ii,4) .EQ. 0) .AND. U_Ro(RYRcluster_states(ii)) .LT. number_channels_open) THEN !(just for testing)
          LCCcluster_states(ii) = index_LCCcluster_numLCCopen(number_channels_open)
       ENDIF
    END DO
 #elif _NUM_CRU_2_TRIGGER == _CRUS_50PERCENT_TRIGGERED
    WRITE (log_string, *) "Trigger all CRUs" 
    CALL write2log(log_string)
    DO ii = 1, NSFU, 2
       IF (U_Lo(LCCcluster_states(ii)) .LT. number_channels_open) THEN
       !IF ((MOD(ii,4) .EQ. 0) .AND. U_Ro(RYRcluster_states(ii)) .LT. number_channels_open) THEN !(just for testing)
          LCCcluster_states(ii) = index_LCCcluster_numLCCopen(number_channels_open)
       ENDIF
    END DO
 #elif _NUM_CRU_2_TRIGGER == _CRUS_25PERCENT_TRIGGERED
    WRITE (log_string, *) "Trigger all CRUs" 
    CALL write2log(log_string)
    DO ii = 1, NSFU, 4
       IF (U_Lo(LCCcluster_states(ii)) .LT. number_channels_open) THEN
       !IF ((MOD(ii,4) .EQ. 0) .AND. U_Ro(RYRcluster_states(ii)) .LT. number_channels_open) THEN !(just for testing)
          LCCcluster_states(ii) = index_LCCcluster_numLCCopen(number_channels_open)
       ENDIF
    END DO
 #endif

    CRU_states(1:NSFU) = state_combine_all(LCCcluster_states, &
               RYRcluster_states, irow_R)
  END SUBROUTINE trigger_CRU_openLCC

  SUBROUTINE trigger_CRU(U_Ro, CRU_states, NSFU, irow_R, LCCcluster_states, RYRcluster_states, &
       is_all_CRU_open)
    ! this version: set a number of RYRs opening and hopefully it can trigger the whole CRU
    IMPLICIT NONE
    REAL(KIND=dp), DIMENSION(:), INTENT(IN) :: U_Ro
    INTEGER, DIMENSION(:):: CRU_states
    INTEGER, INTENT(IN) :: NSFU, irow_R
    INTEGER, DIMENSION(:) :: RYRcluster_states, LCCcluster_states
    LOGICAL, INTENT(INOUT) :: is_all_CRU_open
    ! local
    INTEGER :: ii
    INTEGER :: number_channels_open
    INTEGER :: counter = 0
    INTEGER, PARAMETER :: num_timesteps_holdingchannels = 20

    PRINT *, "start trigger_CRU()"
    number_channels_open = 8  !set 8 channels open

    
    CALL state_breakdown_all(CRU_states(1:NSFU), irow_R, LCCcluster_states, RYRcluster_states)
 #if _NUM_CRU_2_TRIGGER == _CRU3_TRIGGERED
    WRITE (log_string, *) "Trigger only CRU #3" 
    CALL write2log(log_string)
    ii = 3
    IF (U_Ro(RYRcluster_states(ii)) .LT. number_channels_open) THEN
       !IF ((MOD(ii,4) .EQ. 0) .AND. U_Ro(RYRcluster_states(ii)) .LT. number_channels_open) THEN !(just for testing)
       RYRcluster_states(ii) = index_RYRcluster_numRYRopen(number_channels_open)
    ENDIF
 #elif _NUM_CRU_2_TRIGGER == _CRUS_ALL_TRIGGERED
    WRITE (log_string, *) "Trigger all CRUs" 
    CALL write2log(log_string)
    DO ii = 1, NSFU
       IF (U_Ro(RYRcluster_states(ii)) .LT. number_channels_open) THEN
       !IF ((MOD(ii,4) .EQ. 0) .AND. U_Ro(RYRcluster_states(ii)) .LT. number_channels_open) THEN !(just for testing)
          RYRcluster_states(ii) = index_RYRcluster_numRYRopen(number_channels_open)
       ENDIF
    END DO
 #elif _NUM_CRU_2_TRIGGER == _CRUS_50PERCENT_TRIGGERED
    WRITE (log_string, *) "Trigger all CRUs" 
    CALL write2log(log_string)
    DO ii = 1, NSFU, 2
       IF (U_Ro(RYRcluster_states(ii)) .LT. number_channels_open) THEN
       !IF ((MOD(ii,4) .EQ. 0) .AND. U_Ro(RYRcluster_states(ii)) .LT. number_channels_open) THEN !(just for testing)
          RYRcluster_states(ii) = index_RYRcluster_numRYRopen(number_channels_open)
       ENDIF
    END DO
 #elif _NUM_CRU_2_TRIGGER == _CRUS_25PERCENT_TRIGGERED
    WRITE (log_string, *) "Trigger all CRUs" 
    CALL write2log(log_string)
    DO ii = 1, NSFU, 4
       IF (U_Ro(RYRcluster_states(ii)) .LT. number_channels_open) THEN
       !IF ((MOD(ii,4) .EQ. 0) .AND. U_Ro(RYRcluster_states(ii)) .LT. number_channels_open) THEN !(just for testing)
          RYRcluster_states(ii) = index_RYRcluster_numRYRopen(number_channels_open)
       ENDIF
    END DO
 #endif

    CRU_states(1:NSFU) = state_combine_all(LCCcluster_states, &
               RYRcluster_states, irow_R)

    !!!!NOTE: Instead of allowing channel to change state immediately, we wait for a certain time
    counter = counter + 1
    IF (counter .EQ. num_timesteps_holdingchannels) THEN
       is_all_CRU_open = .TRUE.
       WRITE(log_string, *) "Complete trigger CRU"
       CALL write2log(log_string)
       counter = 0
    ENDIF
    !!is_all_CRU_open = .TRUE.
  END SUBROUTINE trigger_CRU

  SUBROUTINE get_index_XXXcluster_numXXXopen(N_XXX, irow_XXX, NSFU, U_XXXo, &
       index_XXXcluster_numXXXopen)
    ! 1. return the XXXcluster index where the #XXXopen is  1..N_XXX
    !    output is saved into index_XXXcluster_numXXXopen
    ! with XXX = RYR or LCC
    ! NOTE: This function will be replaced for get_index_RYRcluster_numRYRopen()
    IMPLICIT NONE
    INTEGER, INTENT(IN) :: N_XXX, irow_XXX, NSFU
    REAL(KIND=dp), DIMENSION(:) :: U_XXXo
    INTEGER, DIMENSION(0:N_XXX) :: index_XXXcluster_numXXXopen
    ! local
    INTEGER :: ii, jj
    
    DO jj = 0, N_XXX
       DO ii = 1, irow_XXX
          IF (ABS(U_XXXo(ii) .EQ. jj)) THEN 
             index_XXXcluster_numXXXopen(jj) = ii
             EXIT
          END IF
       END DO
    ENDDO    
  END SUBROUTINE get_index_XXXcluster_numXXXopen

  SUBROUTINE write_SL_flux(time, Ca_myo, accumulate_J_in, accumulate_J_out, Vm)
    !!! print out the total flux (across SL) at the end of each second
    REAL(KIND=dp), INTENT(IN) :: time, accumulate_J_out, accumulate_J_in, Ca_myo, Vm
    !!local
    REAL(KIND=dp), SAVE :: checkpoint_time = 0.0d0
    REAL(KIND=dp), SAVE :: checkpoint_J_out = 0.0d0, checkpoint_J_in = 0.0d0
    LOGICAL, SAVE :: done_this_second = .FALSE.
    INTEGER, SAVE :: fileID
    CHARACTER(MAX_STR_LEN), SAVE :: file_out    
    CHARACTER(MAX_STR_LEN), SAVE :: formatstring

    IF (checkpoint_time .EQ. 0.0d0) THEN
       WRITE(file_out, "(a,'/flux_SL.dat')") TRIM(where2save)         
       PRINT *, 'delete flux_SL.dat'
       sys_command = "rm " // TRIM(file_out)
       CALL system(sys_command)
       formatstring = "(10(2x,1pe21.13))"
       fileID = get_file_ID()
       OPEN(unit=fileID, file=TRIM(file_out), ACCESS='append')
       checkpoint_time = time_start_spark_analysis+2*EPSILON(1.0d0)
    ENDIF
    !IF (time .GE. checkpoint_time .AND. done_this_second .EQ. .FALSE.) THEN
    IF (time .GE. checkpoint_time) THEN
       !! write data
       checkpoint_J_in = accumulate_J_in - checkpoint_J_in
       checkpoint_J_out = accumulate_J_out - checkpoint_J_out
       WRITE(fileID, formatstring) time, Ca_myo, checkpoint_J_in, checkpoint_J_out, Vm
       CALL FLUSH(fileID)
       checkpoint_J_in = accumulate_J_in
       checkpoint_J_out = accumulate_J_out
       !!wait until next second
       checkpoint_time = checkpoint_time + 1.0d0
       !checkpoint_time = checkpoint_time + 0.10d0
       !done_this_second = .TRUE.
    ENDIF
    IF (checkpoint_time .GT. t_end - 1.0d0 .AND. checkpoint_time .LT. t_end) THEN
       PRINT *, 'complete flux_SL.dat'       
       CLOSE(fileID)
       checkpoint_time  = checkpoint_time + 1.0d0
    ENDIF
    
  END SUBROUTINE write_SL_flux

  SUBROUTINE write_SR_flux(time, Ca_nsr, accumulate_J_in, accumulate_J_out, Ca_myo)
    !!! print out the total flux (across SR) at the end of each second
    REAL(KIND=dp), INTENT(IN) :: time, accumulate_J_out, accumulate_J_in, Ca_nsr, Ca_myo
    !!local
    REAL(KIND=dp), SAVE :: checkpoint_time = 0.0d0
    REAL(KIND=dp), SAVE :: checkpoint_J_out = 0.0d0, checkpoint_J_in = 0.0d0
    LOGICAL, SAVE :: done_this_second = .FALSE.
    INTEGER, SAVE :: fileID
    CHARACTER(MAX_STR_LEN), SAVE :: formatstring, file_out

    IF (checkpoint_time .EQ. 0.0d0) THEN
       WRITE(file_out, "(a,'/flux_SR.dat')") TRIM(where2save)         
       PRINT *, 'delete flux_SR.dat'
       sys_command = "rm " // TRIM(file_out)
       CALL system(sys_command)
       formatstring = "(10(2x,1pe21.13))"
       fileID = get_file_ID()
       OPEN(unit=fileID, file=TRIM(file_out), ACCESS='append')
       checkpoint_time = time_start_spark_analysis+2*EPSILON(1.0d0)
    ENDIF
    !IF (time .GE. checkpoint_time .AND. done_this_second .EQ. .FALSE.) THEN
    IF (time .GE. checkpoint_time) THEN
       !! write data
       checkpoint_J_in = accumulate_J_in - checkpoint_J_in
       checkpoint_J_out = accumulate_J_out - checkpoint_J_out
       !!! time, Cansr, Jsr_in, Jsr_out, Cmyo
       WRITE(fileID, formatstring) time, Ca_nsr, checkpoint_J_in, checkpoint_J_out, Ca_myo
       CALL FLUSH(fileID)
       checkpoint_J_in = accumulate_J_in
       checkpoint_J_out = accumulate_J_out
       !!wait until next second
       checkpoint_time = checkpoint_time + 1.0d0
       !checkpoint_time = checkpoint_time + 0.10d0
       !done_this_second = .TRUE.
    ENDIF
    IF (checkpoint_time .GT. t_end - 1.0d0 .AND. checkpoint_time .LT. t_end) THEN
       PRINT *, 'complete flux_SR.dat'       
       CLOSE(fileID)
       checkpoint_time  = checkpoint_time + 1.0d0
    ENDIF
    
  END SUBROUTINE write_SR_flux
  
#if _LCC_MODEL_ == _SUN2000  
  attributes(global) SUBROUTINE get_compPdt_Sun2000()
    USE cudafor
    IMPLICIT NONE
    
    ! local
    REAL(KIND=dp) :: bigC2, chiC, chiO, phi, dtemp, dtemp2,new
    INTEGER:: tx, ii !,  ct   ! ct = current state
    REAL(KIND=dp), DIMENSION(BLOCKSIZE_1D), shared :: s_cads, s_ryrgate
    INTEGER,VOLATILE :: lcc_ct, ryr_ct
    INTEGER :: jj, fromto, from, to
    REAL(KIND=dp) :: dfactor
    
    ii =  (blockIdx%x-1) * blockDim%x + threadIdx%x
    
    IF (ii .LE. int_ca(1)) THEN
       tx = threadIdx%x
       s_cads(tx) = Ca_ds_dev(ii)
       s_ryrgate(tx) = RYRgate_dev(ii)
    ENDIF
    CALL syncthreads()
       
    IF (ii .LE. int_ca(1)) THEN
       CALL state_breakdown_ker(CRU_states_dev(ii), int_ca(10), lcc_ct, ryr_ct)
       !bigC2 = Ca_ds_dev(ii)**eta_RyR
       dtemp = LOG(s_cads(tx))
       bigC2 = EXP(dtemp*dp_ca(14))
       
       !chiC = EXP(-Ej*0.5d0*((N_R-RYRgate_dev(ii))*Ecc - (RYRgate_dev(ii)-1d0)*Eoo))
       dtemp = (int_ca(2)-s_ryrgate(tx))*dp_ca(2)
       dtemp2 = dtemp - (s_ryrgate(tx) - 1.d0)*dp_ca(3)
       dtemp = dp_ca(1)*dtemp2
       chiC = EXP(dtemp)
       
       !chiO = EXP(-Ej*0.5d0*(RYRgate_dev(ii)*Eoo - (N_R-RYRgate_dev(ii)- 1d0)*Ecc))
       dtemp = s_ryrgate(tx)*dp_ca(3)
       dtemp2 = dtemp - (int_ca(2) - s_ryrgate(tx) - 1d0)*dp_ca(2)
       dtemp = dp_ca(1)*dtemp2
       chiO = EXP(dtemp)

       !phi = k_jsr0*Ca_jsr_dev(ii) + k_jsr1
       phi = dp_ca(4)*Ca_jsr_dev(ii) + dp_ca(5)       
   !      phi = dp_ca(4)*Ca_jsr_dev(ii) + dp_ca(5)     
       
! use 1st column for negative value
       dtemp = chiC * ryr_compKfree_dev(ryr_ct,1)  
       dtemp2 = dtemp + bigC2 * chiO * phi * ryr_compKcal_act_dev(ryr_ct,1)
       DO jj=1, lcc_indxK_dev(lcc_ct,0)
          !integer :: from, to
          fromto = lcc_stateFromTo_dev(lcc_ct,jj)
          to = IBITS(fromto, 0,8)                
          from = IBITS(fromto, 8,8)
          !!no need thermo(1,5)
          IF  (from.EQ.5 .AND. to.EQ.4) THEN
             !!thermo(5,4)
            #if _CaCM_STRATEGY_ == _SUN_STRATEGY || _CaCM_STRATEGY_ == _STRATEGY2
             dfactor = lcc_Ktransitionrate_dev(4,5)*lcc_Ktransitionrate_dev(5,2)*lcc_Krateconstant_dev(2,4) * CaCM_ds_dev(ii)/&
                  (lcc_Ktransitionrate_dev(4,2)*lcc_Ktransitionrate_dev(2,5))
            #elif _CaCM_STRATEGY_ == _STRATEGY3
             dfactor = lcc_Ktransitionrate_dev(4,5)*lcc_Ktransitionrate_dev(5,2)*lcc_Krateconstant_dev(2,4) * CaCM_ds_dev(ii)**dp_ca(15)/&
                  (lcc_Ktransitionrate_dev(4,2)*lcc_Ktransitionrate_dev(2,5))
            #elif _CaCM_STRATEGY_ == _STRATEGY1
             dfactor = lcc_Ktransitionrate_dev(4,5)*lcc_Ktransitionrate_dev(5,2)*lcc_Krateconstant_dev(2,4) * CaCM_ds_dev(ii)**dp_ca(15)/&
                  (lcc_Ktransitionrate_dev(4,2)*lcc_Ktransitionrate_dev(2,5))
            #endif
          ELSE IF (from.EQ. 2 .AND. to.EQ. 4) THEN
             !the case of CaCM-dependent
           #if _CaCM_STRATEGY_ == _SUN_STRATEGY || _CaCM_STRATEGY_ == _STRATEGY2
             dfactor = lcc_Krateconstant_dev(2,4) * CaCM_ds_dev(ii)
           #elif _CaCM_STRATEGY_ == _STRATEGY3
             dfactor = lcc_Krateconstant_dev(2,4) * CaCM_ds_dev(ii)**dp_ca(15)
           #elif _CaCM_STRATEGY_ == _STRATEGY1
             dfactor = lcc_Krateconstant_dev(2,4) * CaCM_ds_dev(ii)**dp_ca(15)
           #endif
             !!TUAN: test with cut-off 15.0(uM)
             !not work: 5.0(uM)
             !dfactor = lcc_Krateconstant_dev(2,4) * MIN(CaCM_ds_dev(ii), 15.0d0) 
          ELSE
             dfactor = lcc_Ktransitionrate_dev(from,to)
          ENDIF

          !NOTE: subtract to get negative value
          dtemp = dfactor * lcc_StateSpace_dev(lcc_ct,from)
          lcc_compK_dev(ii,jj) =  dtemp
          dtemp2 = dtemp2 - dtemp

       ENDDO
       compPdt_dev(ii) = dtemp2
    END IF
  END SUBROUTINE get_compPdt_Sun2000

  attributes(global) SUBROUTINE update_SFU_Sun2000(Xrnd, dt, &
       Vm, dp_arg1, dp_arg2, Ca_myo, Ca_nsr,new)
    !new method
    USE cudafor
    IMPLICIT NONE
    REAL(KIND=dp), DIMENSION(:), device :: Xrnd
    REAL(KIND=dp), VALUE :: dt, Ca_nsr, Ca_myo,new, Vm, dp_arg1, dp_arg2

    !local
    INTEGER :: ii, jj, kk, tx
    REAL(KIND=dp) :: dtemp, dtemp2, compP, J_ryr
    !INTEGER :: nt, ct, cnt
    !nt = new state, ct = current state, cnt= candidate next state
    REAL(KIND=dp) :: bigC1, bigC2, chiC, chiO, phi, J_dhpr, J_CaCM, d_CaCM
    REAL(KIND=dp) :: nryr ! new RYRgate
    REAL(KIND=dp) :: ndhpr !new dhprgate
    INTEGER :: itemp
    REAL(KIND=dp), DIMENSION(BLOCKSIZE_1D), shared :: s_cads, s_ryrgate
    INTEGER, VOLATILE :: lcc_ct, ryr_ct
    INTEGER :: lcc_nt, ryr_nt
    INTEGER :: is_examining
    LOGICAL :: found_nt
    INTEGER :: candidate_new_state !can be any cluster types (LCC or RyR)
  
    ! thread Index
    ii =  (blockIdx%x-1) * blockDim%x + threadIdx%x

    IF (ii .LE. int_ca(1)) THEN
       tx = threadIdx%x
       s_cads(tx) = Ca_ds_dev(ii)
       s_ryrgate(tx) = RYRgate_dev(ii)
    ENDIF
    CALL syncthreads()

    IF (ii .LE. int_ca(1)) THEN
       CALL state_breakdown_ker(CRU_states_dev(ii), int_ca(10), lcc_ct, ryr_ct)

       lcc_nt = lcc_ct
       ryr_nt = ryr_ct
       found_nt = .FALSE.
       compP = compPdt_dev(ii) * dt  + 1.0d0

       IF (compP .GE. Xrnd(ii)) THEN ! channel doesn't change state (easy)
          found_nt = .TRUE.
       ELSE ! channel change to some new state
          dtemp = LOG(s_cads(tx))
          dtemp2 = dtemp *dp_ca(14)
          bigC2 = EXP(dtemp2)

          dtemp = (int_ca(2)-s_ryrgate(tx))*dp_ca(2)
          dtemp2 = dtemp - (s_ryrgate(tx) - 1.d0)*dp_ca(3)
          dtemp = dp_ca(1)*dtemp2
          chiC = EXP(dtemp)
          !dtemp = (int_ca(2)-RYRgate_dev(ii))*dp_ca(2)
          !dtemp2 = dtemp - (RYRgate_dev(ii) - 1d0)*dp_ca(3)
          !dtemp = dp_ca(1)*dtemp2
          !chiC = EXP(dtemp)

          dtemp = s_ryrgate(tx)*dp_ca(3)
          dtemp2 = dtemp - (int_ca(2) - s_ryrgate(tx) - 1d0)*dp_ca(2)
          dtemp = dp_ca(1)*dtemp2
          chiO = EXP(dtemp)
          !dtemp = RYRgate_dev(ii)*dp_ca(3)
          !dtemp2= dtemp - (int_ca(2)-RYRgate_dev(ii)- 1d0)*dp_ca(2)
          !dtemp = dp_ca(1)*dtemp2
          !chiO = EXP(dtemp)

          phi = dp_ca(4)*Ca_jsr_dev(ii) + dp_ca(5)

       !  phi = dp_ca(4)*Ca_jsr_dev(ii) + dp_ca(5)

          DO jj=2, int_ca(3)+int_ca(4)-1
             IF (found_nt .EQ. .FALSE.) THEN
                IF (jj .LE. int_ca(3)) THEN
                   ![Ca]ds-act RYR
                   kk = jj
                   dtemp = bigC2 * chiO * phi * ryr_compKcal_act_dev(ryr_ct,kk)
                   candidate_new_state = ryr_indxKcal_act_dev(ryr_ct,kk)
                ELSE IF (jj .LE. int_ca(3)+int_ca(4)-1) THEN
                   !inact RYR
                   kk = jj-(int_ca(3)-1)
                   dtemp = chiC * ryr_compKfree_dev(ryr_ct,kk)
                   candidate_new_state = ryr_indxKfree_dev(ryr_ct,kk)
                ELSE
                ENDIF
                compP = compP + dtemp * dt
                IF (compP .GE. Xrnd(ii)) THEN
                   found_nt = .TRUE.
                   ryr_nt = candidate_new_state
                   EXIT
                END IF
             ENDIF
          END DO
          IF (found_nt .EQ. .FALSE.) THEN
             DO jj=1, lcc_indxK_dev(lcc_ct,0)
                compP = compP + lcc_compK_dev(ii,jj) * dt
                IF (compP .GE. Xrnd(ii)) THEN
                   found_nt = .TRUE.
                   !to avoid LCC gating at negative potential
                   !IF (Vm .GE. -70.0d0)  lcc_nt = candidate_new_state
                   lcc_nt = lcc_indxK_dev(lcc_ct,jj)
                   EXIT !current DO-loop
                ENDIF
             ENDDO
          END IF
       END IF

       nryr = U_Ro_dev(ryr_nt)
       ndhpr = U_Lo_dev(lcc_nt)

       J_efflux_dev(ii) = dp_ca(11)*(s_cads(tx)-Ca_myo)
       J_refill_dev(ii) = dp_ca(10)*(Ca_nsr-Ca_jsr_dev(ii))

#if (_RUNMODE_ == _MODE_I_CLAMP || _RUNMODE_ == _MODE_V_CLAMP || _RUNMODE_ == _MODE_V_CLAMP_TEST_LCC)
       !I-clamp or V-clamp
       I_dhpr_dev(ii)  = ndhpr * dp_ca(13) * dp_ca(16) * Vm * &
            (s_cads(tx) * dp_arg1 - dp_ca(18)) * dp_arg2  ! [uA]
       !J_dhpr_dev(ii)  = -I_dhpr_dev(ii) * dp_ca(17)  ! [uA/s]
       J_dhpr  = -I_dhpr_dev(ii) * dp_ca(17)  ! [uA/s]
#else
       I_dhpr_dev(ii)  = 0.0d0
       !J_dhpr_dev(ii) = 0d0       
       J_dhpr = 0d0
#endif

       !!J_ryr = nryr * dp_ca(9) * (Ca_jsr_dev(ii)-Ca_ds_dev(ii)) !J_ryr
       J_ryr = nryr * dp_ca(9) * (Ca_jsr_dev(ii)-s_cads(tx))
       J_ryr_dev(ii) = J_ryr

    #if _CaCM_STRATEGY_ == _SUN_STRATEGY
       d_CaCM = dt * (dp_ca(21) * (s_cads(tx)**dp_ca(24)) * (dp_ca(23)-CaCM_ds_dev(ii)) - dp_ca(22) * CaCM_ds_dev(ii))
       CaCM_ds_dev(ii) = MIN(dp_ca(23), CaCM_ds_dev(ii) + d_CaCM)
       s_cads(tx) = s_cads(tx) + dt * dp_ca(6) * dp_ca(7) * (J_ryr - J_efflux_dev(ii) + J_dhpr)
    #elif _CaCM_STRATEGY_ == _STRATEGY1
       d_CaCM = dt * (dp_ca(21) * (s_cads(tx)**1) * (dp_ca(23)-CaCM_ds_dev(ii)) - dp_ca(22) * CaCM_ds_dev(ii))
       CaCM_ds_dev(ii) = MIN(dp_ca(23), CaCM_ds_dev(ii) + d_CaCM)
       s_cads(tx) = s_cads(tx) + dt * dp_ca(6) * dp_ca(7) * (J_ryr - J_efflux_dev(ii) + J_dhpr)
    #elif _CaCM_STRATEGY_ == _STRATEGY2
       d_CaCM = dt * (dp_ca(21) * (s_cads(tx)**dp_ca(24)) * (dp_ca(23)-CaCM_ds_dev(ii)) - dp_ca(22) * CaCM_ds_dev(ii))
       !!d_CaCM = 0.0d0 !test disable [CaCM]-dependent
       CaCM_ds_dev(ii) = MIN(dp_ca(23), CaCM_ds_dev(ii) + d_CaCM)
       s_cads(tx) = s_cads(tx) + dt * dp_ca(7) * (J_ryr - J_efflux_dev(ii) + J_dhpr)
       !!IF (CaCM_ds_dev(ii) .LT. dp_ca(23)) s_cads(tx) = s_cads(tx) - d_CaCM  !!test only one calcium bind
       IF (CaCM_ds_dev(ii) .LT. dp_ca(23)) s_cads(tx) = s_cads(tx) - dp_ca(24)*d_CaCM
    #elif _CaCM_STRATEGY_ == _STRATEGY3
       d_CaCM = dt * (dp_ca(21) * (s_cads(tx)**dp_ca(24)) * (dp_ca(23)-CaCM_ds_dev(ii)) - dp_ca(22) * CaCM_ds_dev(ii))
       CaCM_ds_dev(ii) = MIN(dp_ca(23), CaCM_ds_dev(ii) + d_CaCM)
       s_cads(tx) = s_cads(tx) + dt * dp_ca(7) * (J_ryr - J_efflux_dev(ii) + J_dhpr)
       IF (CaCM_ds_dev(ii) .LT. dp_ca(23)) s_cads(tx) = s_cads(tx) - dp_ca(24)*d_CaCM
    #endif
       
       !!!WHICH ONE
       !s_cads(tx) = s_cads(tx) + dt * dp_ca(7) * (J_ryr - J_efflux_dev(ii) + J_dhpr) - d_CaCM
       !s_cads(tx) = s_cads(tx) - d_CaCM
       !s_cads(tx) = s_cads(tx) - d_CaCM * 3
       !s_cads(tx) = s_cads(tx) - d_CaCM * 4
       
       dtemp = dp_ca(19) + Ca_jsr_dev(ii)
       dtemp2 = dtemp * dtemp
       dtemp = dtemp2 / (dp_ca(19)*dp_ca(20) + dtemp2)   !beta_jsr
       Ca_jsr_dev(ii) = Ca_jsr_dev(ii) + dt*(dtemp * dp_ca(8)) * (J_refill_dev(ii) - J_ryr)

       CRU_states_dev(ii) =  state_combine_single_ker(lcc_nt, ryr_nt, int_ca(10))
       !CRU_states_dev(ii) = (lcc_nt-1) * int_ca(10) + ryr_nt
       RYRgate_dev(ii) = nryr
       DHPRgate_dev(ii) = ndhpr
       Ca_ds_dev(ii) = s_cads(tx)
    END IF
  END SUBROUTINE update_SFU_Sun2000  
#endif !_LCC_MODEL_ == _SUN2000  

#if _LCC_MODEL_ == _WILLIAMS2013_NEWALGORITHM  
  attributes(global) SUBROUTINE get_compPdt_Williams2013()
    USE cudafor
    IMPLICIT NONE
    
    ! local
    REAL(KIND=dp) :: bigC2, chiC, chiO, phi, dtemp, dtemp2
    INTEGER:: tx, ii !,  ct   ! ct = current state
    REAL(KIND=dp), DIMENSION(BLOCKSIZE_1D), shared :: s_cads, s_ryrgate
    INTEGER,VOLATILE :: lcc_ct, ryr_ct
    INTEGER :: jj, fromto, from, to
    REAL(KIND=dp) :: dfactor
    
    ii =  (blockIdx%x-1) * blockDim%x + threadIdx%x
    
    IF (ii .LE. int_ca(1)) THEN
       tx = threadIdx%x
       s_cads(tx) = Ca_ds_dev(ii)
       s_ryrgate(tx) = RYRgate_dev(ii)
    ENDIF
    CALL syncthreads()
       
    IF (ii .LE. int_ca(1)) THEN
       CALL state_breakdown_ker(CRU_states_dev(ii), int_ca(10), lcc_ct, ryr_ct)
       !bigC2 = Ca_ds_dev(ii)**eta_RyR
       dtemp = LOG(s_cads(tx))
       bigC2 = EXP(dtemp*dp_ca(14))
       
       !chiC = EXP(-Ej*0.5d0*((N_R-RYRgate_dev(ii))*Ecc - (RYRgate_dev(ii)-1d0)*Eoo))
       dtemp = (int_ca(2)-s_ryrgate(tx))*dp_ca(2)
       dtemp2 = dtemp - (s_ryrgate(tx) - 1.d0)*dp_ca(3)
       dtemp = dp_ca(1)*dtemp2
       chiC = EXP(dtemp)
       
       !chiO = EXP(-Ej*0.5d0*(RYRgate_dev(ii)*Eoo - (N_R-RYRgate_dev(ii)- 1d0)*Ecc))
       dtemp = s_ryrgate(tx)*dp_ca(3)
       dtemp2 = dtemp - (int_ca(2) - s_ryrgate(tx) - 1d0)*dp_ca(2)
       dtemp = dp_ca(1)*dtemp2
       chiO = EXP(dtemp)

       !phi = k_jsr0*Ca_jsr_dev(ii) + k_jsr1
       phi = dp_ca(4)*Ca_jsr_dev(ii) + dp_ca(5)       
       
       ! use 1st column for negative value
       dtemp = chiC * ryr_compKfree_dev(ryr_ct,1)  
       dtemp2 = dtemp + bigC2 * chiO * phi * ryr_compKcal_act_dev(ryr_ct,1)
       DO jj=1, lcc_indxK_dev(lcc_ct,0)
          !integer :: from, to
          fromto = lcc_stateFromTo_dev(lcc_ct,jj)
          to = IBITS(fromto, 0,8)                
          from = IBITS(fromto, 8,8)
          !!no need to update thermo(6,7)
          IF  (from.EQ.4 .AND. to.EQ.7) THEN
             !!thermo(4,7)
             dfactor = (lcc_Krateconstant_dev(7,4)*Ca_ds_dev(ii)**dp_ca(15))*lcc_Ktransitionrate_dev(4,3)*lcc_Ktransitionrate_dev(3,2)*lcc_Ktransitionrate_dev(2,5)*lcc_Ktransitionrate_dev(5,7)/ &
                  (lcc_Ktransitionrate_dev(7,5)*lcc_Ktransitionrate_dev(5,2)*(lcc_Krateconstant_dev(2,3)*Ca_ds_dev(ii)**dp_ca(15))*lcc_Ktransitionrate_dev(3,4))
!!$NOTE: previously we don't updat thermo             dfactor = lcc_Ktransitionrate_dev(from,to)
          ELSE IF (from.EQ. 2 .AND. to.EQ. 3) THEN
             !the case of [Ca]ds-dependent 
             dfactor = lcc_Krateconstant_dev(2,3) * Ca_ds_dev(ii)**dp_ca(15)
          ELSEIF (from.EQ. 7 .AND. to.EQ. 4) THEN
             !the case of [Ca]ds-dependent 
             dfactor = lcc_Krateconstant_dev(7,4) * Ca_ds_dev(ii)**dp_ca(15)
          ELSE
             dfactor = lcc_Ktransitionrate_dev(from,to)
          ENDIF

          !NOTE: subtract to get negative value
          dtemp = dfactor * lcc_StateSpace_dev(lcc_ct,from)
          lcc_compK_dev(ii,jj) =  dtemp
          dtemp2 = dtemp2 - dtemp

       ENDDO
       compPdt_dev(ii) = dtemp2
    END IF
  END SUBROUTINE get_compPdt_Williams2013

  attributes(global) SUBROUTINE update_SFU_Williams2013(Xrnd, dt, &
       Vm, dp_arg1, dp_arg2, Ca_myo, Ca_nsr)
    !new method
    USE cudafor
    IMPLICIT NONE
    REAL(KIND=dp), DIMENSION(:), device :: Xrnd
    REAL(KIND=dp), VALUE :: dt, Ca_nsr, Ca_myo, Vm, dp_arg1, dp_arg2

    !local
    INTEGER :: ii, jj, kk, tx
    REAL(KIND=dp) :: dtemp, dtemp2, compP, J_ryr
    !INTEGER :: nt, ct, cnt
    !nt = new state, ct = current state, cnt= candidate next state
    REAL(KIND=dp) :: bigC1, bigC2, chiC, chiO, phi, J_dhpr
    REAL(KIND=dp) :: nryr ! new RYRgate
    REAL(KIND=dp) :: ndhpr !new dhprgate
    INTEGER :: itemp
    REAL(KIND=dp), DIMENSION(BLOCKSIZE_1D), shared :: s_cads, s_ryrgate
    INTEGER, VOLATILE :: lcc_ct, ryr_ct
    INTEGER :: lcc_nt, ryr_nt
    INTEGER :: is_examining
    LOGICAL :: found_nt
    INTEGER :: candidate_new_state !can be any cluster types (LCC or RyR)
  
    ! thread Index
    ii =  (blockIdx%x-1) * blockDim%x + threadIdx%x

    IF (ii .LE. int_ca(1)) THEN
       tx = threadIdx%x
       s_cads(tx) = Ca_ds_dev(ii)
       s_ryrgate(tx) = RYRgate_dev(ii)
    ENDIF
    CALL syncthreads()

    IF (ii .LE. int_ca(1)) THEN
       CALL state_breakdown_ker(CRU_states_dev(ii), int_ca(10), lcc_ct, ryr_ct)

       lcc_nt = lcc_ct
       ryr_nt = ryr_ct
       found_nt = .FALSE.
       compP = compPdt_dev(ii) * dt  + 1.0d0

       IF (compP .GE. Xrnd(ii)) THEN ! channel doesn't change state (easy)
          found_nt = .TRUE.
       ELSE ! channel change to some new state
          dtemp = LOG(s_cads(tx))
          dtemp2 = dtemp *dp_ca(14)
          bigC2 = EXP(dtemp2)

          dtemp = (int_ca(2)-s_ryrgate(tx))*dp_ca(2)
          dtemp2 = dtemp - (s_ryrgate(tx) - 1.d0)*dp_ca(3)
          dtemp = dp_ca(1)*dtemp2
          chiC = EXP(dtemp)
          !dtemp = (int_ca(2)-RYRgate_dev(ii))*dp_ca(2)
          !dtemp2 = dtemp - (RYRgate_dev(ii) - 1d0)*dp_ca(3)
          !dtemp = dp_ca(1)*dtemp2
          !chiC = EXP(dtemp)

          dtemp = s_ryrgate(tx)*dp_ca(3)
          dtemp2 = dtemp - (int_ca(2) - s_ryrgate(tx) - 1d0)*dp_ca(2)
          dtemp = dp_ca(1)*dtemp2
          chiO = EXP(dtemp)
          !dtemp = RYRgate_dev(ii)*dp_ca(3)
          !dtemp2= dtemp - (int_ca(2)-RYRgate_dev(ii)- 1d0)*dp_ca(2)
          !dtemp = dp_ca(1)*dtemp2
          !chiO = EXP(dtemp)

          phi = dp_ca(4)*Ca_jsr_dev(ii) + dp_ca(5)

          DO jj=2, int_ca(3)+int_ca(4)-1
             IF (found_nt .EQ. .FALSE.) THEN
                IF (jj .LE. int_ca(3)) THEN
                   ![Ca]ds-act RYR
                   kk = jj
                   dtemp = bigC2 * chiO * phi * ryr_compKcal_act_dev(ryr_ct,kk)
                   candidate_new_state = ryr_indxKcal_act_dev(ryr_ct,kk)
                ELSE IF (jj .LE. int_ca(3)+int_ca(4)-1) THEN
                   !inact RYR
                   kk = jj-(int_ca(3)-1)
                   dtemp = chiC * ryr_compKfree_dev(ryr_ct,kk)
                   candidate_new_state = ryr_indxKfree_dev(ryr_ct,kk)
                ELSE
                ENDIF
                compP = compP + dtemp * dt
                IF (compP .GE. Xrnd(ii)) THEN
                   found_nt = .TRUE.
                   ryr_nt = candidate_new_state
                   EXIT
                END IF
             ENDIF
          END DO
          IF (found_nt .EQ. .FALSE.) THEN
             DO jj=1, lcc_indxK_dev(lcc_ct,0)
                compP = compP + lcc_compK_dev(ii,jj) * dt
                IF (compP .GE. Xrnd(ii)) THEN
                   found_nt = .TRUE.
                   !to avoid LCC gating at negative potential
                   !IF (Vm .GE. -70.0d0)  lcc_nt = candidate_new_state
                   lcc_nt = lcc_indxK_dev(lcc_ct,jj)
                   EXIT !current DO-loop
                ENDIF
             ENDDO
          END IF
       END IF

       nryr = U_Ro_dev(ryr_nt)
       ndhpr = U_Lo_dev(lcc_nt)

       J_efflux_dev(ii) = dp_ca(11)*(s_cads(tx)-Ca_myo)
       J_refill_dev(ii) = dp_ca(10)*(Ca_nsr-Ca_jsr_dev(ii))

#if (_RUNMODE_ == _MODE_I_CLAMP || _RUNMODE_ == _MODE_V_CLAMP || _RUNMODE_ == _MODE_V_CLAMP_TEST_LCC)
       !I-clamp or V-clamp
       I_dhpr_dev(ii)  = ndhpr * dp_ca(13) * dp_ca(16) * Vm * &
            (s_cads(tx) * dp_arg1 - dp_ca(18)) * dp_arg2  ! [uA]
       !J_dhpr_dev(ii)  = -I_dhpr_dev(ii) * dp_ca(17)  ! [uA/s]
       J_dhpr  = -I_dhpr_dev(ii) * dp_ca(17)  ! [uA/s]
#else
       I_dhpr_dev(ii)  = 0.0d0
       !J_dhpr_dev(ii) = 0d0       
       J_dhpr = 0d0
#endif

       !!J_ryr = nryr * dp_ca(9) * (Ca_jsr_dev(ii)-Ca_ds_dev(ii)) !J_ryr
       J_ryr = nryr * dp_ca(9) * (Ca_jsr_dev(ii)-s_cads(tx))
       J_ryr_dev(ii) = J_ryr

       !exact formula
       !s_cads(tx) = s_cads(tx) + dt * dp_ca(6) * dp_ca(7) * (J_ryr - J_efflux_dev(ii) + J_dhpr_dev(ii))
       s_cads(tx) = s_cads(tx) + dt * dp_ca(6) * dp_ca(7) * (J_ryr - J_efflux_dev(ii) + J_dhpr)  
       
       dtemp = dp_ca(19) + Ca_jsr_dev(ii)
       dtemp2 = dtemp * dtemp
       dtemp = dtemp2 / (dp_ca(19)*dp_ca(20) + dtemp2)   !beta_jsr
       Ca_jsr_dev(ii) = Ca_jsr_dev(ii) + dt*(dtemp * dp_ca(8)) * (J_refill_dev(ii) - J_ryr)

       CRU_states_dev(ii) =  state_combine_single_ker(lcc_nt, ryr_nt, int_ca(10))
       !CRU_states_dev(ii) = (lcc_nt-1) * int_ca(10) + ryr_nt
       RYRgate_dev(ii) = nryr
       DHPRgate_dev(ii) = ndhpr
       Ca_ds_dev(ii) = s_cads(tx)
    END IF
  END SUBROUTINE update_SFU_Williams2013
#endif !_LCC_MODEL_ == _WILLIAMS2013_NEWALGORITHM
  
END MODULE ecc_gpu_utility

